import time

from solace.messaging.errors.pubsubplus_client_error import PubSubPlusClientError
from solace.messaging.utils.manageable import Metric

from pypubsubtools.core.rpc_parameter_keys import RpcParameterKeys
from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.core.runtime_property_builder import RuntimePropertyBuilder
from pypubsubtools.utils.log import get_logger
from sdkperf.config.constants import Constants
from sdkperf.config.endpoint_access_type import EndpointAccessType
from sdkperf.config.ep_config_properties import EpConfigProperties
from sdkperf.config.flow_status import FlowStatus
from sdkperf.config.subscriber_destinations_type import SubscriberDestinationsType
from sdkperf.core.last_message_details import LastMessageDetails
from sdkperf.core.sdkperf_client_collection import SdkPerfClientCollection

logger = get_logger()

"""
TO TEST WITH PYTHON FROM COMMAND LINE
Run the server from main below

import xmlrpc.client
proxy = xmlrpc.client.ServerProxy("http://localhost:5678/")
proxy.configure({"a":"b"})
Etc.
"""


class PerfClientProxy:

    def __init__(self, unique_id, tool_mode):
        self._id = unique_id
        self._tm = tool_mode
        self._pb = RuntimePropertyBuilder()
        self._collection = None

    def destroy(self):
        if self._collection is not None:
            self._collection.destroy()

    def _is_init(self):
        return self._collection is not None

    def _convert_last_message_details_to_dict(self, msg):
        table = dict()
        if msg.get_destination() is not None:
            table["destination"] = msg.get_destination()
        if msg.get_destination_type() is not None:
            table["destinationType"] = msg.get_destination_type()
        if msg.get_http_content_type() is not None:
            table["httpContentType"] = msg.get_http_content_type()
        if msg.get_http_content_encoding() is not None:
            table["httpContentEncoding"] = msg.get_http_content_encoding()
        if msg.get_correlation_id() is not None:
            table["correlationId"] = msg.get_correlation_id()
        if msg.get_expiration() is not None:
            table["expiration"] = msg.get_expiration()
        if msg.get_priority() is not None:
            table["priority"] = msg.get_priority()
        if msg.get_application_message_id() is not None:
            table["applicationMessageId"] = msg.get_application_message_id()
        if msg.get_application_message_type() is not None:
            table["applicationMessageType"] = msg.get_application_message_type()
        if msg.get_sender_timestamp() is not None:
            table["senderTimestamp"] = msg.get_sender_timestamp()
        if msg.get_redelivered() is not None:
            table["redelivered"] = msg.get_redelivered()
        if msg.get_cos() is not None:
            table["cos"] = msg.get_cos()
        if msg.get_creation_trace_id() is not None:
            table["creationContextTraceId"] = msg.get_creation_trace_id()
        if msg.get_creation_span_id() is not None:
            table["creationContextSpanId"] = msg.get_creation_span_id()
        if msg.get_creation_sampled() is not None:
            table["creationContextSampled"] = msg.get_creation_sampled()
        if msg.get_creation_trace_state() is not None:
            table["creationContextTraceState"] = msg.get_creation_trace_state()
        if msg.get_transport_trace_id() is not None:
            table["transportContextTraceId"] = msg.get_transport_trace_id()
        if msg.get_transport_span_id() is not None:
            table["transportContextSpanId"] = msg.get_transport_span_id()
        if msg.get_transport_sampled() is not None:
            table["transportContextSampled"] = msg.get_transport_sampled()
        if msg.get_transport_trace_state() is not None:
            table["transportContextTraceState"] = msg.get_transport_trace_state()
        if msg.get_baggage() is not None:
            table["baggage"] = msg.get_baggage()
        if msg.get_binary_payload_size() is not None:
            table["binaryPayloadSize"] = msg.get_binary_payload_size()
        # Add more message properties below here

        return table

    def Configure(self, username: str, password: str, rtrHost: str, dummyInt: int, dummyBool: bool, keepAlive: int,
                  reconAtt: int, reconInterval: int):
        self._pb.set_client_username(username)
        self._pb.set_client_password(password)

        self._pb.set_client_ip_addr(rtrHost)

        if keepAlive != -1:
            self._pb.set_keepalive_interval_msec(keepAlive)

        self._pb.set_reconnect_attempts(reconAtt)
        self._pb.set_reconnect_interval_msec(reconInterval)
        return Constants.SUCCESS_RETURN

    def CustomConfigure(self, params: dict):
        if RpcParameterKeys.CLIENT_MODE in params:
            self._pb.set_client_mode(params[RpcParameterKeys.CLIENT_MODE])
        if RpcParameterKeys.CLIENT_PASSWORD_FILE in params:
            current_pass = getattr(self._pb, RuntimeProperties.CLIENT_PASSWORD, None)
            if current_pass is None or len(current_pass) == 0:
                try:
                    with open(params[RpcParameterKeys.CLIENT_PASSWORD_FILE], "r") as pass_file:
                        self._pb.set_client_password(pass_file.read())
                except Exception as e:
                    logger.warning(
                        f"Failed to load password from file '{params[RpcParameterKeys.CLIENT_PASSWORD_FILE]}' {str(e)}")
        if RpcParameterKeys.CLIENT_COUNT in params:
            self._pb.set_num_clients(params[RpcParameterKeys.CLIENT_COUNT])
        if RpcParameterKeys.PUBS_PER_SESSION in params:
            self._pb.set_num_pubs_per_session(params[RpcParameterKeys.PUBS_PER_SESSION])
        if RpcParameterKeys.CLIENT_NAME_PREFIX in params:
            self._pb.set_client_name_prefix(params[RpcParameterKeys.CLIENT_NAME_PREFIX])
        if RpcParameterKeys.CLIENT_VPN in params:
            self._pb.set_client_vpn(params[RpcParameterKeys.CLIENT_VPN])
        if RpcParameterKeys.CLIENT_DESCRIPTION in params:
            self._pb.set_client_description(params[RpcParameterKeys.CLIENT_DESCRIPTION])
        if RpcParameterKeys.CLIENT_COMPRESSION_LEVEL in params and \
                params[RpcParameterKeys.CLIENT_COMPRESSION_LEVEL] != -1:
            self._pb.set_client_compression_level(params[RpcParameterKeys.CLIENT_COMPRESSION_LEVEL])
        if RpcParameterKeys.PAYLOAD_COMPRESSION_LEVEL in params and \
                params[RpcParameterKeys.PAYLOAD_COMPRESSION_LEVEL] != -1:
            self._pb.set_payload_compression_level(params[RpcParameterKeys.PAYLOAD_COMPRESSION_LEVEL])
        if RpcParameterKeys.WANT_PER_MESSAGE_DETAILS in params:
            self._pb.set_want_per_message_details(params[RpcParameterKeys.WANT_PER_MESSAGE_DETAILS])
        if RpcParameterKeys.LAST_MESSAGES_QUEUE_DEPTH in params:
            self._pb.set_want_per_message_details(True)
            self._pb.set_last_message_queue_depth(params[RpcParameterKeys.LAST_MESSAGES_QUEUE_DEPTH])
        if RpcParameterKeys.WANT_MESSAGE_PRIORITY_STATS in params:
            self._pb.set_want_message_priority_stats(params[RpcParameterKeys.WANT_MESSAGE_PRIORITY_STATS])
        if RpcParameterKeys.REQUEST_REPLY_WAIT_TIME in params:
            self._pb.set_request_reply_wait_time(params[RpcParameterKeys.REQUEST_REPLY_WAIT_TIME])
        if RpcParameterKeys.AD_PUB_WINDOW in params and params[RpcParameterKeys.AD_PUB_WINDOW] != -1:
            self._pb.set_ad_pub_window_size(params[RpcParameterKeys.AD_PUB_WINDOW])
        if RpcParameterKeys.AD_PUB_ACKTIME in params and params[RpcParameterKeys.AD_PUB_ACKTIME] != -1:
            self._pb.set_ad_pub_ack_timeout(params[RpcParameterKeys.AD_PUB_ACKTIME])
        if RpcParameterKeys.PUB_MAX_RESENDS in params and params[RpcParameterKeys.PUB_MAX_RESENDS] != -1:
            self._pb.set_ad_pub_max_resend(params[RpcParameterKeys.PUB_MAX_RESENDS])

        if RpcParameterKeys.AD_SUB_WINDOW in params and params[RpcParameterKeys.AD_SUB_WINDOW] != -1:
            self._pb.set_ad_sub_window_size(params[RpcParameterKeys.AD_SUB_WINDOW])
        if RpcParameterKeys.AD_SUB_ACKTIME in params and params[RpcParameterKeys.AD_SUB_ACKTIME] != -1:
            self._pb.set_ad_sub_ack_timeout(params[RpcParameterKeys.AD_SUB_ACKTIME])
        if RpcParameterKeys.AD_WIN_THRESH in params and params[RpcParameterKeys.AD_WIN_THRESH] != -1:
            self._pb.set_ad_sub_window_threshold(params[RpcParameterKeys.AD_WIN_THRESH])

        if RpcParameterKeys.WANT_SMOOTH_PUB in params:
            self._pb.set_want_smooth_pub_calc_lat(params[RpcParameterKeys.WANT_SMOOTH_PUB])
        elif RpcParameterKeys.PSM_CALC_LAT in params:
            self._pb.set_want_smooth_pub_calc_lat(params[RpcParameterKeys.PSM_CALC_LAT])
        else:
            self._pb.set_want_smooth_pub_calc_lat(False)

        if RpcParameterKeys.WANT_LATENCY in params:
            self._pb.set_want_latency(params[RpcParameterKeys.WANT_LATENCY])
        if RpcParameterKeys.SUB_LATENCY_BUCKETS in params:
            self._pb.set_latency_buckets(params[RpcParameterKeys.SUB_LATENCY_BUCKETS])
        if RpcParameterKeys.SUB_LATENCY_GRANULARITY in params:
            self._pb.set_latency_granularity(params[RpcParameterKeys.SUB_LATENCY_GRANULARITY])
        if RpcParameterKeys.SUB_LATENCY_WARMUP in params:
            self._pb.set_latency_warmup_in_secs(params[RpcParameterKeys.SUB_LATENCY_WARMUP])
        if RpcParameterKeys.WANT_PAYLOAD_CHECK in params:
            self._pb.set_want_payload_check(params[RpcParameterKeys.WANT_PAYLOAD_CHECK])
        if RpcParameterKeys.WANT_ORDER_CHECK in params:
            self._pb.set_want_order_check(params[RpcParameterKeys.WANT_ORDER_CHECK])
        if RpcParameterKeys.WANT_ORDER_MEMORY in params:
            self._pb.set_want_order_memory(params[RpcParameterKeys.WANT_ORDER_MEMORY])
        if RpcParameterKeys.SUB_WANT_FULL_INSPECTION in params:
            self._pb.set_want_message_examine_content(params[RpcParameterKeys.SUB_WANT_FULL_INSPECTION])
        if RpcParameterKeys.TCP_NO_DELAY in params:
            self._pb.set_want_tcp_nodelay(params[RpcParameterKeys.TCP_NO_DELAY])
        if RpcParameterKeys.SUB_WANT_PERCID_STATS in params:
            self._pb.set_cid_stats(params[RpcParameterKeys.SUB_WANT_PERCID_STATS])
        if RpcParameterKeys.WANT_DI_STATS in params:
            self._pb.set_want_di_stats(params[RpcParameterKeys.WANT_DI_STATS])
        if RpcParameterKeys.WANT_USER_PROP_TOOL_DATA in params:
            self._pb.set_want_user_prop_tool_data(params[RpcParameterKeys.WANT_USER_PROP_TOOL_DATA])
        if RpcParameterKeys.WANT_NON_BLOCKING_CONNECT in params:
            self._pb.set_want_non_blocking_connect(params[RpcParameterKeys.WANT_NON_BLOCKING_CONNECT])

        if RpcParameterKeys.WANT_STOP_ON_ERROR in params:
            self._pb.set_want_stop_on_error(params[RpcParameterKeys.WANT_STOP_ON_ERROR])
        if RpcParameterKeys.SUB_SLOW_DELAY_MILLIS in params:
            self._pb.set_subscriber_sub_delay(params[RpcParameterKeys.SUB_SLOW_DELAY_MILLIS])

        if RpcParameterKeys.SUB_SLOW_DELAY_COUNT in params:
            self._pb.set_slow_sub_delay_count(params[RpcParameterKeys.SUB_SLOW_DELAY_COUNT])

        if RpcParameterKeys.SUBSCRIBER_RATE_INTERVAL in params:
            self._pb.set_subscriber_rate_interval(params[RpcParameterKeys.SUBSCRIBER_RATE_INTERVAL])
        else:
            self._pb.set_subscriber_rate_interval(1)

        if RpcParameterKeys.CLIENT_WANT_AUTO_GEN_SDM in params:
            self._pb.set_want_struct_msg_check(params[RpcParameterKeys.CLIENT_WANT_AUTO_GEN_SDM])
        elif RpcParameterKeys.CLIENT_WANT_STRUCT_MSG_CHECK in params:
            self._pb.set_want_struct_msg_check(params[RpcParameterKeys.CLIENT_WANT_STRUCT_MSG_CHECK])
        else:
            self._pb.set_want_struct_msg_check(False)

        if RpcParameterKeys.SDM_FILE_ARRAY in params:
            self._pb.set_sdm_prop_list_from_files(params[RpcParameterKeys.SDM_FILE_ARRAY])
            self._pb.set_want_struct_msg_check(True)
        if RpcParameterKeys.SDM_STRINGS_ARRAY in params:
            self._pb.set_sdm_prop_list_from_strings(params[RpcParameterKeys.SDM_STRINGS_ARRAY])
            self._pb.set_want_struct_msg_check(True)

        if RpcParameterKeys.REAPPLY_SUBSCRIPTIONS in params:
            self._pb.set_want_subscription_memory(params[RpcParameterKeys.REAPPLY_SUBSCRIPTIONS])
        if RpcParameterKeys.IGNORE_EXISTS_ERRORS in params:
            self._pb.set_ignore_exists_errors(params[RpcParameterKeys.IGNORE_EXISTS_ERRORS])
        if RpcParameterKeys.NO_LOCAL in params:
            self._pb.set_want_no_local(params[RpcParameterKeys.NO_LOCAL])

        if RpcParameterKeys.WANT_CLIENT_ACK in params:
            self._pb.set_want_client_ack(params[RpcParameterKeys.WANT_CLIENT_ACK])
        if RpcParameterKeys.WANT_CLIENT_ACK_THREAD in params:
            self._pb.set_want_client_ack_thread(params[RpcParameterKeys.WANT_CLIENT_ACK_THREAD])
        if RpcParameterKeys.CLIENT_ACK_SKIP_NUM in params and params[RpcParameterKeys.CLIENT_ACK_SKIP_NUM] != -1:
            self._pb.set_client_ack_skip_num(params[RpcParameterKeys.CLIENT_ACK_SKIP_NUM])
        if RpcParameterKeys.CLIENT_ACK_QUEUE_DEPTH in params and params[RpcParameterKeys.CLIENT_ACK_QUEUE_DEPTH] != -1:
            self._pb.set_sub_msg_queue_depth(params[RpcParameterKeys.CLIENT_ACK_QUEUE_DEPTH])
        if RpcParameterKeys.CLIENT_ACK_QUEUE_FLUSH in params:
            self._pb.set_client_ack_flush_queue(params[RpcParameterKeys.CLIENT_ACK_QUEUE_FLUSH])
        if RpcParameterKeys.CLIENT_ACK_QUEUE_REVERSE in params:
            self._pb.set_client_ack_queue_reverse(params[RpcParameterKeys.CLIENT_ACK_QUEUE_REVERSE])
        if RpcParameterKeys.CLIENT_ACK_RANDOM_DEPTH in params:
            self._pb.set_client_ack_random_depth(params[RpcParameterKeys.CLIENT_ACK_RANDOM_DEPTH])
        if RpcParameterKeys.AD_ACK_EVENT_MODE in params:
            self._pb.set_ad_ack_event_mode(params[RpcParameterKeys.AD_ACK_EVENT_MODE])
        if RpcParameterKeys.WANT_PUB_ACK_ORDER_CHECK in params:
            self._pb.set_ad_want_pub_ack_order_check(params[RpcParameterKeys.WANT_PUB_ACK_ORDER_CHECK])
        if RpcParameterKeys.REQUIRED_SETTLEMENTS_STRING in params:
            self._pb.set_required_settlement_string(params[RpcParameterKeys.REQUIRED_SETTLEMENTS_STRING])
        if RpcParameterKeys.SETTLEMENT_PATTERN_STRING in params:
            self._pb.set_settlement_pattern_string(params[RpcParameterKeys.SETTLEMENT_PATTERN_STRING])

        if RpcParameterKeys.SSL_PROTOCOL in params and len(params[RpcParameterKeys.SSL_PROTOCOL]) > 0:
            self._pb.set_ssl_protocol(params[RpcParameterKeys.SSL_PROTOCOL])
        if RpcParameterKeys.SSL_CONNECTION_DOWNGRADE_TO in params:
            self._pb.set_ssl_connection_downgrade_to(params[RpcParameterKeys.SSL_CONNECTION_DOWNGRADE_TO])
        if RpcParameterKeys.SSL_EXCLUDED_PROTOCOL in params and len(params[RpcParameterKeys.SSL_EXCLUDED_PROTOCOL]) > 0:
            self._pb.set_ssl_excluded_protocols(params[RpcParameterKeys.SSL_EXCLUDED_PROTOCOL])
        if RpcParameterKeys.SSL_VALIDATE_CERTIFICATE in params:
            self._pb.set_ssl_validate_certificate(params[RpcParameterKeys.SSL_VALIDATE_CERTIFICATE])
        if RpcParameterKeys.SSL_VALIDATE_CERTIFICATE_DATE in params:
            self._pb.set_ssl_validate_certificate_date(params[RpcParameterKeys.SSL_VALIDATE_CERTIFICATE_DATE])
        if RpcParameterKeys.SSL_VALIDATE_CERTIFICATE_HOSTNAME in params:
            self._pb.set_ssl_validate_certificate_hostname(params[RpcParameterKeys.SSL_VALIDATE_CERTIFICATE_HOSTNAME])
        if RpcParameterKeys.SSL_CIPHER_SUITES in params and params[RpcParameterKeys.SSL_CIPHER_SUITES]:
            self._pb.set_ssl_cipher_suites(params[RpcParameterKeys.SSL_CIPHER_SUITES])
        if RpcParameterKeys.SSL_TRUST_STORE in params and params[RpcParameterKeys.SSL_TRUST_STORE]:
            self._pb.set_ssl_trust_store(params[RpcParameterKeys.SSL_TRUST_STORE])
        if RpcParameterKeys.SSL_TRUST_STORE_DIR in params:
            self._pb.set_ssl_trust_store_dir(params[RpcParameterKeys.SSL_TRUST_STORE_DIR])
        if RpcParameterKeys.SSL_CLIENT_CERTIFICATE_FILE in params:
            self._pb.set_ssl_client_certificate_file(params[RpcParameterKeys.SSL_CLIENT_CERTIFICATE_FILE])
        if RpcParameterKeys.SSL_CLIENT_PRIVATE_KEY_FILE in params:
            self._pb.set_ssl_client_private_key_file(params[RpcParameterKeys.SSL_CLIENT_PRIVATE_KEY_FILE])
        if RpcParameterKeys.SSL_CLIENT_PRIVATE_KEY_FILE_PASSWORD in params and \
                params[RpcParameterKeys.SSL_CLIENT_PRIVATE_KEY_FILE_PASSWORD]:
            self._pb.set_ssl_client_private_key_file_password(
                params[RpcParameterKeys.SSL_CLIENT_PRIVATE_KEY_FILE_PASSWORD])
        if RpcParameterKeys.SSL_TRUST_STORE_PASSWORD in params and params[RpcParameterKeys.SSL_TRUST_STORE_PASSWORD]:
            self._pb.set_ssl_trust_store_password(params[RpcParameterKeys.SSL_TRUST_STORE_PASSWORD])
        if RpcParameterKeys.SSL_TRUST_STORE_FORMAT in params and params[RpcParameterKeys.SSL_TRUST_STORE_FORMAT]:
            self._pb.set_ssl_trust_store_format(params[RpcParameterKeys.SSL_TRUST_STORE_FORMAT])
        if RpcParameterKeys.SSL_TRUSTED_COMMON_NAME_LIST in params and \
                params[RpcParameterKeys.SSL_TRUSTED_COMMON_NAME_LIST]:
            self._pb.set_ssl_trusted_common_name_list(params[RpcParameterKeys.SSL_TRUSTED_COMMON_NAME_LIST])
        if RpcParameterKeys.SSL_KEY_STORE in params and params[RpcParameterKeys.SSL_KEY_STORE]:
            self._pb.set_ssl_key_store(params[RpcParameterKeys.SSL_KEY_STORE])
        if RpcParameterKeys.SSL_KEY_STORE_PASSWORD in params and params[RpcParameterKeys.SSL_KEY_STORE_PASSWORD]:
            self._pb.set_ssl_key_store_password(params[RpcParameterKeys.SSL_KEY_STORE_PASSWORD])
        if RpcParameterKeys.SSL_KEY_STORE_FORMAT in params and params[RpcParameterKeys.SSL_KEY_STORE_FORMAT]:
            self._pb.set_ssl_key_store_format(params[RpcParameterKeys.SSL_KEY_STORE_FORMAT])
        if RpcParameterKeys.SSL_PRIVATE_KEY_ALIAS in params and params[RpcParameterKeys.SSL_PRIVATE_KEY_ALIAS]:
            self._pb.set_ssl_private_key_alias(params[RpcParameterKeys.SSL_PRIVATE_KEY_ALIAS])
        if RpcParameterKeys.SSL_PRIVATE_KEY_PASSWORD in params and params[RpcParameterKeys.SSL_PRIVATE_KEY_PASSWORD]:
            self._pb.set_ssl_private_key_password(params[RpcParameterKeys.SSL_PRIVATE_KEY_PASSWORD])
        if RpcParameterKeys.AUTHENTICATION_SCHEME in params:
            self._pb.set_authentication_scheme(params[RpcParameterKeys.AUTHENTICATION_SCHEME])

        if RpcParameterKeys.OAUTH_ACCESS_TOKEN in params:
            self._pb.set_oauth_access_token(params[RpcParameterKeys.OAUTH_ACCESS_TOKEN])
        if RpcParameterKeys.OAUTH_ISSUER_ID in params:
            self._pb.set_oauth_issuer_id(params[RpcParameterKeys.OAUTH_ISSUER_ID])
        if RpcParameterKeys.OAUTH_ID_TOKEN in params:
            self._pb.set_oauth_id_token(params[RpcParameterKeys.OAUTH_ID_TOKEN])

        if RpcParameterKeys.NUM_MESSAGE_PRODUCERS in params:
            self._pb.set_num_message_producer(params[RpcParameterKeys.NUM_MESSAGE_PRODUCERS])

        if RpcParameterKeys.PUBLISHER_BACKPRESSURE_TYPE in params:
            # The publisher bp should be a list so split on ","
            self._pb.set_publisher_backpressure_type(
                params[RpcParameterKeys.PUBLISHER_BACKPRESSURE_TYPE].split(sep=","))
        if RpcParameterKeys.PUBLISHER_BACKPRESSURE_BUFFER in params:
            self._pb.set_publisher_backpressure_buffer_size(params[RpcParameterKeys.PUBLISHER_BACKPRESSURE_BUFFER])

        if RpcParameterKeys.RECEIVER_BACKPRESSURE_TYPE in params:
            self._pb.set_subscriber_backpressure_type(params[RpcParameterKeys.RECEIVER_BACKPRESSURE_TYPE])
        if RpcParameterKeys.RECEIVER_BACKPRESSURE_BUFFER in params:
            self._pb.set_subscriber_backpressure_buffer_size(params[RpcParameterKeys.RECEIVER_BACKPRESSURE_BUFFER])
        if RpcParameterKeys.RECEIVER_SLEEP_TIME in params:
            self._pb.set_receiver_sleep_time(params[RpcParameterKeys.RECEIVER_SLEEP_TIME])
        if RpcParameterKeys.DISABLE_TERMINATION_LISTENERS in params:
            self._pb.set_want_termination_listeners(not params[RpcParameterKeys.DISABLE_TERMINATION_LISTENERS])
        if RpcParameterKeys.WANT_UNGRACEFUL_TERMINATE in params:
            self._pb.set_want_ungraceful_terminate(params[RpcParameterKeys.WANT_UNGRACEFUL_TERMINATE])
        if RpcParameterKeys.WANT_UNGRACEFUL_DISCONNECT in params:
            self._pb.set_want_ungraceful_disconnect(params[RpcParameterKeys.WANT_UNGRACEFUL_DISCONNECT])

        # In python we cannot specify anything about the reply-to topic so if it's wanted then we're just gonna use it
        if RpcParameterKeys.WANT_REPLY_TO_TOPIC in params:
            self._pb.set_want_reply_topic(params[RpcParameterKeys.WANT_REPLY_TO_TOPIC])
        if RpcParameterKeys.WANT_REPLY_TO_STRING in params and params[RpcParameterKeys.WANT_REPLY_TO_STRING] != "":
            self._pb.set_want_reply_topic(True)
        if RpcParameterKeys.WANT_REPLY_TO_POSTFIX in params and params[RpcParameterKeys.WANT_REPLY_TO_POSTFIX] != "":
            self._pb.set_want_reply_topic(True)

        if RpcParameterKeys.CACHE_PROP_NAME in params:
            self._pb.set_cache_prop_name(params[RpcParameterKeys.CACHE_PROP_NAME])
            self._pb.set_cache_want_stats(True)
        else:
            self._pb.set_cache_want_stats(False)
        if RpcParameterKeys.CACHE_PROP_MAX_MSGS_PER_TOPIC in params \
                and params[RpcParameterKeys.CACHE_PROP_MAX_MSGS_PER_TOPIC] != -1:
            self._pb.set_cache_prop_max_msgs_per_topic(params[RpcParameterKeys.CACHE_PROP_MAX_MSGS_PER_TOPIC])
        if RpcParameterKeys.CACHE_PROP_MAX_AGE in params \
                and params[RpcParameterKeys.CACHE_PROP_MAX_AGE] != -1:
            self._pb.set_cache_prop_max_age(params[RpcParameterKeys.CACHE_PROP_MAX_AGE])
        if RpcParameterKeys.CACHE_PROP_TIMEOUT_IN_MSEC in params \
                and params[RpcParameterKeys.CACHE_PROP_TIMEOUT_IN_MSEC] != -1:
            self._pb.set_cache_prop_timeout_in_msec(params[RpcParameterKeys.CACHE_PROP_TIMEOUT_IN_MSEC])

        if RpcParameterKeys.TRACE_EXPORTER_CONNECTION_STRING in params:
            self._pb.set_trace_exporter_connection_string(params[RpcParameterKeys.TRACE_EXPORTER_CONNECTION_STRING])
            self._pb.set_want_trace_context_prop(True)
        if RpcParameterKeys.TRACE_COPY_RECEIVE_CONTEXT in params \
                and params[RpcParameterKeys.TRACE_COPY_RECEIVE_CONTEXT]:
            self._pb.set_want_trace_copy_receive_context(True)
            self._pb.set_want_trace_context_prop(True)
        if RpcParameterKeys.MESSAGE_OBJECT_TYPE_CHECK in params:
            self._pb.set_want_msg_obj_type_check(params[RpcParameterKeys.MESSAGE_OBJECT_TYPE_CHECK])

        # Sometimes the API properties we want to set are passed in through the -extraPropsList
        # arg instead of their specific argument.
        if RpcParameterKeys.EXTRA_PARAMETER_LIST in params:
            if len(params[RpcParameterKeys.EXTRA_PARAMETER_LIST]) % 2 == 0:
                for i in range(0, len(params[RpcParameterKeys.EXTRA_PARAMETER_LIST]), 2):
                    prop_name = params[RpcParameterKeys.EXTRA_PARAMETER_LIST][i]
                    prop_val = params[RpcParameterKeys.EXTRA_PARAMETER_LIST][i + 1]
                    if prop_name == "solace.messaging.authentication.kerberos.instance-name":
                        prop_val = str(prop_val)
                        self._pb.set_krb_service_name(prop_val)
                    elif prop_name == "solace.messaging.transport.keep-alive-interval":
                        prop_val = int(prop_val)
                        self._pb.set_keepalive_interval_msec(prop_val)
                    elif prop_name == "solace.messaging.transport.reconnection-attempts":
                        prop_val = int(prop_val)
                        self._pb.set_reconnect_attempts(prop_val)
                    elif prop_name == "solace.messaging.transport.reconnection-attempts-wait-interval":
                        prop_val = int(prop_val)
                        self._pb.set_reconnect_interval_msec(prop_val)

        if self._tm == "sdkperf":
            try:
                self._collection = SdkPerfClientCollection(self._pb.props)
            except Exception as e:
                logger.error(f"CustomConfigure: {str(e)}")
                return {"rc": 400, "errorInfo": str(e)}
        else:
            return {"rc": 400, "errorInfo": "Tool Mode not understood"}
        return Constants.SUCCESS_RETURN

    def Connect(self):
        try:
            self._collection.connect()
        except Exception as e:
            logger.error(f"Connect: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def Disconnect(self):
        try:
            self._collection.disconnect()
        except Exception as e:
            logger.error(f"Disconnect: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def Start(self):
        try:
            self._collection.start()
        except Exception as e:
            logger.error(f"Start: {str(e)}")
            return {"rc": 1, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def Stop(self):
        try:
            self._collection.stop()
        except Exception as e:
            logger.error(f"Stop: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def OpenTransactedSession(self, params: dict):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}
        transaction_prop_builder = RuntimePropertyBuilder.create_default_properties()
        client_index = Constants.ALL_CLIENT_INDEX

        table = dict(Constants.SUCCESS_RETURN)
        if RpcParameterKeys.CLIENT_IDX in params:
            client_index = params[RpcParameterKeys.CLIENT_IDX]
        if RpcParameterKeys.TRANSACTION_SIZE in params:
            transaction_prop_builder.set_ad_transaction_size(params[RpcParameterKeys.TRANSACTION_SIZE])
        if RpcParameterKeys.TRANSACTION_IDLE_TIME in params:
            transaction_prop_builder.set_ad_transaction_idle_time(params[RpcParameterKeys.TRANSACTION_IDLE_TIME])
        if RpcParameterKeys.TRANSACTION_ROLLBACK_INTERVAL in params:
            transaction_prop_builder.set_ad_transaction_rollback_interval(
                params[RpcParameterKeys.TRANSACTION_ROLLBACK_INTERVAL])
        try:
            session_name = self._collection.open_transacted_session(client_index, transaction_prop_builder.props)
            table["sessionName"] = session_name
        except Exception as e:
            logger.error(f"OpenTransactedSession: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return table

    def CloseTransactedSession(self, session_name: str, client_index: int):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}

        try:
            self._collection.close_transacted_session(client_index, session_name)
        except Exception as e:
            logger.error(f"CloseTransactedSession: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def CloseAllTransactedSessions(self, client_index: int):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}

        try:
            self._collection.close_all_transacted_sessions(client_index)
        except Exception as e:
            logger.error(f"CloseAllTransactedSession: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def CommitTransaction(self, session_name: str, want_rollback: bool, client_index: int):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}

        try:
            self._collection.commit_transaction(client_index, session_name, want_rollback)
        except Exception as e:
            logger.error(f"CommitTransaction: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def GetLastMessageDetails(self, client_index: int):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}
        table = dict()
        try:
            msg: LastMessageDetails = self._collection.get_last_message_details(client_index)
            table = self._convert_last_message_details_to_dict(msg)
        except Exception as e:
            logger.error(f"GetLastMessageDetails: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        table["rc"] = 0

        return table

    def GetLastMessagesDetails(self, client_index: int, msgs_to_get: int):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}
        table = dict()
        try:
            msgs = list()
            for msg in self._collection.get_last_messages_details(client_index, msgs_to_get):
                msgs.append(self._convert_last_message_details_to_dict(msg))
            table["details"] = msgs
        except Exception as e:
            logger.error(f"GetLastMessageDetails: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        table["rc"] = 0

        return table

    @staticmethod
    def SetClientName(index: int, name: str):
        # We are not able to change the name of the client after it has been created so this is basically a noop
        return Constants.SUCCESS_RETURN

    def UpdateAuthOnReconnect(self, params: dict):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}

        client_index = -1
        access_token = None
        id_token = None

        if RpcParameterKeys.CLIENT_IDX in params:
            client_index = params[RpcParameterKeys.CLIENT_IDX]
        if RpcParameterKeys.OAUTH_ACCESS_TOKEN in params:
            access_token = params[RpcParameterKeys.OAUTH_ACCESS_TOKEN]
        if RpcParameterKeys.OAUTH_ID_TOKEN in params:
            id_token = params[RpcParameterKeys.OAUTH_ID_TOKEN]

        try:
            self._collection.update_auth_on_reconnect(client_index, access_token, id_token)
        except Exception as e:
            logger.error(f"UpdateAuthOnReconnect: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}

        return Constants.SUCCESS_RETURN

    def GetLastError(self, index: int):
        table = dict(Constants.SUCCESS_RETURN)
        try:
            ex = self._collection.get_last_exception(index)
        except Exception as e:
            return {"rc": 400, "errorInfo": str(e)}
        if ex is None:
            table["subCode"] = ""
            table["rtrRespCode"] = ""
            table["errorStr"] = ""
        else:
            if isinstance(ex, PubSubPlusClientError):
                table["subCode"] = ex.sub_code if ex.sub_code is not None else "UNKNOWN"
                table["rtrRespCode"] = "0"
                table["errorStr"] = str(ex)
            else:
                table["subCode"] = "UNKNOWN"
                table["rtrRespCode"] = "0"
                table["errorStr"] = str(ex)
        return table

    def ClearLastError(self, index: int):
        self._collection.clear_last_exception(index)
        return Constants.SUCCESS_RETURN

    def GetClientName(self, index: int):
        if self._collection is not None:
            table = dict(Constants.SUCCESS_RETURN)
            table["client-names"] = self._collection.get_client_name(index)
            return table
        else:
            return {'rc': 100, "errorInfo": "PerfProxy not configured."}

    def GetChannelState(self):
        if self._collection is not None:
            table = dict(Constants.SUCCESS_RETURN)
            table["state"] = self._collection.get_channel_state().value
            return table
        else:
            return {'rc': 100, "errorInfo": "PerfProxy not configured."}

    def PublishMsg(self, params: dict):
        pub_props_builder = RuntimePropertyBuilder.create_default_properties()

        self._common_publish_props(pub_props_builder, params)

        try:
            self._collection.publish_message(pub_props_builder.props, Constants.ALL_CLIENT_INDEX)
        except Exception as e:
            logger.error(f"PublishMsg: {str(e)}")
            return {"rc": 1, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def StartPublishing(self, params: dict):
        pub_props_builder = RuntimePropertyBuilder.create_default_properties()
        self._common_publish_props(pub_props_builder, params)
        try:
            self._collection.start_publishing(pub_props_builder.props)
        except Exception as e:
            logger.error(f"StartPublishing: {str(e)}")
            return {"rc": 1, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def StopPublishing(self):
        try:
            self._collection.stop_publishing()
        except Exception as e:
            logger.error(f"StopPublishing: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def IsPublishing(self):
        table = dict(Constants.SUCCESS_RETURN)
        table["isPublishing"] = self._collection.is_publishing()
        return table

    def MapTopics(self, queue: str, topics_array: list, client_index: int, is_adding: bool):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured."}
        temp_topics_list = list(topics_array)
        if len(temp_topics_list) == 1 and temp_topics_list[0] == "":
            temp_topics_list = list()
        try:
            self._collection.map_topics(queue, temp_topics_list, is_adding, client_index)
        except Exception as e:
            logger.error(f"MapTopics: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def SubscriptionUpdate(self, params: dict):
        subscriptions = list()
        if RpcParameterKeys.SUBSCRIPTIONS in params:
            subscriptions = params[RpcParameterKeys.SUBSCRIPTIONS]
        if RpcParameterKeys.SUBSCRIPTIONS_FILE in params:
            subscriptions = []
            with open(params[RpcParameterKeys.SUBSCRIPTIONS_FILE]) as my_file:
                for line in my_file:
                    subscriptions.append(line)

        if params[RpcParameterKeys.IS_ADDING]:
            ep = EpConfigProperties.create_for_subscription_add(subscriptions,
                                                                params[RpcParameterKeys.CLIENT_NAME],
                                                                SubscriberDestinationsType.TOPIC)
        else:
            ep = EpConfigProperties.create_for_subscription_remove(subscriptions,
                                                                   params[RpcParameterKeys.CLIENT_NAME],
                                                                   SubscriberDestinationsType.TOPIC)
        table = dict(Constants.SUCCESS_RETURN)
        try:
            before = time.time()
            self._collection.subscription_update(ep, params[RpcParameterKeys.CLIENT_IDX])
            after = time.time()
            table["rate"] = f"{len(subscriptions) / (after - before):.2f}"
        except Exception as e:
            logger.error(f"SubscriptionUpdate: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}

        return table

    def RequestReply(self, topic: str, request: str, client_index: int):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured."}
        table = dict(Constants.SUCCESS_RETURN)
        try:
            reply = self._collection.request_reply(topic, request, client_index)
            table["reply"] = reply
        except Exception as e:
            logger.error(f"RequestReply: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return table

    @staticmethod
    def TopicUpdate(params: dict):
        return {"rc": 400, "errorInfo": "TopicUpdate not supported by python API"}

    def UnbindAllTempEndpoints(self, params: dict):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured."}
        client_index = Constants.ALL_CLIENT_INDEX
        try:
            if RpcParameterKeys.CLIENT_IDX in params:
                client_index = params[RpcParameterKeys.CLIENT_IDX]
            self._collection.unbind_all_temp_endpoints(client_index)
        except Exception as e:
            logger.error(f"UnbindAllTempEndpoints: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def TempEndpointUpdate(self, params: dict):
        ep = EpConfigProperties.create_default()
        index = -1

        table = dict(Constants.SUCCESS_RETURN)
        try:
            if RpcParameterKeys.SUBSCRIPTIONS in params:
                ep.subscriptions = params[RpcParameterKeys.SUBSCRIPTIONS]

            if RpcParameterKeys.SELECTORS in params:
                ep.selectors = params[RpcParameterKeys.SELECTORS]
            if len(ep.selectors) == 1 and ep.selectors[0] == "":
                ep.selectors = list()

            if RpcParameterKeys.IS_TOPIC_ENDPOINT in params:
                if params[RpcParameterKeys.IS_TOPIC_ENDPOINT]:
                    return {"rc": 400, "errorInfo": "Topic TempEndpointUpdate not supported by python API"}
                else:
                    ep.type = SubscriberDestinationsType.QUEUE

            if RpcParameterKeys.CLIENT_IDX in params:
                index = params[RpcParameterKeys.CLIENT_IDX]

            if RpcParameterKeys.NUM_END_POINTS in params:
                ep.num_temp_endpoints = params[RpcParameterKeys.NUM_END_POINTS]
            if RpcParameterKeys.NO_LOCAL in params:
                ep.no_local = params[RpcParameterKeys.NO_LOCAL]
            if RpcParameterKeys.SESSION_NAME in params:
                ep.transacted_session_name = params[RpcParameterKeys.SESSION_NAME]
            if RpcParameterKeys.TRANSACTION_SIZE in params:
                ep.transaction_size = params[RpcParameterKeys.TRANSACTION_SIZE]
            if RpcParameterKeys.ACTIVE_FLOW_IND in params:
                ep.want_active_flow_indication = params[RpcParameterKeys.ACTIVE_FLOW_IND]
            if RpcParameterKeys.WANT_CUT_THROUGH_PERSISTENCE in params:
                ep.want_cut_through_persistence = params[RpcParameterKeys.WANT_CUT_THROUGH_PERSISTENCE]
            if RpcParameterKeys.WANT_MSG_PRIORITY_ORDER_CHECKING in params:
                ep.want_message_priority_order_checking = params[RpcParameterKeys.WANT_MSG_PRIORITY_ORDER_CHECKING]
            if RpcParameterKeys.WANT_REPLAY in params:
                ep.want_replay = params[RpcParameterKeys.WANT_REPLAY]
            if RpcParameterKeys.WANT_REPLAY_FROM_DATE in params and \
                    params[RpcParameterKeys.WANT_REPLAY_FROM_DATE] != "":
                ep.want_replay = True
                ep.want_replay_from_date = params[RpcParameterKeys.WANT_REPLAY_FROM_DATE]
            if RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID in params and \
                    params[RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID] != "":
                ep.want_replay = True
                ep.want_replay_from_msg_id = params[RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID]
            if RpcParameterKeys.REQUIRED_SETTLEMENTS_STRING in params:
                ep.required_settlements = params[RpcParameterKeys.REQUIRED_SETTLEMENTS_STRING]
            if RpcParameterKeys.SETTLEMENT_PATTERN_STRING in params:
                ep.settlement_patterns = params[RpcParameterKeys.SETTLEMENT_PATTERN_STRING]

            # Ignoring Replay and Reconnect Information
            endpoints = self._collection.temp_endpoint_update(ep, index)

            if RpcParameterKeys.SKIP_ENDPOINT_NAMES in params and params[RpcParameterKeys.SKIP_ENDPOINT_NAMES]:
                table["endpoints"] = ""
            else:
                table["endpoints"] = endpoints
        except Exception as e:
            logger.error(f"TempEndpointUpdate: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return table

    def QueueUpdate(self, params: dict):
        ep = EpConfigProperties.create_default()
        index = -1

        try:
            if RpcParameterKeys.QUEUES_ARRAY in params:
                ep.ep_names = params[RpcParameterKeys.QUEUES_ARRAY]

            if RpcParameterKeys.SELECTORS in params:
                ep.selectors = params[RpcParameterKeys.SELECTORS]
            if len(ep.selectors) == 1 and ep.selectors[0] == "":
                ep.selectors = list()

            if RpcParameterKeys.IS_ADDING in params:
                ep.is_adding = params[RpcParameterKeys.IS_ADDING]
            if RpcParameterKeys.CLIENT_IDX in params:
                index = params[RpcParameterKeys.CLIENT_IDX]
            if RpcParameterKeys.NO_LOCAL in params:
                ep.no_local = params[RpcParameterKeys.NO_LOCAL]
            if RpcParameterKeys.SESSION_NAME in params:
                ep.transacted_session_name = params[RpcParameterKeys.SESSION_NAME]
            if RpcParameterKeys.TRANSACTION_SIZE in params:
                ep.transaction_size = params[RpcParameterKeys.TRANSACTION_SIZE]
            if RpcParameterKeys.ACTIVE_FLOW_IND in params:
                ep.want_active_flow_indication = params[RpcParameterKeys.ACTIVE_FLOW_IND]
            if RpcParameterKeys.WANT_CUT_THROUGH_PERSISTENCE in params:
                ep.want_cut_through_persistence = params[RpcParameterKeys.WANT_CUT_THROUGH_PERSISTENCE]
            if RpcParameterKeys.WANT_MSG_PRIORITY_ORDER_CHECKING in params:
                ep.want_message_priority_order_checking = params[RpcParameterKeys.WANT_MSG_PRIORITY_ORDER_CHECKING]
            if RpcParameterKeys.WANT_REPLAY in params:
                ep.want_replay = params[RpcParameterKeys.WANT_REPLAY]
            if RpcParameterKeys.WANT_REPLAY_FROM_DATE in params and \
                    params[RpcParameterKeys.WANT_REPLAY_FROM_DATE] != "":
                ep.want_replay = True
                ep.want_replay_from_date = params[RpcParameterKeys.WANT_REPLAY_FROM_DATE]
            if RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID in params and \
                    params[RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID] != "":
                ep.want_replay = True
                ep.want_replay_from_msg_id = params[RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID]
            if RpcParameterKeys.REQUIRED_SETTLEMENTS_STRING in params:
                ep.required_settlements = params[RpcParameterKeys.REQUIRED_SETTLEMENTS_STRING]
            if RpcParameterKeys.SETTLEMENT_PATTERN_STRING in params:
                ep.settlement_patterns = params[RpcParameterKeys.SETTLEMENT_PATTERN_STRING]

            # Ignoring Reconnect Information
            self._collection.queue_update(ep, index)
        except Exception as e:
            logger.error(f"QueueUpdate: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def EndpointProvisioning(self, params: dict):
        index = -1

        endpoints_array = list()
        is_topic = False
        is_adding = False
        access_type = EndpointAccessType.ACCESSTYPE_API_DEFAULT
        max_msg_size = -1
        quota = -1
        permission = -1
        respect_ttl = -1
        max_msg_redelivery = -1
        want_active_flow_indication = False

        try:
            if RpcParameterKeys.CLIENT_IDX in params:
                index = params[RpcParameterKeys.CLIENT_IDX]
            if RpcParameterKeys.EP_LIST in params:
                endpoints_array = params[RpcParameterKeys.EP_LIST]
            if RpcParameterKeys.IS_TOPIC_ENDPOINT in params:
                is_topic = params[RpcParameterKeys.IS_TOPIC_ENDPOINT]
            if RpcParameterKeys.IS_ADDING in params:
                is_adding = params[RpcParameterKeys.IS_ADDING]
            if RpcParameterKeys.ACCESS_TYPE in params:
                if params[RpcParameterKeys.ACCESS_TYPE] == 1:
                    access_type = EndpointAccessType.ACCESSTYPE_EXCLUSIVE
                elif params[RpcParameterKeys.ACCESS_TYPE] == 0:
                    access_type = EndpointAccessType.ACCESSTYPE_NONEXCLUSIVE
                else:
                    logger.error(f"Invalid Endpoint Access Type value provided: {params[RpcParameterKeys.ACCESS_TYPE]}")
            if RpcParameterKeys.MAX_MSG_SIZE in params:
                max_msg_size = params[RpcParameterKeys.MAX_MSG_SIZE]
            if RpcParameterKeys.QUOTA in params:
                quota = params[RpcParameterKeys.QUOTA]
            if RpcParameterKeys.PERMISSION in params:
                permission = params[RpcParameterKeys.PERMISSION]
            if RpcParameterKeys.RESPECT_TTL in params:
                if params[RpcParameterKeys.RESPECT_TTL]:
                    respect_ttl = 1
                else:
                    respect_ttl = 0
            if RpcParameterKeys.ACTIVE_FLOW_IND in params:
                want_active_flow_indication = params[RpcParameterKeys.ACTIVE_FLOW_IND]
            if RpcParameterKeys.MAX_MSG_REDELIVERY in params:
                max_msg_redelivery = params[RpcParameterKeys.MAX_MSG_REDELIVERY]

            ep = EpConfigProperties.create_for_ep_provision(endpoints_array,
                                                            is_topic,
                                                            is_adding,
                                                            access_type,
                                                            max_msg_size,
                                                            quota,
                                                            permission,
                                                            respect_ttl,
                                                            max_msg_redelivery,
                                                            want_active_flow_indication)
            if RpcParameterKeys.WANT_REPLAY in params:
                ep.want_replay = params[RpcParameterKeys.WANT_REPLAY]

            if RpcParameterKeys.WANT_REPLAY_FROM_DATE in params and \
                    params[RpcParameterKeys.WANT_REPLAY_FROM_DATE] != "":
                ep.want_replay = True
                ep.want_replay_from_date = params[RpcParameterKeys.WANT_REPLAY_FROM_DATE]

            if RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID in params and \
                    params[RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID] != "":
                ep.want_replay = True
                ep.want_replay_from_msg_id = params[RpcParameterKeys.WANT_REPLAY_FROM_MSG_ID]

            self._collection.endpoint_provision(ep, index)
        except Exception as e:
            logger.error(f"EndpointProvisioning: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}
        return Constants.SUCCESS_RETURN

    def GetFlowStatus(self, params: dict):
        ep_name = ""
        client_idx = -1

        if RpcParameterKeys.ENDPOINT_NAME in params:
            ep_name = params[RpcParameterKeys.ENDPOINT_NAME]
        if RpcParameterKeys.CLIENT_IDX in params:
            client_idx = params[RpcParameterKeys.CLIENT_IDX]

        try:
            flow_status = self._collection.get_flow_status(ep_name, client_idx)
        except Exception:
            flow_status = FlowStatus()
        table = dict(Constants.SUCCESS_RETURN)
        table["flowUp"] = flow_status.get_is_up()
        table["afiState"] = flow_status.get_afi_state_string()
        return table

    def GetLastAckedMsgId(self, client_index: int):
        table = dict(Constants.SUCCESS_RETURN)
        last_acked_msg_list = self._collection.get_last_acked_message_id(client_index)
        table["last-acked-msg-ids"] = list(last_acked_msg_list)
        return table

    def ResetStats(self):
        if not self._is_init():
            return {"rc": 100, "errorInfo": "PerfProxy not configured"}

        self._collection.clear_stats(Constants.ALL_CLIENT_INDEX)

        return Constants.SUCCESS_RETURN

    def GetStats(self, clientIndex: int, reset: bool = False):
        table = self._get_stats_table(clientIndex)

        if reset:
            self._collection.clear_stats(clientIndex)
        return table

    @staticmethod
    def GetCsmpStats(stats_vector: list, index: int):
        table = dict(Constants.SUCCESS_RETURN)
        for stat in stats_vector:
            table[stat] = True
        return table

    def GetOrderCheckDetails(self, params: dict):
        client_index = Constants.ALL_CLIENT_INDEX
        stream_id = None

        if RpcParameterKeys.STREAM_OFFSET in params:
            stream_id = params[RpcParameterKeys.STREAM_OFFSET]
        if RpcParameterKeys.CLIENT_IDX in params:
            client_index = params[RpcParameterKeys.CLIENT_IDX]

        if client_index == Constants.ALL_CLIENT_INDEX:
            client_ids = list()
            for i in range(len(self._collection.get_client_list())):
                client_ids.append(i)
        else:
            client_ids = [client_index]

        buf = ""
        # for client_id in client_ids:
        #     sub_stats = self._collection.get_rx_stats(client_id)
        #
        #     tot_ooo_msgs = sub_stats.get_stat("oooMsgs")
        #     tot_lost_msgs = 0
        #     tot_dup_msgs = 0
        #     tot_red_dup_msgs = 0
        #
        #     if tot_ooo_msgs > 0 or tot_lost_msgs > 0 or tot_dup_msgs > 0 or tot_red_dup_msgs > 0:
        #         buf += "Message Order Details Client: " + sub_stats.get_name() + "\n"
        #
        #         if stream_id is not None:
        #             buf += "  Stream: streamId \n"
        #             buf += sub_stats.getOrderData().get(stream_id).toString()
        #         else:
        #             # Always output the topic sequence first followed by default stream ID.
        #             if sub_stats.getOrderData().containsKey(PerfStats.C_TOPIC_SEQUENCE_STREAM_ID):
        #                 buf += "  Stream: TOPIC_SEQUENCE \n"
        #                 buf += sub_stats.getOrderData().get(PerfStats.C_TOPIC_SEQUENCE_STREAM_ID).toString()
        #
        #             if sub_stats.getOrderData().containsKey(PerfStats.C_DEFAULT_STREAM_ID):
        #                 buf += "  Stream: DEFAULT \n"
        #                 buf += sub_stats.getOrderData().get(PerfStats.C_DEFAULT_STREAM_ID).toString()
        #
        #             for oooEntry in sub_stats.getOrderData().entrySet():
        #                 # We've printed the special streams above.
        #                 if oooEntry.getKey().equals(PerfStats.C_DEFAULT_STREAM_ID) or oooEntry.getKey().equals(
        #                         PerfStats.C_TOPIC_SEQUENCE_STREAM_ID):
        #                     continue
        #
        #                 buf += "  Stream: " + oooEntry.getKey() + "\n"
        #                 buf += oooEntry.getValue().toString()
        table = dict(Constants.SUCCESS_RETURN)
        table["details"] = buf
        return table

    def GetCustomStats(self, clientIndex: int, statsFilter: str, reset: bool = False):
        try:
            table = dict(Constants.SUCCESS_RETURN)

            stats = self._collection.get_rx_stats(clientIndex)
            bin_crc_ok = self._collection.get_perf_stat("binCrcOk", clientIndex)
            bin_crc_fail = self._collection.get_perf_stat("binCrcFail", clientIndex)
            crc_ok = self._collection.get_perf_stat("crcOk", clientIndex)
            crc_fail = self._collection.get_perf_stat("crcFail", clientIndex)
            struct_ok = self._collection.get_perf_stat("structCheckOk", clientIndex)
            struct_fail = self._collection.get_perf_stat("structCheckFail", clientIndex)

            table["numMsgsCrcOk"] = crc_ok
            table["numMsgsCrcFail"] = crc_fail
            table["numMsgsRedelivered"] = self._collection.get_perf_stat("numMsgRedelivered", clientIndex)
            table["latencyMin"] = self._collection.get_perf_stat("latMin", clientIndex)
            table["latencyAvg"] = self._collection.get_perf_stat("latAvg", clientIndex)
            table["latencyMax"] = self._collection.get_perf_stat("latMax", clientIndex)
            table["latency50th"] = stats.get_percentile_latency_in_usec(50)
            table["latency95th"] = stats.get_percentile_latency_in_usec(95)
            table["latency99th"] = stats.get_percentile_latency_in_usec(99)
            table["latency99_9th"] = stats.get_percentile_latency_in_usec(99.9)
            table["latencyStdDev"] = stats.get_latency_std_dev_in_usec()
            table["latencyBucketSize"] = stats.get_latency_bucket_size()
            table["numMsgsDIFlag"] = self._collection.get_perf_stat("numMsgDISet", clientIndex)
            table["psmMaxUSecBehind"] = 0
            table["psmNumTimesBehind"] = 0
            table["numMsgsRecStructuredTotal"] = struct_ok
            table["numMsgsRecStructuredText"] = 0
            table["numMsgsRecStructuredMap"] = 0
            table["numMsgsRecStructuredStream"] = 0
            table["numMsgsRecUnstructuredTotal"] = 0
            table["numMsgsRecUnstructuredBytes"] = 0
            table["numMsgsRecUnstructuredXml"] = 0

            if "orderCheck" in statsFilter or "all" in statsFilter:
                order_check_ok = True
                num_ooo_msgs = 0
                num_dup_msgs = 0
                num_redelivered_dup_msgs = 0
                num_republished_dup_msgs = 0
                num_miss_msgs = 0

                for k, v in stats._ooo_data.items():
                    num_ooo_msgs += len(v.ooo_msg_ids)
                    num_miss_msgs += len(v.lost_msg_ids)
                    num_dup_msgs += len(v.dup_msg_ids)
                    num_redelivered_dup_msgs += len(v.redelivered_dup_msg_ids)
                    num_republished_dup_msgs += len(v.republished_dup_msg_ids)

                if num_ooo_msgs > 0 or num_dup_msgs > 0 or num_miss_msgs > 0:
                    order_check_ok = False

                table["orderCheckOk"] = order_check_ok
                table["numOrderCheckedMsgs"] = self._collection.get_perf_stat("numOrderCheckedMsgs", clientIndex)
                table["numOOOMsgs"] = num_ooo_msgs
                table["numMissingMsgs"] = num_miss_msgs
                table["numDuplicateMsgs"] = num_dup_msgs
                table["numRedeliveredDuplicateMsgs"] = num_redelivered_dup_msgs
                table["numRepublishedDuplicateMsgs"] = num_republished_dup_msgs
                # These 3 stats are intended to be used to verify windowed AD ACKs which rely on
                # order checking being enabled.
                table["numAckEvents"] = self._collection.get_perf_stat("numAckEvents", clientIndex)
                table["numNackEvents"] = 0
                table["numPubMsgsAcked"] = 0

                # Replication Group Message Id duplicate detection stats
                table["numReplicationGroupMessageIdDuplicates"] = self._collection.get_perf_stat(
                    "numReplicationGroupMessageIdDuplicates",
                    clientIndex
                )
                table["numReplicationGroupMessageIdSpoolMessageIdChanges"] = self._collection.get_perf_stat(
                    "numReplicationGroupMessageIdSpoolMessageIdChanges",
                    clientIndex
                )

            if "integrity" in statsFilter or "all" in statsFilter:
                table["numMsgsCrcXmlPayOk"] = crc_ok
                table["numMsgsCrcXmlPayFail"] = crc_fail
                table["numMsgsCrcBinAttOk"] = bin_crc_ok
                table["numMsgsCrcBinAttFail"] = bin_crc_fail
                table["numMsgsCrcUDOk"] = crc_ok
                table["numMsgsCrcUDFail"] = crc_fail
                table["numMsgsCrcSDMOk"] = struct_ok
                table["numMsgsCrcSDMFail"] = struct_fail

            if "latency" in statsFilter or "all" in statsFilter:
                buckets = stats.get_latency_buckets()
                ret_str = ""
                for bucket in buckets:
                    ret_str = f"{ret_str}{bucket} "
                table["latencyBuckets"] = ret_str

            if "messagePriority" in statsFilter or "all" in statsFilter:
                rx = stats.rx_priority
                tx = stats.tx_priority

                total_rx_msg = 0
                rx_priority_array = list()
                total_tx_msg = 0
                tx_priority_array = list()

                for k, v in rx.items():
                    rx_priority_array.append(k)
                    rx_priority_array.append(v)
                    total_rx_msg += v

                for k, v in tx.items():
                    tx_priority_array.append(k)
                    tx_priority_array.append(v)
                    total_tx_msg += v

                table["numMsgsWithPriorityRecv"] = str(total_rx_msg)
                table["numMsgsWithPrioritySent"] = str(total_tx_msg)
                table["msgPrioritiesRecv"] = str(rx_priority_array)
                table["msgPrioritiesSent"] = str(tx_priority_array)
                table["numMsgPriorityOrderCheckingErrors"] = str(0)

            if "cache" in statsFilter or "all" in statsFilter:
                table["numCacheReqIncompNoData"] = self._collection.get_perf_stat("numCacheReqIncompNoData",
                                                                                  clientIndex)
                table["numCacheReqOk"] = self._collection.get_perf_stat("numCacheReqOk", clientIndex)
                table["numCacheReqIncompSuspect"] = self._collection.get_perf_stat("numCacheReqIncompSuspect",
                                                                                   clientIndex)
                table["numLiveMsgsRecv"] = self._collection.get_perf_stat("numLiveMsgsRecv", clientIndex)
                table["numCacheReqError"] = self._collection.get_perf_stat("numCacheReqError", clientIndex)
                table["numCacheMsgsRecv"] = self._collection.get_perf_stat("numCacheMsgsRecv", clientIndex)
                table["numCacheReqSent"] = self._collection.get_sdk_stat(Metric.CACHE_REQUESTS_SENT, clientIndex)
                table["numSuspectRecv"] = self._collection.get_perf_stat("numSuspectRecv", clientIndex)
                table["numCacheRespRecv"] = self._collection.get_perf_stat("numCacheRespRecv", clientIndex)
                table["numCacheReqIncompTimeout"] = self._collection.get_perf_stat("numCacheReqIncompTimeout",
                                                                                   clientIndex)
                table["numCacheReqIncompCancelled"] = 0
                table["numDiscardedResp"] = 0
                table["numLiveDataFulfill"] = self._collection.get_sdk_stat(Metric.CACHE_REQUESTS_SUCCEEDED,
                                                                            clientIndex) - \
                                              self._collection.get_perf_stat("numCacheReqOk", clientIndex)

            if "nextGenStats" in statsFilter or "all" in statsFilter:
                table["numPublishMsgsTerminationDiscarded"] = self._collection \
                    .get_sdk_stat(Metric.PUBLISH_MESSAGES_TERMINATION_DISCARDED, clientIndex)
                table["numReceivedMsgsTerminationDiscarded"] = self._collection \
                    .get_sdk_stat(Metric.RECEIVED_MESSAGES_TERMINATION_DISCARDED, clientIndex)
                table["numReceivedMessagesBackpressureDiscarded"] = self._collection \
                    .get_sdk_stat(Metric.RECEIVED_MESSAGES_BACKPRESSURE_DISCARDED, clientIndex)
                table["numInternalDiscardNotifications"] = self._collection \
                    .get_sdk_stat(Metric.INTERNAL_DISCARD_NOTIFICATIONS, clientIndex)
                table["numDirectMessagesReceived"] = self._collection \
                    .get_sdk_stat(Metric.DIRECT_MESSAGES_RECEIVED, clientIndex)
                table["numTerminationEvent"] = self._collection.get_perf_stat("numTerminationEvent", clientIndex)
                table["numTerminationRegistered"] = self._collection.get_num_termination_registered(clientIndex)

            if "messageContent" in statsFilter or "all" in statsFilter:
                table["numMsgsObjTypeOk"] = self._collection.get_perf_stat("numMsgsObjTypeOk", clientIndex)
                table["numMsgsObjTypeFail"] = self._collection.get_perf_stat("numMsgsObjTypeFail", clientIndex)

            if "nackStats" in statsFilter or "all" in statsFilter:
                table["reliablePersistentMsgsRecvedAcked"] = self._collection.get_sdk_stat(Metric.PERSISTENT_ACKNOWLEDGE_SENT, clientIndex)
                table["reliableMsgsRecvedAcked"] = self._collection.get_sdk_stat(Metric.PERSISTENT_ACKNOWLEDGE_SENT, clientIndex)
                table["reliablePersistentMsgsRecvedAccepted"] = self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_ACCEPTED, clientIndex)
                table["reliableMsgsRecvedAccepted"] = self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_ACCEPTED, clientIndex)
                table["reliablePersistentMsgsRecvedRejected"] = self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_REJECTED, clientIndex)
                table["reliableMsgsRecvedRejected"] = self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_REJECTED, clientIndex)
                table["reliablePersistentMsgsRecvedFailed"] = self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_FAILED, clientIndex)
                table["reliableMsgsRecvedFailed"] = self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_FAILED, clientIndex)


            if reset:
                self._collection.clear_stats(clientIndex)

            logger.warn(f"Return from GetCustomStats follows:\n{table}")

            return table
        except Exception as e:
            logger.error(f"GetCustomStats: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}

    def StartCacheRequesting(self, params: dict):
        topics = []
        num_msgs = 0
        cache_rate = 0
        live_data_action = ""
        want_subscribe = False
        wait_for_confirm = False
        if RpcParameterKeys.TOPICS_ARRAY in params:
            topics = params[RpcParameterKeys.TOPICS_ARRAY]
        if RpcParameterKeys.TOPICS_FILE in params:
            with open(params[RpcParameterKeys.TOPICS_FILE], 'r') as topic_file:
                for line in topic_file:
                    topics.append(line.strip())
        if RpcParameterKeys.MSG_NUM in params:
            num_msgs = params[RpcParameterKeys.MSG_NUM]
        if RpcParameterKeys.MSG_RATE in params:
            cache_rate = params[RpcParameterKeys.MSG_RATE]
        if RpcParameterKeys.CACHE_REQ_LIVE_DATA_ACTION in params:
            live_data_action = params[RpcParameterKeys.CACHE_REQ_LIVE_DATA_ACTION]
        if RpcParameterKeys.CACHE_REQ_WANT_SUBSCRIBE in params:
            want_subscribe = params[RpcParameterKeys.CACHE_REQ_WANT_SUBSCRIBE]
        if RpcParameterKeys.CACHE_REQ_WAIT_FOR_CONFIRM in params:
            wait_for_confirm = params[RpcParameterKeys.CACHE_REQ_WAIT_FOR_CONFIRM]

        self._collection.start_cache_requesting(topics, want_subscribe, live_data_action, wait_for_confirm, num_msgs,
                                                cache_rate)

        table = dict(Constants.SUCCESS_RETURN)
        return table

    def StopCacheRequesting(self):
        self._collection.stop_cache_requesting()

        table = dict(Constants.SUCCESS_RETURN)
        return table

    def IsCacheRequesting(self):
        ret = self._collection.is_cache_requesting()

        table = dict(Constants.SUCCESS_RETURN)
        table["isCacheRequesting"] = ret
        return table

    def CacheRequest(self, params: dict):
        topics = []
        client_index = -1
        live_data_action = ""
        want_subscribe = False
        wait_for_confirm = False
        if RpcParameterKeys.TOPICS_ARRAY in params:
            topics = params[RpcParameterKeys.TOPICS_ARRAY]
        if RpcParameterKeys.TOPICS_FILE in params:
            with open(params[RpcParameterKeys.TOPICS_FILE], 'r') as topic_file:
                for line in topic_file:
                    topics.append(line.strip())
        if RpcParameterKeys.CACHE_REQ_LIVE_DATA_ACTION in params:
            live_data_action = params[RpcParameterKeys.CACHE_REQ_LIVE_DATA_ACTION]
        if RpcParameterKeys.CACHE_REQ_WANT_SUBSCRIBE in params:
            want_subscribe = params[RpcParameterKeys.CACHE_REQ_WANT_SUBSCRIBE]
        if RpcParameterKeys.CACHE_REQ_WAIT_FOR_CONFIRM in params:
            wait_for_confirm = params[RpcParameterKeys.CACHE_REQ_WAIT_FOR_CONFIRM]

        try:
            self._collection.cache_request(topics, want_subscribe, live_data_action, wait_for_confirm, client_index)
        except Exception as e:
            logger.error(f"CacheRequest: {str(e)}")
            return {"rc": 400, "errorInfo": str(e)}

        table = dict(Constants.SUCCESS_RETURN)
        return table

    @staticmethod
    def GetCpuUsage():
        table = dict(Constants.SUCCESS_RETURN)
        table["cpuUsage"] = "50.00"
        return table

    def _get_stats_table(self, index):
        table = dict(Constants.SUCCESS_RETURN)

        rx_msg_rate = 0
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._collection.get_client_list():
                rx_msg_rate += client.get_rx_stats().get_thru_put(
                    self._collection.get_perf_stat("rxMsgs", client._client_id - 1))
        else:
            c = self._collection.get_client(index)
            rx_msg_rate += c.get_rx_stats().get_thru_put(
                self._collection.get_perf_stat("rxMsgs", index))

        table["txMsgs"] = f"{self._collection.get_sdk_stat(Metric.TOTAL_MESSAGES_SENT, index)}"
        table["txBytes"] = f"{self._collection.get_sdk_stat(Metric.TOTAL_BYTES_SENT, index)}"
        table["txMsgRate"] = f"{self._collection.get_tx_throughput():.2f}"
        table["rxMsgs"] = f'{self._collection.get_perf_stat("rxMsgs", index)}'
        table["rxBytes"] = f'{self._collection.get_perf_stat("rxBytes", index)}'
        table["rxMsgRate"] = f"{rx_msg_rate:.2f}"

        logger.debug(f"Stats: {table}")
        return table

    def _common_publish_props(self, prop_builder, params):
        if RpcParameterKeys.TOPICS_ARRAY in params:
            prop_builder.set_publish_topic_list(params[RpcParameterKeys.TOPICS_ARRAY])
        if RpcParameterKeys.TOPICS_FILE in params:
            topics = []
            with open(params[RpcParameterKeys.TOPICS_FILE], 'r') as topic_file:
                for line in topic_file:
                    topics.append(line.strip())
            if len(topics) > 0:
                prop_builder.set_publish_topic_list(topics)
        if RpcParameterKeys.QUEUES_ARRAY in params:
            prop_builder.set_publish_queue_list(params[RpcParameterKeys.QUEUES_ARRAY])
        if RpcParameterKeys.MSG_TYPE in params:
            prop_builder.set_pub_message_type(params[RpcParameterKeys.MSG_TYPE])
        if RpcParameterKeys.MESSAGE_OBJECT_TYPE in params:
            prop_builder.set_msg_obj_type_list(params[RpcParameterKeys.MESSAGE_OBJECT_TYPE])
        if RpcParameterKeys.MSG_RATE in params:
            rate = params[RpcParameterKeys.MSG_RATE]
            num_clients = getattr(self._pb.props, RuntimeProperties.NUM_CLIENTS, 1)
            num_pubs_per_session = getattr(self._pb.props, RuntimeProperties.NUM_PUBS_PER_SESSION, 1)
            publish_rate = rate / num_clients / num_pubs_per_session
            prop_builder.set_publish_rate_per_pub(publish_rate)
        if RpcParameterKeys.MSG_RATE_IS_MAX in params:
            prop_builder.set_msg_rate_is_max(True)
        if RpcParameterKeys.MSG_NUM in params:
            prop_builder.set_num_msgs_to_publish(params[RpcParameterKeys.MSG_NUM])
        if RpcParameterKeys.MESSAGE_ID in params:
            prop_builder.set_message_id(params[RpcParameterKeys.MESSAGE_ID])
        if RpcParameterKeys.USER_COS in params and params[RpcParameterKeys.USER_COS] != -1:
            prop_builder.set_publish_cos(params[RpcParameterKeys.USER_COS])
        if RpcParameterKeys.WANT_LATENCY in params:
            prop_builder.set_want_latency(params[RpcParameterKeys.WANT_LATENCY])
        if RpcParameterKeys.WANT_ORDER_CHECK in params:
            prop_builder.set_want_order_check(params[RpcParameterKeys.WANT_ORDER_CHECK])
        if RpcParameterKeys.PUB_STREAM_OFFSET in params:
            prop_builder.set_pub_stream_id_offset(params[RpcParameterKeys.PUB_STREAM_OFFSET])
            prop_builder.set_want_order_memory(True)
        if RpcParameterKeys.PUB_ORDER_OFFSET in params:
            prop_builder.set_pub_order_offset(params[RpcParameterKeys.PUB_ORDER_OFFSET])
        if RpcParameterKeys.WANT_PAYLOAD_CHECK in params:
            prop_builder.set_want_payload_check(params[RpcParameterKeys.WANT_PAYLOAD_CHECK])

        if RpcParameterKeys.ATTACHMENT_SIZE in params:
            prop_builder.set_attachment_size_list([params[RpcParameterKeys.ATTACHMENT_SIZE]])
        if RpcParameterKeys.ATTACHMENT_SIZE_LIST in params:
            prop_builder.set_attachment_size_list(params[RpcParameterKeys.ATTACHMENT_SIZE_LIST])
        if RpcParameterKeys.ATTACHMENT_FILE_ARRAY in params:
            prop_builder.set_attachment_file_list(params[RpcParameterKeys.ATTACHMENT_FILE_ARRAY])
        if RpcParameterKeys.ATTACHMENT_SIZE_SWEEP in params:
            prop_builder.set_attachment_sweep(params[RpcParameterKeys.ATTACHMENT_SIZE_SWEEP])

        if RpcParameterKeys.DOC_FILE_ARRAY in params:
            prop_builder.set_xml_file_list(params[RpcParameterKeys.DOC_FILE_ARRAY])

        if RpcParameterKeys.SDM_FILE_ARRAY in params:
            prop_builder.set_sdm_prop_list_from_files(params[RpcParameterKeys.SDM_FILE_ARRAY])
            prop_builder.set_want_struct_msg_check(True)
        if RpcParameterKeys.SDM_STRINGS_ARRAY in params:
            prop_builder.set_sdm_prop_list_from_strings(params[RpcParameterKeys.SDM_STRINGS_ARRAY])
            prop_builder.set_want_struct_msg_check(True)

        if RpcParameterKeys.DMQ_ELIGIBLE in params:
            prop_builder.set_msg_dmq_eligible(params[RpcParameterKeys.DMQ_ELIGIBLE])
        if RpcParameterKeys.TTL in params:
            prop_builder.set_msg_ttl(params[RpcParameterKeys.TTL])
        if RpcParameterKeys.ELIDING in params:
            prop_builder.set_msg_eliding_eligible(params[RpcParameterKeys.ELIDING])
        if RpcParameterKeys.MESSAGE_PRIORITY in params:
            prop_builder.set_publish_message_priority([int(params[RpcParameterKeys.MESSAGE_PRIORITY]), 1])
        if RpcParameterKeys.PUBLISH_MESSAGE_PRIORITY_LIST in params:
            str_list = str(params[RpcParameterKeys.PUBLISH_MESSAGE_PRIORITY_LIST])
            int_list = [int(x) for x in str_list.split(",")]
            prop_builder.set_publish_message_priority(int_list)

        if RpcParameterKeys.TOPIC_PREFIX in params:
            prefix = params[RpcParameterKeys.TOPIC_PREFIX]
            tc = 1
            if RpcParameterKeys.TOPIC_COUNT in params and params[RpcParameterKeys.TOPIC_COUNT] > 0:
                tc = params[RpcParameterKeys.TOPIC_COUNT]
            topics = list()
            for i in range(tc, tc + 1):
                topics.append(f"{prefix}{i:06d}")
            prop_builder.set_publish_topic_list(topics)

        if RpcParameterKeys.CLIENT_MODE in params:
            prop_builder.set_client_mode(params[RpcParameterKeys.CLIENT_MODE])
        if RpcParameterKeys.XA_SESSION_NAMES_ARRAY in params:
            prop_builder.set_transacted_session_name_list(params[RpcParameterKeys.XA_SESSION_NAMES_ARRAY])
        if RpcParameterKeys.TRANSACTION_SIZE in params:
            prop_builder.set_ad_transaction_size(params[RpcParameterKeys.TRANSACTION_SIZE])
        if RpcParameterKeys.TRANSACTION_ROLLBACK_INTERVAL in params:
            prop_builder.set_ad_transaction_rollback_interval(params[RpcParameterKeys.TRANSACTION_ROLLBACK_INTERVAL])
        if RpcParameterKeys.SKIP_FINAL_COMMIT in params:
            prop_builder.set_skip_final_commit(params[RpcParameterKeys.SKIP_FINAL_COMMIT])
        if RpcParameterKeys.BURST_DURATION in params:
            prop_builder.set_burst_duration(params[RpcParameterKeys.BURST_DURATION])
        if RpcParameterKeys.INTER_BURST_DURATION in params:
            prop_builder.set_inter_burst_duration(params[RpcParameterKeys.INTER_BURST_DURATION])
        if RpcParameterKeys.MSG_RATE_IS_MAX in params:
            prop_builder.set_msg_rate_is_max(params[RpcParameterKeys.MSG_RATE_IS_MAX])
        if RpcParameterKeys.WANT_SMOOTH_PUB in params:
            prop_builder.set_want_smooth_pub_calc_lat(params[RpcParameterKeys.WANT_SMOOTH_PUB])

        if RpcParameterKeys.HTTP_CONTENT_ENCODING in params:
            prop_builder.set_http_content_encoding(params[RpcParameterKeys.HTTP_CONTENT_ENCODING])
        if RpcParameterKeys.HTTP_CONTENT_TYPE in params:
            prop_builder.set_http_content_type(params[RpcParameterKeys.HTTP_CONTENT_TYPE])

        # In python we cannot specify anything about the reply-to topic so if it's wanted then we're just gonna use it
        if RpcParameterKeys.WANT_REPLY_TO_TOPIC in params:
            prop_builder.set_want_reply_topic(params[RpcParameterKeys.WANT_REPLY_TO_TOPIC])
        if RpcParameterKeys.WANT_REPLY_TO_STRING in params and params[RpcParameterKeys.WANT_REPLY_TO_STRING] != "":
            prop_builder.set_want_reply_topic(True)
        if RpcParameterKeys.WANT_REPLY_TO_POSTFIX in params and params[RpcParameterKeys.WANT_REPLY_TO_POSTFIX] != "":
            prop_builder.set_want_reply_topic(True)
        if RpcParameterKeys.PARTITION_KEY_LIST in params and params[RpcParameterKeys.PARTITION_KEY_LIST] != "":
            prop_builder.set_partition_key_list(params[RpcParameterKeys.PARTITION_KEY_LIST])

        if RpcParameterKeys.WANT_TRACE_CREATION_CONTEXT in params \
                and params[RpcParameterKeys.WANT_TRACE_CREATION_CONTEXT]:
            prop_builder.set_want_trace_creation_context(True)
        if RpcParameterKeys.WANT_TRACE_TRANSPORT_CONTEXT in params \
                and params[RpcParameterKeys.WANT_TRACE_TRANSPORT_CONTEXT]:
            prop_builder.set_want_trace_transport_context(True)
        if RpcParameterKeys.BAGGAGE in params and params[RpcParameterKeys.BAGGAGE] != "":
            prop_builder.set_trace_baggage(params[RpcParameterKeys.BAGGAGE])

        if RpcParameterKeys.CUSTOM_PROPERTY_LIST in params:
            prop_builder.set_custom_properties_list(params[RpcParameterKeys.CUSTOM_PROPERTY_LIST])

    @staticmethod
    def create(unique_id: str, mode: str):
        proxy = PerfClientProxy(unique_id, mode)
        return proxy
