from solace.messaging.config.transport_security_strategy import TLS

from pypubsubtools.core.runtime_properties import RuntimeProperties
from sdkperf.config.authentication_scheme import AuthenticationScheme
from sdkperf.config.publisher_back_pressure_type import PublisherBackPressureType
from sdkperf.config.subscriber_back_pressure_type import SubscriberBackPressureType
from sdkperf.config.client_mode_type import ClientModeType
from sdkperf.config.endpoint_access_type import EndpointAccessType
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.config.message_object_type import MessageObjectType
from sdkperf.core.message_priority_generator import MessagePriorityGenerator
from sdkperf.core.sdm_properties import SdmProperties


class RuntimePropertyBuilder:
    SSL_PROTOCOLS_MAP = {'sslv3': TLS.SecureProtocols.SSLv3,
                         'tlsv1': TLS.SecureProtocols.TLSv1,
                         'tlsv1.1': TLS.SecureProtocols.TLSv1_1,
                         'tlsv1.2': TLS.SecureProtocols.TLSv1_2}

    def __init__(self):
        self.props = RuntimeProperties()

    def set_request_reply_wait_time(self, request_reply_wait_time):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.REQUEST_REPLY_WAIT_TIME, request_reply_wait_time)

    def set_num_clients(self, num_clients):
        """ Auto-generated property setter """
        if num_clients < 0:
            raise Exception("Number of Client (-cc) argument must be greater than 0")
        setattr(self.props, RuntimeProperties.NUM_CLIENTS, num_clients)

    def set_num_context_threads(self, num_context_threads):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_CONTEXT_THREADS, num_context_threads)

    def set_num_pub_threads(self, num_pub_threads):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_PUB_THREADS, num_pub_threads)

    def set_num_pubs_per_session(self, num_pubs_per_session):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_PUBS_PER_SESSION, num_pubs_per_session)

    def set_publish_end_delay_in_sec(self, publish_end_delay_in_sec):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_END_DELAY_IN_SEC, publish_end_delay_in_sec)

    def set_publish_rate_per_pub(self, publish_rate_per_pub):
        """ Auto-generated property setter """
        if publish_rate_per_pub < 0:
            raise Exception("Message Rate (-mr) argument must be greater or equal 0")
        setattr(self.props, RuntimeProperties.PUBLISH_RATE_PER_PUB, publish_rate_per_pub)

    def set_num_msgs_to_publish(self, num_msgs_to_publish):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_MSGS_TO_PUBLISH, num_msgs_to_publish)

    def set_xml_payload_size_list(self, xml_payload_size_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XML_PAYLOAD_SIZE_LIST, xml_payload_size_list)

    def set_attachment_size_list(self, attachment_size_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.ATTACHMENT_SIZE_LIST, attachment_size_list)

    def set_attachment_file_list(self, attachment_file_list):
        doc_list = list()
        for file in attachment_file_list:
            try:
                with open(file, "rb") as f:
                    doc_list.append(bytearray(f.read()))
            except Exception as e:
                print(f"ERROR OPENING ATTACHMENT FILE {file} {e}")
        setattr(self.props, RuntimeProperties.ATTACHMENT_FILE_LIST, doc_list)

    def set_xml_file_list(self, xml_file_list):
        doc_list = list()
        for file in xml_file_list:
            try:
                with open(file, "rb") as f:
                    doc_list.append(f.read())
            except Exception as e:
                print(f"ERROR OPENING XML FILE {file} {e}")
        setattr(self.props, RuntimeProperties.XML_FILE_LIST, doc_list)

    def set_sdm_prop_list_from_files(self, sdm_file_list):
        sdm_props = list()
        for file in sdm_file_list:
            prop = SdmProperties()
            try:
                prop.parse_input_file(file)
            except Exception as e:
                print(f"ERROR SETTING SDM PROPS {file} {e}")
            sdm_props.append(prop)
        setattr(self.props, RuntimeProperties.SDM_PROP_LIST, sdm_props)

    def set_sdm_prop_list_from_strings(self, sdm_string_list):
        prop = SdmProperties()
        try:
            prop.parse_input_lines(sdm_string_list)
        except Exception as e:
            print(f"ERROR SETTING SDM PROPS {e}")
        setattr(self.props, RuntimeProperties.SDM_PROP_LIST, [prop])

    def set_attachment_sweep(self, value):
        if not isinstance(value, str):
            raise Exception(f"Attachment sweep must be a string, not {type(value)}")
        values = list(map(int, value.split(',')))
        if len(values) != 3:
            raise Exception(f"ERROR: Must specify start,step,end in bytes for msa-sweep: {value}")
        ints = list()
        for i in range(values[0], values[2] + 1, values[1]):
            ints.append(i)
        setattr(self.props, RuntimeProperties.ATTACHMENT_SIZE_LIST, ints)

    def set_client_ip_addr(self, client_ip_addr):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_IP_ADDR, client_ip_addr)

    def set_client_username(self, client_username):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_USERNAME, client_username)

    def set_client_vpn(self, client_vpn):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_VPN, client_vpn)

    def set_client_name_prefix(self, client_name_prefix):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_NAME_PREFIX, client_name_prefix)

    def set_change_client_names_flag(self, change_client_names_flag):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CHANGE_CLIENT_NAMES_FLAG, change_client_names_flag)

    def set_client_password(self, client_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_PASSWORD, client_password)

    def set_client_description(self, client_description):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_DESCRIPTION, client_description)

    def set_client_dto_local_priority(self, client_dto_local_priority):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_DTO_LOCAL_PRIORITY, client_dto_local_priority)

    def set_client_dto_network_priority(self, client_dto_network_priority):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_DTO_NETWORK_PRIORITY, client_dto_network_priority)

    def set_client_compression_level(self, client_compression_level):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_COMPRESSION_LEVEL, client_compression_level)

    def set_payload_compression_level(self, payload_compression_level):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PAYLOAD_COMPRESSION_LEVEL, payload_compression_level)

    def set_want_tcp_nodelay(self, want_tcp_nodelay):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_TCP_NODELAY, want_tcp_nodelay)

    def set_publish_file_list(self, publish_file_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_FILE_LIST, publish_file_list)

    def set_publish_attach_list(self, publish_attach_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_ATTACH_LIST, publish_attach_list)

    def set_smf_binary_file_list(self, smf_binary_file_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SMF_BINARY_FILE_LIST, smf_binary_file_list)

    def set_subscription_refresh(self, subscription_refresh):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUBSCRIPTION_REFRESH, subscription_refresh)

    def set_publish_topic_list(self, publish_topic_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_TOPIC_LIST, publish_topic_list)

    def set_publish_queue_list(self, publish_queue_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_QUEUE_LIST,
                [f"#P2P/QUE/{q}" if "#P2P" not in q else q for q in publish_queue_list])

    def set_num_message_producer(self, num_message_producer):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_MESSAGE_PRODUCER, num_message_producer)

    def set_publish_message_priority(self, publish_message_priority):
        pri_obj = MessagePriorityGenerator(publish_message_priority)
        setattr(self.props, RuntimeProperties.PUBLISH_MESSAGE_PRIORITY, pri_obj)

    def set_selector_list(self, selector_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SELECTOR_LIST, selector_list)

    def set_want_remove_subscriber(self, want_remove_subscriber):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_REMOVE_SUBSCRIBER, want_remove_subscriber)

    def set_want_subscription_memory(self, want_subscription_memory):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_SUBSCRIPTION_MEMORY, want_subscription_memory)

    def set_want_per_sub_thru_stats(self, want_per_sub_thru_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_PER_SUB_THRU_STATS, want_per_sub_thru_stats)

    def set_want_subscription_rate_stats(self, want_subscription_rate_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_SUBSCRIPTION_RATE_STATS, want_subscription_rate_stats)

    def set_ad_pub_window_size(self, ad_pub_window_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_PUB_WINDOW_SIZE, ad_pub_window_size)

    def set_ad_pub_ack_timeout(self, ad_pub_ack_timeout):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_PUB_ACK_TIMEOUT, ad_pub_ack_timeout)

    def set_ad_sub_window_size(self, ad_sub_window_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_SUB_WINDOW_SIZE, ad_sub_window_size)

    def set_ad_sub_ack_timeout(self, ad_sub_ack_timeout):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_SUB_ACK_TIMEOUT, ad_sub_ack_timeout)

    def set_ad_sub_window_threshold(self, ad_sub_window_threshold):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_SUB_WINDOW_THRESHOLD, ad_sub_window_threshold)

    def set_ad_transaction_size(self, ad_transaction_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_TRANSACTION_SIZE, ad_transaction_size)

    def set_ad_want_xa_transaction(self, ad_want_xa_transaction):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_WANT_XA_TRANSACTION, ad_want_xa_transaction)

    def set_xa_want_one_phase_commit_transaction(self, xa_want_one_phase_commit_transaction):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_WANT_ONE_PHASE_COMMIT_TRANSACTION,
                xa_want_one_phase_commit_transaction)

    def set_xa_num_msgs_per_transaction_segment(self, xa_num_msgs_per_transaction_segment):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_NUM_MSGS_PER_TRANSACTION_SEGMENT, xa_num_msgs_per_transaction_segment)

    def set_xa_num_suspended_transactions(self, xa_num_suspended_transactions):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_NUM_SUSPENDED_TRANSACTIONS, xa_num_suspended_transactions)

    def set_ad_want_producer_consumers_transaction(self, ad_want_producer_consumers_transaction):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_WANT_PRODUCER_CONSUMERS_TRANSACTION,
                ad_want_producer_consumers_transaction)

    def set_xa_transaction_idle_timeout(self, xa_transaction_idle_timeout):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_TRANSACTION_IDLE_TIMEOUT, xa_transaction_idle_timeout)

    def set_xa_transaction_phase_delay_before_suspending(self, xa_transaction_phase_delay_before_suspending):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_TRANSACTION_PHASE_DELAY_BEFORE_SUSPENDING,
                xa_transaction_phase_delay_before_suspending)

    def set_xa_transaction_phase_delay_before_ending(self, xa_transaction_phase_delay_before_ending):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_TRANSACTION_PHASE_DELAY_BEFORE_ENDING,
                xa_transaction_phase_delay_before_ending)

    def set_xa_transaction_phase_delay_before_preparing(self, xa_transaction_phase_delay_before_preparing):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_TRANSACTION_PHASE_DELAY_BEFORE_PREPARING,
                xa_transaction_phase_delay_before_preparing)

    def set_xa_transaction_phase_delay_before_committing(self, xa_transaction_phase_delay_before_committing):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.XA_TRANSACTION_PHASE_DELAY_BEFORE_COMMITTING,
                xa_transaction_phase_delay_before_committing)

    def set_ad_transaction_idle_time(self, ad_transaction_idle_time):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_TRANSACTION_IDLE_TIME, ad_transaction_idle_time)

    def set_ad_transaction_rollback_interval(self, ad_transaction_rollback_interval):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_TRANSACTION_ROLLBACK_INTERVAL, ad_transaction_rollback_interval)

    def set_ad_want_active_flow_indication(self, ad_want_active_flow_indication):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_WANT_ACTIVE_FLOW_INDICATION, ad_want_active_flow_indication)

    def set_ad_ack_immediately_interval(self, ad_ack_immediately_interval):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_ACK_IMMEDIATELY_INTERVAL, ad_ack_immediately_interval)

    def set_ad_ack_event_mode(self, ad_ack_event_mode):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_ACK_EVENT_MODE, ad_ack_event_mode)

    def set_ad_want_pub_ack_order_check(self, ad_want_pub_ack_order_check):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_WANT_PUB_ACK_ORDER_CHECK, ad_want_pub_ack_order_check)

    def set_cid_stats(self, cid_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CID_STATS, cid_stats)

    def set_want_di_stats(self, want_di_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_DI_STATS, want_di_stats)

    def set_ad_pub_max_resend(self, ad_pub_max_resend):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.AD_PUB_MAX_RESEND, ad_pub_max_resend)

    def set_want_verbose(self, want_verbose):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_VERBOSE, want_verbose)

    def set_want_quiet(self, want_quiet):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_QUIET, want_quiet)

    def set_want_ud(self, want_ud):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_UD, want_ud)

    def set_want_sync_receive(self, want_sync_receive):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_SYNC_RECEIVE, want_sync_receive)

    def set_pub_message_type(self, pub_message_type):
        if type(pub_message_type) == GenericMessageDeliveryMode:
            setattr(self.props, RuntimeProperties.PUB_MESSAGE_TYPE, pub_message_type)
        elif pub_message_type.upper() == GenericMessageDeliveryMode.DIRECT.name:
            setattr(self.props, RuntimeProperties.PUB_MESSAGE_TYPE, GenericMessageDeliveryMode.DIRECT)
        elif pub_message_type.upper() == GenericMessageDeliveryMode.PERSISTENT.name:
            setattr(self.props, RuntimeProperties.PUB_MESSAGE_TYPE, GenericMessageDeliveryMode.PERSISTENT)
        elif pub_message_type.upper() == GenericMessageDeliveryMode.NON_PERSISTENT.name:
            setattr(self.props, RuntimeProperties.PUB_MESSAGE_TYPE, GenericMessageDeliveryMode.NON_PERSISTENT)
        elif pub_message_type.upper() == "DEFAULT":
            setattr(self.props, RuntimeProperties.PUB_MESSAGE_TYPE, GenericMessageDeliveryMode.DIRECT)
        else:
            raise Exception(f"Invalid value \"{pub_message_type}\" for Message Delivery Mode.")
        
    def set_msg_obj_type_list(self, msg_obj_type_pattern):
        pattern = msg_obj_type_pattern.lower()
        msg_obj_type_list = []
        for char in pattern:
            if char == 'b':
                msg_obj_type_list.append(MessageObjectType.BYTES)
            elif char == 't':
                msg_obj_type_list.append(MessageObjectType.TEXT)
            elif char == 's':
                msg_obj_type_list.append(MessageObjectType.STREAM)
            elif char == 'm':
                msg_obj_type_list.append(MessageObjectType.MAP)

        if msg_obj_type_list:
            setattr(self.props, RuntimeProperties.MSG_OBJ_TYPE_LIST, msg_obj_type_list)

    def set_want_msg_obj_type_check(self, want_msg_obj_type_check):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_MSG_OBJ_TYPE_CHECK, want_msg_obj_type_check)

    def set_want_smooth_pub_calc_lat(self, want_smooth_pub_calc_lat):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_SMOOTH_PUB_CALC_LAT, want_smooth_pub_calc_lat)

    def set_want_struct_msg_check(self, want_struct_msg_check):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_STRUCT_MSG_CHECK, want_struct_msg_check)

    def set_struct_data_msgs_list(self, struct_data_msgs_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.STRUCT_DATA_MSGS_LIST, struct_data_msgs_list)

    def set_sub_notification_queue_depth(self, sub_notification_queue_depth):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUB_NOTIFICATION_QUEUE_DEPTH, sub_notification_queue_depth)

    def set_publish_cos(self, publish_cos):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_COS, publish_cos)

    def set_publish_to_one(self, publish_to_one):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISH_TO_ONE, publish_to_one)

    def set_burst_duration(self, burst_duration):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.BURST_DURATION, burst_duration)

    def set_inter_burst_duration(self, inter_burst_duration):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.INTER_BURST_DURATION, inter_burst_duration)

    def set_pub_send_vect_size(self, pub_send_vect_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUB_SEND_VECT_SIZE, pub_send_vect_size)

    def set_msg_dmq_eligible(self, msg_dmq_eligible):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MSG_DMQ_ELIGIBLE, msg_dmq_eligible)

    def set_msg_ttl(self, msg_ttl):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MSG_TTL, msg_ttl)

    def set_msg_eliding_eligible(self, msg_eliding_eligible):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MSG_ELIDING_ELIGIBLE, msg_eliding_eligible)

    def set_pub_stream_id_offset(self, pub_stream_id_offset):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUB_STREAM_ID_OFFSET, pub_stream_id_offset)

    def set_pub_order_offset(self, pub_order_offset):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUB_ORDER_OFFSET, pub_order_offset)

    def set_subscription_client_name(self, subscription_client_name):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUBSCRIPTION_CLIENT_NAME, subscription_client_name)

    def set_msg_rate_is_max(self, msg_rate_is_max):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MSG_RATE_IS_MAX, msg_rate_is_max)

    def set_ignore_exists_errors(self, ignore_exists_errors):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.IGNORE_EXISTS_ERRORS, ignore_exists_errors)

    def set_sub_msg_queue_depth(self, sub_msg_queue_depth):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUB_MSG_QUEUE_DEPTH, sub_msg_queue_depth)

    def set_want_client_ack(self, want_client_ack):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_CLIENT_ACK, want_client_ack)

    def set_want_client_ack_thread(self, want_client_ack_thread):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_CLIENT_ACK_THREAD, want_client_ack_thread)

    def set_client_ack_skip_num(self, client_ack_skip_num):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_ACK_SKIP_NUM, client_ack_skip_num)

    def set_client_ack_flush_queue(self, client_ack_flush_queue):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_ACK_FLUSH_QUEUE, client_ack_flush_queue)

    def set_client_ack_queue_reverse(self, client_ack_queue_reverse):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_ACK_QUEUE_REVERSE, client_ack_queue_reverse)

    def set_client_ack_random_depth(self, client_ack_random_depth):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_ACK_RANDOM_DEPTH, client_ack_random_depth)

    def set_settlement_pattern_string(self, settlement_pattern_string):
        setattr(self.props, RuntimeProperties.SETTLEMENT_PATTERN_STRING, settlement_pattern_string)

    def set_required_settlement_string(self, required_settlement_string):
        setattr(self.props, RuntimeProperties.REQUIRED_SETTLEMENTS_STRING, required_settlement_string)

    def set_transacted_session_name_list(self, transacted_session_name_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.TRANSACTED_SESSION_NAME_LIST, transacted_session_name_list)

    def set_web_transport_protocol(self, web_transport_protocol):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_TRANSPORT_PROTOCOL, web_transport_protocol)

    def set_web_transport_protocol_list(self, web_transport_protocol_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_TRANSPORT_PROTOCOL_LIST, web_transport_protocol_list)

    def set_transport_protocol_downgrade_timeout_ms(self, transport_protocol_downgrade_timeout_ms):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.TRANSPORT_PROTOCOL_DOWNGRADE_TIMEOUT_MS,
                transport_protocol_downgrade_timeout_ms)

    def set_skip_final_commit(self, skip_final_commit):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SKIP_FINAL_COMMIT, skip_final_commit)

    def set_correlation_id(self, correlation_id):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CORRELATION_ID, correlation_id)

    def set_message_id(self, message_id):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MESSAGE_ID, message_id)

    def set_want_message_priority_stats(self, want_message_priority_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_MESSAGE_PRIORITY_STATS, want_message_priority_stats)

    def set_want_message_priority_order_checking(self, want_message_priority_order_checking):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_MESSAGE_PRIORITY_ORDER_CHECKING,
                want_message_priority_order_checking)

    def set_want_replay(self, want_replay):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_REPLAY, want_replay)

    def set_want_replay_from_date(self, want_replay_from_date):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_REPLAY_FROM_DATE, want_replay_from_date)

    def set_want_replay_from_msg_id(self, want_replay_from_msg_id):
        setattr(self.props, RuntimeProperties.WANT_REPLAY_FROM_MSG_ID, want_replay_from_msg_id)

    def set_reconnect_tries(self, reconnect_tries):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RECONNECT_TRIES, reconnect_tries)

    def set_reconnect_retry_interval_in_msecs(self, reconnect_retry_interval_in_msecs):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RECONNECT_RETRY_INTERVAL_IN_MSECS, reconnect_retry_interval_in_msecs)

    def set_partition_key_list(self, pk_string):
        if pk_string is not None and len(pk_string) != 0:
            partitions = str.split(pk_string, ",")
            setattr(self.props, RuntimeProperties.PARTITION_KEY_LIST, partitions)

    def set_want_per_message_details(self, want_per_message_details):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_PER_MESSAGE_DETAILS, want_per_message_details)

    def set_want_provisioned_endpoint(self, want_provisioned_endpoint):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_PROVISIONED_ENDPOINT, want_provisioned_endpoint)

    def set_pe_permission(self, pe_permission):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PE_PERMISSION, pe_permission)

    def set_pe_access_type(self, pe_access_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PE_ACCESS_TYPE, pe_access_type)

    def set_pe_quota_mb(self, pe_quota_mb):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PE_QUOTA_MB, pe_quota_mb)

    def set_pe_maxmsg_size(self, pe_maxmsg_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PE_MAXMSG_SIZE, pe_maxmsg_size)

    def set_pe_respect_ttl(self, pe_respect_ttl):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PE_RESPECT_TTL, pe_respect_ttl)

    def set_discard_notify_sender(self, discard_notify_sender):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.DISCARD_NOTIFY_SENDER, discard_notify_sender)

    def set_pe_max_msg_redelivery(self, pe_max_msg_redelivery):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PE_MAX_MSG_REDELIVERY, pe_max_msg_redelivery)

    def set_sub_destination_type(self, sub_destination_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUB_DESTINATION_TYPE, sub_destination_type)

    # def set_sub_xpe_lists(self, sub_xpe_lists):
    #     """ Auto-generated property setter """
    #     setattr(self.props, RuntimeProperties.SUB_XPE_LISTS, sub_xpe_lists)

    def set_sub_topic_lists(self, sub_topic_lists):
        self._split_topic_list_among_clients(RuntimeProperties.SUB_TOPIC_LISTS, sub_topic_lists)

    def set_sub_topic_list_from_file_list(self, sub_topic_file_list):
        split_file_list = sub_topic_file_list.split(sep=',')
        topic_lists_list = list()
        for file in split_file_list:
            with open(file, "r") as open_file:
                topic_list = list()
                for line in open_file:
                    topic_list.append(line.strip())
                topic_lists_list.append(topic_list)
        setattr(self.props, RuntimeProperties.SUB_TOPIC_LISTS, topic_lists_list)

    def set_sub_dte_lists(self, sub_dte_lists):
        self._split_topic_list_among_clients(RuntimeProperties.SUB_DTE_LISTS, sub_dte_lists)

    def set_sub_queue_lists(self, sub_queue_lists):
        self._split_topic_list_among_clients(RuntimeProperties.SUB_QUEUE_LISTS, sub_queue_lists)

    def set_sub_queue_list_from_file_list(self, sub_queue_file_list):
        queue_lists_list = list()
        for file in sub_queue_file_list:
            with open(file, "r") as open_file:
                queue_list = list()
                for line in open_file:
                    queue_list.append(str.strip(line))
                queue_lists_list.append(queue_list)
        setattr(self.props, RuntimeProperties.SUB_QUEUE_LISTS, queue_lists_list)

    def set_num_temp_queue_endpoints(self, num_temp_queue_endpoints):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_TEMP_QUEUE_ENDPOINTS, num_temp_queue_endpoints)

    def set_num_temp_topic_endpoints(self, num_temp_topic_endpoints):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUM_TEMP_TOPIC_ENDPOINTS, num_temp_topic_endpoints)

    def set_want_no_local(self, want_no_local):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_NO_LOCAL, want_no_local)

    def set_want_versionprint(self, want_versionprint):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_VERSIONPRINT, want_versionprint)

    def set_want_message_examine_content(self, want_message_examine_content):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_MESSAGE_EXAMINE_CONTENT, want_message_examine_content)

    def set_want_order_check(self, want_order_check):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_ORDER_CHECK, want_order_check)

    def set_want_loss_and_duplicate_detection(self, want_loss_and_duplicate_detection):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_LOSS_AND_DUPLICATE_DETECTION, want_loss_and_duplicate_detection)

    def set_want_order_memory(self, want_order_memory):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_ORDER_MEMORY, want_order_memory)

    def set_want_payload_check(self, want_payload_check):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_PAYLOAD_CHECK, want_payload_check)

    def set_option_file(self, option_file):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.OPTION_FILE, option_file)

    def set_want_latency(self, want_latency):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_LATENCY, want_latency)

    def set_want_full_latency_stats(self, want_full_latency_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_FULL_LATENCY_STATS, want_full_latency_stats)

    def set_latency_buckets(self, latency_buckets):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.LATENCY_BUCKETS, latency_buckets)

    def set_latency_granularity(self, latency_granularity):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.LATENCY_GRANULARITY, latency_granularity)

    def set_latency_warmup_in_secs(self, latency_warmup_in_secs):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.LATENCY_WARMUP_IN_SECS, latency_warmup_in_secs)

    def set_jms_suback_mode(self, jms_suback_mode):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.JMS_SUBACK_MODE, jms_suback_mode)

    def set_jms_connection_factory(self, jms_connection_factory):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.JMS_CONNECTION_FACTORY, jms_connection_factory)

    def set_want_jndi_for_jms(self, want_jndi_for_jms):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_JNDI_FOR_JMS, want_jndi_for_jms)

    def set_solace_jms_optimize_direct(self, solace_jms_optimize_direct):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SOLACE_JMS_OPTIMIZE_DIRECT, solace_jms_optimize_direct)

    def set_application_server_ra_transaction(self, application_server_ra_transaction):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.APPLICATION_SERVER_RA_TRANSACTION, application_server_ra_transaction)

    def set_subscriber_sub_delay(self, subscriber_sub_delay):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUBSCRIBER_SUB_DELAY, subscriber_sub_delay)

    def set_slow_sub_delay_count(self, slow_sub_delay_count):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SLOW_SUB_DELAY_COUNT, slow_sub_delay_count)

    def set_subscriber_rate_interval(self, subscriber_rate_interval):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUBSCRIBER_RATE_INTERVAL, subscriber_rate_interval)

    def set_cache_prop_name(self, cache_prop_name):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_PROP_NAME, cache_prop_name)

    def set_cache_prop_max_msgs_per_topic(self, cache_prop_max_msgs_per_topic):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_PROP_MAX_MSGS_PER_TOPIC, cache_prop_max_msgs_per_topic)

    def set_cache_prop_max_age(self, cache_prop_max_age):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_PROP_MAX_AGE, cache_prop_max_age)

    def set_cache_prop_timeout_in_msec(self, cache_prop_timeout_in_msec):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_PROP_TIMEOUT_IN_MSEC, cache_prop_timeout_in_msec)

    def set_cache_want_requests_on_subscribe(self, cache_want_requests_on_subscribe):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_WANT_REQUESTS_ON_SUBSCRIBE, cache_want_requests_on_subscribe)

    def set_cache_req_msg_rate(self, cache_req_msg_rate):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_REQ_MSG_RATE, cache_req_msg_rate)

    def set_cache_num_req(self, cache_num_req):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_NUM_REQ, cache_num_req)

    def set_cache_live_data_action(self, cache_live_data_action):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_LIVE_DATA_ACTION, cache_live_data_action)

    def set_cache_want_stats(self, cache_want_stats):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_WANT_STATS, cache_want_stats)

    def set_cache_req_min_sequence_num(self, cache_req_min_sequence_num):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_REQ_MIN_SEQUENCE_NUM, cache_req_min_sequence_num)

    def set_cache_req_max_sequence_num(self, cache_req_max_sequence_num):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_REQ_MAX_SEQUENCE_NUM, cache_req_max_sequence_num)

    def set_cache_req_wait_for_confirm(self, cache_req_wait_for_confirm):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CACHE_REQ_WAIT_FOR_CONFIRM, cache_req_wait_for_confirm)

    def set_ssl_protocol(self, ssl_protocol):
        excluded = list()
        if isinstance(ssl_protocol, str):
            allowed_protocols = [x.lower() for x in ssl_protocol.split(sep=",")]
        elif isinstance(ssl_protocol, list):
            allowed_protocols = ssl_protocol
        else:
            allowed_protocols = list()

        for protocol in self.SSL_PROTOCOLS_MAP.keys():
            if protocol.lower() not in allowed_protocols:
                excluded.append(self.SSL_PROTOCOLS_MAP[protocol])
        if len(excluded) > 0:
            self.set_ssl_excluded_protocols(excluded)

    def set_ssl_connection_downgrade_to(self, ssl_connection_downgrade_to):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_CONNECTION_DOWNGRADE_TO, ssl_connection_downgrade_to)

    def set_ssl_excluded_protocols(self, ssl_excluded_protocols):
        if isinstance(ssl_excluded_protocols, str):
            excluded = list()
            excluded_protocols = [x.lower() for x in ssl_excluded_protocols.split(sep=",")]
            for protocol in self.SSL_PROTOCOLS_MAP.keys():
                if protocol.lower() in excluded_protocols:
                    excluded.append(self.SSL_PROTOCOLS_MAP[protocol])
            if len(excluded) > 0:
                setattr(self.props, RuntimeProperties.SSL_EXLCUDED_PROTOCOLS, excluded)
        elif isinstance(ssl_excluded_protocols, list):
            setattr(self.props, RuntimeProperties.SSL_EXLCUDED_PROTOCOLS, ssl_excluded_protocols)
        else:
            raise Exception("Invalid input to set_ssl_exlcuded_protocols, must pass string or list")

    def set_ssl_client_certificate_file(self, ssl_client_certificate_file):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_CLIENT_CERTIFICATE_FILE, ssl_client_certificate_file)

    def set_ssl_client_private_key_file(self, ssl_client_private_key_file):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_CLIENT_PRIVATE_KEY_FILE, ssl_client_private_key_file)

    def set_ssl_client_private_key_file_password(self, ssl_client_private_key_file_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_CLIENT_PRIVATE_KEY_FILE_PASSWORD,
                ssl_client_private_key_file_password)

    def set_ssl_validate_certificate(self, ssl_validate_certificate):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_VALIDATE_CERTIFICATE, ssl_validate_certificate)

    def set_ssl_validate_certificate_date(self, ssl_validate_certificate_date):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_VALIDATE_CERTIFICATE_DATE, ssl_validate_certificate_date)

    def set_ssl_validate_certificate_hostname(self, ssl_validate_certificate_hostname):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_VALIDATE_CERTIFICATE_HOSTNAME, ssl_validate_certificate_hostname)

    def set_ssl_cipher_suites(self, ssl_cipher_suites):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_CIPHER_SUITES, ssl_cipher_suites)

    def set_ssl_trust_store(self, ssl_trust_store):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_TRUST_STORE, ssl_trust_store)

    def set_ssl_trust_store_dir(self, ssl_trust_store_dir):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_TRUST_STORE_DIR, ssl_trust_store_dir)

    def set_ssl_trust_store_password(self, ssl_trust_store_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_TRUST_STORE_PASSWORD, ssl_trust_store_password)

    def set_ssl_trust_store_format(self, ssl_trust_store_format):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_TRUST_STORE_FORMAT, ssl_trust_store_format)

    def set_ssl_trusted_common_name_list(self, ssl_trusted_common_name_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_TRUSTED_COMMON_NAME_LIST, ssl_trusted_common_name_list)

    def set_ssl_key_store(self, ssl_key_store):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_KEY_STORE, ssl_key_store)

    def set_ssl_key_store_password(self, ssl_key_store_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_KEY_STORE_PASSWORD, ssl_key_store_password)

    def set_ssl_key_store_format(self, ssl_key_store_format):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_KEY_STORE_FORMAT, ssl_key_store_format)

    def set_ssl_private_key_alias(self, ssl_private_key_alias):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_PRIVATE_KEY_ALIAS, ssl_private_key_alias)

    def set_ssl_private_key_password(self, ssl_private_key_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_PRIVATE_KEY_PASSWORD, ssl_private_key_password)

    def set_ssl_rest_key_store(self, ssl_rest_key_store):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_KEY_STORE, ssl_rest_key_store)

    def set_ssl_rest_key_store_password(self, ssl_rest_key_store_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_KEY_STORE_PASSWORD, ssl_rest_key_store_password)

    def set_ssl_rest_trust_store(self, ssl_rest_trust_store):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_TRUST_STORE, ssl_rest_trust_store)

    def set_ssl_rest_trust_store_password(self, ssl_rest_trust_store_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_TRUST_STORE_PASSWORD, ssl_rest_trust_store_password)

    def set_ssl_rest_private_key_password(self, ssl_rest_private_key_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_PRIVATE_KEY_PASSWORD, ssl_rest_private_key_password)

    def set_ssl_rest_want_client_auth(self, ssl_rest_want_client_auth):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_WANT_CLIENT_AUTH, ssl_rest_want_client_auth)

    def set_ssl_rest_need_client_auth(self, ssl_rest_need_client_auth):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_NEED_CLIENT_AUTH, ssl_rest_need_client_auth)

    def set_ssl_rest_excluded_cipher_suites(self, ssl_rest_excluded_cipher_suites):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SSL_REST_EXCLUDED_CIPHER_SUITES, ssl_rest_excluded_cipher_suites)

    def set_jaas_login_context(self, jaas_login_context):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.JAAS_LOGIN_CONTEXT, jaas_login_context)

    def set_jaas_config_file_reload_enabled(self, jaas_config_file_reload_enabled):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.JAAS_CONFIG_FILE_RELOAD_ENABLED, jaas_config_file_reload_enabled)

    def set_krb_service_name(self, krb_service_name):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.KRB_SERVICE_NAME, krb_service_name)

    def set_authentication_scheme(self, authentication_scheme):
        if type(authentication_scheme) == AuthenticationScheme:
            setattr(self.props, RuntimeProperties.AUTHENTICATION_SCHEME, authentication_scheme)
        if type(authentication_scheme) == str:
            if authentication_scheme.upper() == AuthenticationScheme.get_name(AuthenticationScheme.BASIC):
                setattr(self.props, RuntimeProperties.AUTHENTICATION_SCHEME, AuthenticationScheme.BASIC)
            elif authentication_scheme.upper() == AuthenticationScheme.get_name(
                    AuthenticationScheme.CLIENT_CERTIFICATE):
                setattr(self.props, RuntimeProperties.AUTHENTICATION_SCHEME, AuthenticationScheme.CLIENT_CERTIFICATE)
            elif authentication_scheme.upper() == AuthenticationScheme.get_name(
                    AuthenticationScheme.GSS_KRB):
                setattr(self.props, RuntimeProperties.AUTHENTICATION_SCHEME, AuthenticationScheme.GSS_KRB)
            elif authentication_scheme.upper() == AuthenticationScheme.get_name(AuthenticationScheme.OAUTH2):
                setattr(self.props, RuntimeProperties.AUTHENTICATION_SCHEME, AuthenticationScheme.OAUTH2)
            else:
                raise Exception(f"Invalid authentication scheme: \"{authentication_scheme}\". See help for valid "
                                f"authentication schemes.")

    def set_oauth_access_token(self, oauth_token):
        setattr(self.props, RuntimeProperties.OAUTH_ACCESS_TOKEN, oauth_token)

    def set_oauth_issuer_id(self, issuer_id):
        setattr(self.props, RuntimeProperties.OAUTH_ISSUER_ID, issuer_id)

    def set_oauth_id_token(self, id_token):
        setattr(self.props, RuntimeProperties.OAUTH_ID_TOKEN, id_token)

    def set_read_timeout_in_msec(self, read_timeout_in_msec):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.READ_TIMEOUT_IN_MSEC, read_timeout_in_msec)

    def set_keepalive_interval_msec(self, keepalive_interval_msec):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.KEEPALIVE_INTERVAL_MSEC, keepalive_interval_msec)

    def set_reconnect_attempts(self, reconnect_attempts):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RECONNECT_ATTEMPTS, reconnect_attempts)

    def set_reconnect_fail_action(self, reconnect_fail_action):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RECONNECT_FAIL_ACTION, reconnect_fail_action)

    def set_reconnect_interval_msec(self, reconnect_interval_msec):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RECONNECT_INTERVAL_MSEC, reconnect_interval_msec)

    def set_want_queue_browser(self, want_queue_browser):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_QUEUE_BROWSER, want_queue_browser)

    def set_qb_delete_frequency(self, qb_delete_frequency):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.QB_DELETE_FREQUENCY, qb_delete_frequency)

    def set_qb_delete_buffer(self, qb_delete_buffer):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.QB_DELETE_BUFFER, qb_delete_buffer)

    def set_web_messaging_goto_url(self, web_messaging_goto_url):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_MESSAGING_GOTO_URL, web_messaging_goto_url)

    def set_web_messaging_browser_type(self, web_messaging_browser_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_MESSAGING_BROWSER_TYPE, web_messaging_browser_type)

    def set_web_messaging_transport_scheme(self, web_messaging_transport_scheme):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_MESSAGING_TRANSPORT_SCHEME, web_messaging_transport_scheme)

    def set_web_messaging_use_proxy(self, web_messaging_use_proxy):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_MESSAGING_USE_PROXY, web_messaging_use_proxy)

    def set_want_dispatch_session(self, want_dispatch_session):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_DISPATCH_SESSION, want_dispatch_session)

    def set_web_messaging_send_buffer_max_size(self, web_messaging_send_buffer_max_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_MESSAGING_SEND_BUFFER_MAX_SIZE, web_messaging_send_buffer_max_size)

    def set_web_messaging_max_web_payload(self, web_messaging_max_web_payload):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WEB_MESSAGING_MAX_WEB_PAYLOAD, web_messaging_max_web_payload)

    def set_want_context_per_cient(self, want_context_per_cient):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_CONTEXT_PER_CIENT, want_context_per_cient)

    def set_want_stop_on_error(self, want_stop_on_error):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_STOP_ON_ERROR, want_stop_on_error)

    def set_extra_prop_dict(self, extra_prop_list):
        if len(extra_prop_list) % 2 != 0:
            raise Exception(f"Extra prop list size must be a multiple of 2, given size is {len(extra_prop_list)}")
        extra_prop_dict = dict()
        for i in range(0, len(extra_prop_list), 2):
            extra_prop_dict[extra_prop_list[i]] = extra_prop_list[i + 1]
        setattr(self.props, RuntimeProperties.EXTRA_PROP_DICT, extra_prop_dict)

    def set_want_msg_dump(self, want_msg_dump):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_MSG_DUMP, want_msg_dump)

    def set_msg_dump_dir(self, msg_dump_dir):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MSG_DUMP_DIR, msg_dump_dir)

    def set_client_mode(self, client_mode):
        if client_mode == "sink":
            setattr(self.props, RuntimeProperties.CLIENT_MODE, ClientModeType.SINK)
            self.set_want_msg_reflect(False)
        else:
            setattr(self.props, RuntimeProperties.CLIENT_MODE, ClientModeType.REPLY)
            self.set_want_msg_reflect(True)

    def set_want_msg_reflect(self, want_msg_reflect):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_MSG_REFLECT, want_msg_reflect)

    def set_want_user_prop_tool_data(self, want_user_prop_tool_data):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_USER_PROP_TOOL_DATA, want_user_prop_tool_data)

    def set_want_non_blocking_connect(self, want_non_blocking_connect):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_NON_BLOCKING_CONNECT, want_non_blocking_connect)

    def set_last_message_queue_depth(self, last_message_queue_depth):
        setattr(self.props, RuntimeProperties.LAST_MESSAGE_QUEUE_DEPTH, last_message_queue_depth)

    def set_message_callback_on_reactor(self, message_callback_on_reactor):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MESSAGE_CALLBACK_ON_REACTOR, message_callback_on_reactor)

    def set_want_reply_topic(self, want_reply_topic):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_REPLY_TOPIC, want_reply_topic)

    def set_want_reply_temporary_queue(self, want_reply_temporary_queue):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_REPLY_TEMPORARY_QUEUE, want_reply_temporary_queue)

    def set_reply_to_postfix(self, reply_to_postfix):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.REPLY_TO_POSTFIX, reply_to_postfix)

    def set_reply_to_string(self, reply_to_string):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.REPLY_TO_STRING, reply_to_string)

    def set_want_replication_reconnect(self, want_replication_reconnect):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_REPLICATION_RECONNECT, want_replication_reconnect)

    def set_want_cut_through_persistence(self, want_cut_through_persistence):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_CUT_THROUGH_PERSISTENCE, want_cut_through_persistence)

    def set_tool_mode(self, tool_mode):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.TOOL_MODE, tool_mode)

    def set_api_mode(self, api_mode):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.API_MODE, api_mode)

    def set_tool_third_party_external_client_class(self, tool_third_party_external_client_class):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.TOOL_THIRD_PARTY_EXTERNAL_CLIENT_CLASS,
                tool_third_party_external_client_class)

    def set_number_of_msgs_to_pub_on_rcv(self, number_of_msgs_to_pub_on_rcv):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.NUMBER_OF_MSGS_TO_PUB_ON_RCV, number_of_msgs_to_pub_on_rcv)

    def set_por_topic_list(self, por_topic_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.POR_TOPIC_LIST, por_topic_list)

    def set_por_queue_list(self, por_queue_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.POR_QUEUE_LIST, por_queue_list)

    def set_por_message_type(self, por_message_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.POR_MESSAGE_TYPE, por_message_type)

    def set_por_attachment_size_list(self, por_attachment_size_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.POR_ATTACHMENT_SIZE_LIST, por_attachment_size_list)

    def set_por_reflect_message(self, por_reflect_message):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.POR_REFLECT_MESSAGE, por_reflect_message)

    def set_application_server_command(self, application_server_command):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.APPLICATION_SERVER_COMMAND, application_server_command)

    def set_server_port_list(self, server_port_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SERVER_PORT_LIST, server_port_list)

    def set_rest_client_mode(self, rest_client_mode):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTCLIENT_MODE, rest_client_mode)

    def set_rest_data_mode(self, rest_data_mode):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.REST_DATA_MODE, rest_data_mode)

    def set_restserver_connection_type(self, restserver_connection_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_CONNECTION_TYPE, restserver_connection_type)

    def set_restserver_auth_username(self, restserver_auth_username):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_AUTH_USERNAME, restserver_auth_username)

    def set_restserver_auth_password(self, restserver_auth_password):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_AUTH_PASSWORD, restserver_auth_password)

    def set_restserver_auth_http_header_name(self, restserver_auth_http_header_name):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_AUTH_HTTP_HEADER_NAME, restserver_auth_http_header_name)

    def set_restserver_auth_http_header_value(self, restserver_auth_http_header_value):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_AUTH_HTTP_HEADER_VALUE, restserver_auth_http_header_value)

    def set_restserver_response_content(self, restserver_response_content):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_RESPONSE_CONTENT, restserver_response_content)

    def set_restserver_num_threads(self, restserver_num_threads):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_NUM_THREADS, restserver_num_threads)

    def set_rest_ip_list(self, rest_ip_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.REST_IP_LIST, rest_ip_list)

    def set_restserver_prob_resp(self, restserver_prob_resp):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.RESTSERVER_PROB_RESP, restserver_prob_resp)

    def set_message_reply_delay_list(self, message_reply_delay_list):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MESSAGE_REPLY_DELAY_LIST, message_reply_delay_list)

    def set_mqtt_publish_qos(self, mqtt_publish_qos):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_PUBLISH_QOS, mqtt_publish_qos)

    def set_mqtt_want_publish_retain(self, mqtt_want_publish_retain):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_WANT_PUBLISH_RETAIN, mqtt_want_publish_retain)

    def set_mqtt_disable_ssl_hostname_verification(self, mqtt_disable_ssl_hostname_verification):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_DISABLE_SSL_HOSTNAME_VERIFICATION,
                mqtt_disable_ssl_hostname_verification)

    def set_mqtt_subscribe_qos(self, mqtt_subscribe_qos):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_SUBSCRIBE_QOS, mqtt_subscribe_qos)

    def set_mqtt_will_topic(self, mqtt_will_topic):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_WILL_TOPIC, mqtt_will_topic)

    def set_mqtt_will_qos(self, mqtt_will_qos):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_WILL_QOS, mqtt_will_qos)

    def set_mqtt_will_retained(self, mqtt_will_retained):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_WILL_RETAINED, mqtt_will_retained)

    def set_mqtt_will_msg_size(self, mqtt_will_msg_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_WILL_MSG_SIZE, mqtt_will_msg_size)

    def set_mqtt_want_clean_connect(self, mqtt_want_clean_connect):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT_WANT_CLEAN_CONNECT, mqtt_want_clean_connect)

    def set_mqtt5_session_expiry_interval(self, mqtt5_session_expiry_interval):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT5_SESSION_EXPIRY_INTERVAL, mqtt5_session_expiry_interval)

    def set_mqtt5_receive_maximum(self, mqtt5_receive_maximum):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT5_RECEIVE_MAXIMUM, mqtt5_receive_maximum)

    def set_mqtt5_request_response_info(self, mqtt5_request_response_info):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT5_REQUEST_RESPONSE_INFO, mqtt5_request_response_info)

    def set_mqtt5_request_problem_info(self, mqtt5_request_problem_info):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT5_REQUEST_PROBLEM_INFO, mqtt5_request_problem_info)

    def set_mqtt5_send_reason_messages(self, mqtt5_send_reason_messages):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT5_SEND_REASON_MESSAGES, mqtt5_send_reason_messages)

    def set_mqtt5_will_delay_interval(self, mqtt5_will_delay_interval):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.MQTT5_WILL_DELAY_INTERVAL, mqtt5_will_delay_interval)

    def set_custom_properties_list(self, custom_properties_list):
        """ Auto-generated property setter """
        custom_props = dict()
        if custom_properties_list is not None and len(custom_properties_list) != 0:
            tokens = str(custom_properties_list).split(",")
            if len(tokens) % 3 != 0:
                raise Exception("Custom props list length must be a multiple of 3 (Type,Key,Value)")
            for i in range(0, len(tokens), 3):
                if str(tokens[i]).upper() == "STRING":
                    custom_props[tokens[i+1]] = str(tokens[i+2])
                elif str(tokens[i]).upper() == "BOOLEAN":
                    custom_props[tokens[i+1]] = bool(tokens[i+2])
                elif str(tokens[i]).upper().startswith("INT") or \
                        str(tokens[i]).upper() == "SHORT" or \
                        str(tokens[i]).upper() == "LONG":
                    custom_props[tokens[i + 1]] = int(tokens[i + 2])
                elif str(tokens[i]).upper() == "DOUBLE" or \
                        str(tokens[i]).upper() == "FLOAT":
                    custom_props[tokens[i + 1]] = float(tokens[i + 2])

        if len(custom_props) != 0:
            setattr(self.props, RuntimeProperties.CUSTOM_PROPERTIES_LIST, custom_props)

    def set_client_version(self, client_version):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.CLIENT_VERSION, client_version)

    def set_http_content_encoding(self, http_content_encoding):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.HTTP_CONTENT_ENCODING, http_content_encoding)

    def set_http_content_type(self, http_content_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.HTTP_CONTENT_TYPE, http_content_type)

    def set_publisher_backpressure_type(self, publisher_backpressure_type):
        backpressure_types = list()
        for pbp_type in publisher_backpressure_type:
            if isinstance(pbp_type, PublisherBackPressureType):
                backpressure_types.append(pbp_type)
            elif isinstance(pbp_type, str):
                if pbp_type.upper() == PublisherBackPressureType.NO.name:
                    backpressure_types.append(PublisherBackPressureType.NO)
                elif pbp_type.upper() == PublisherBackPressureType.REJECT.name:
                    backpressure_types.append(PublisherBackPressureType.REJECT)
                elif pbp_type.upper() == PublisherBackPressureType.WAIT.name:
                    backpressure_types.append(PublisherBackPressureType.WAIT)
                elif pbp_type.upper() == PublisherBackPressureType.ELASTIC.name:
                    backpressure_types.append(PublisherBackPressureType.ELASTIC)
                else:
                    raise Exception(f"Invalid publisher backpressure type: \"{pbp_type}\". See help for valid "
                                    f"backpressure types.")
        setattr(self.props, RuntimeProperties.PUBLISHER_BACKPRESSURE_TYPE, backpressure_types)

    def set_subscriber_backpressure_type(self, subscriber_backpressure_type):

        if isinstance(subscriber_backpressure_type, SubscriberBackPressureType):
            setattr(self.props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_TYPE, subscriber_backpressure_type)
        elif isinstance(subscriber_backpressure_type, str):
            if subscriber_backpressure_type.upper() == SubscriberBackPressureType.ELASTIC.name:
                setattr(self.props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_TYPE, SubscriberBackPressureType.ELASTIC)
            elif subscriber_backpressure_type.upper() == SubscriberBackPressureType.OLDEST.name:
                setattr(self.props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_TYPE, SubscriberBackPressureType.OLDEST)
            elif subscriber_backpressure_type.upper() == SubscriberBackPressureType.LATEST.name:
                setattr(self.props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_TYPE, SubscriberBackPressureType.LATEST)
            else:
                raise Exception(f"invalid subscriber backpressure type: \"{subscriber_backpressure_type}\" "
                                f"See help for valid backpressure types")

    def set_publisher_backpressure_buffer_size(self, publisher_backpressure_buffer_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.PUBLISHER_BACKPRESSURE_BUFFER_SIZE, publisher_backpressure_buffer_size)

    def set_subscriber_backpressure_buffer_size(self, subscriber_backpressure_buffer_size):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_BUFFER_SIZE, subscriber_backpressure_buffer_size)

    def set_receiver_sleep_time(self, receiver_sleep_time):
        setattr(self.props, RuntimeProperties.RECEIVER_SLEEP_TIME, receiver_sleep_time)

    def set_want_termination_listeners(self, want_termination_listeners):
        setattr(self.props, RuntimeProperties.WANT_TERMINATION_LISTENERS, want_termination_listeners)

    def set_want_ungraceful_terminate(self, want_ungraceful_terminate):
        setattr(self.props, RuntimeProperties.WANT_UNGRACEFUL_TERMINATE, want_ungraceful_terminate)

    def set_want_ungraceful_disconnect(self, want_ungraceful_disconnect):
        setattr(self.props, RuntimeProperties.WANT_UNGRACEFUL_DISCONNECT, want_ungraceful_disconnect)

    def set_grace_period(self, grace_period):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.GRACE_PERIOD, grace_period)

    def set_want_producer_listeners(self, want_producer_listeners):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_PRODUCER_LISTENERS, want_producer_listeners)

    def set_want_publish_await_ack(self, want_publish_await_ack):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_PUBLISH_AWAIT_ACK, want_publish_await_ack)

    def set_want_single_pub_type(self, want_single_pub_type):
        """ Auto-generated property setter """
        setattr(self.props, RuntimeProperties.WANT_SINGLE_PUB_TYPE, want_single_pub_type)

    def set_want_trace_creation_context(self, want_trace_creation_context):
        setattr(self.props, RuntimeProperties.WANT_TRACE_CREATION_CONTEXT, want_trace_creation_context)

    def set_trace_baggage(self, trace_baggage):
        setattr(self.props, RuntimeProperties.TRACE_BAGGAGE, trace_baggage)

    def set_want_trace_transport_context(self, want_trace_transport_context):
        setattr(self.props, RuntimeProperties.WANT_TRACE_TRANSPORT_CONTEXT, want_trace_transport_context)

    def set_want_trace_copy_receive_context(self, want_trace_copy_receive_context):
        setattr(self.props, RuntimeProperties.WANT_TRACE_COPY_RECEIVE_CONTEXT, want_trace_copy_receive_context)

    def set_trace_exporter_connection_string(self, trace_exporter_connection_string):
        setattr(self.props, RuntimeProperties.TRACE_EXPORTER_CONNECTION_STRING, trace_exporter_connection_string)

    def set_trace_exporter_batch_size(self, trace_exporter_batch_size):
        setattr(self.props, RuntimeProperties.TRACE_EXPORTER_BATCH_SIZE, trace_exporter_batch_size)

    def set_trace_exporter_schedule_delay(self, trace_exporter_schedule_delay):
        setattr(self.props, RuntimeProperties.TRACE_EXPORTER_SCHEDULE_DELAY, trace_exporter_schedule_delay)

    def set_trace_exporter_queue_size(self, trace_exporter_queue_size):
        setattr(self.props, RuntimeProperties.TRACE_EXPORTER_QUEUE_SIZE, trace_exporter_queue_size)

    def set_want_trace_context_prop(self, want_trace_context_prop):
        setattr(self.props, RuntimeProperties.WANT_TRACE_CONTEXT_PROP, want_trace_context_prop)

    def _split_topic_list_among_clients(self, prop, topic_str):
        topic_lists = list()
        i = 0
        if isinstance(topic_str, str):
            topics = topic_str.split(sep=",")
        elif isinstance(topic_str, list):
            topics = topic_str
        else:
            raise Exception("Unknown type to _split_topic_list_among_clients")
        num_clients = getattr(self.props, RuntimeProperties.NUM_CLIENTS, 1)
        for item in topics:
            if i < num_clients:
                inner_list = list()
                inner_list.append(item)
                topic_lists.append(inner_list)
            else:
                topic_lists[i % num_clients].append(item)
            i += 1
        setattr(self.props, prop, topic_lists)

    @classmethod
    def create_default_properties(cls):
        ret = cls()

        ret.set_reconnect_attempts(3)
        ret.set_reconnect_interval_msec(3000)
        ret.set_num_clients(0)
        ret.set_tool_mode("sdkperf")
        ret.set_api_mode("python")
        ret.set_want_payload_check(False)
        ret.set_want_ud(False)
        ret.set_want_order_check(False)
        ret.set_want_loss_and_duplicate_detection(False)
        ret.set_want_order_memory(False)
        ret.set_want_message_examine_content(False)
        ret.set_num_pubs_per_session(1)
        ret.set_num_msgs_to_publish(0)
        ret.set_publish_end_delay_in_sec(2)
        ret.set_client_username("")
        ret.set_client_vpn("")
        ret.set_client_name_prefix("")
        ret.set_client_password("")
        ret.set_client_description("")  # Not accessible via cli.
        ret.set_change_client_names_flag(False)
        ret.set_want_tcp_nodelay(True)  # DO NOT set the nodelay option
        ret.set_pub_message_type(GenericMessageDeliveryMode.DIRECT)
        ret.set_want_msg_obj_type_check(False)
        ret.set_subscriber_sub_delay(0)
        ret.set_slow_sub_delay_count(-1)
        ret.set_subscriber_rate_interval(1)
        ret.set_want_remove_subscriber(True)
        ret.set_want_subscription_memory(True)
        ret.set_want_per_sub_thru_stats(False)
        ret.set_want_subscription_rate_stats(False)
        ret.set_sub_notification_queue_depth(-1)
        ret.set_cid_stats(False)
        ret.set_want_di_stats(False)
        ret.set_want_verbose(False)
        ret.set_want_quiet(False)
        ret.set_want_versionprint(False)
        ret.set_want_latency(False)
        ret.set_want_full_latency_stats(True)
        ret.set_latency_buckets(1024)
        ret.set_latency_granularity(0)
        ret.set_latency_warmup_in_secs(30.0)
        ret.set_want_smooth_pub_calc_lat(False)
        ret.set_want_struct_msg_check(False)
        ret.set_publish_rate_per_pub(1)
        ret.set_cache_want_stats(False)
        ret.set_cache_num_req(0)
        ret.set_cache_req_min_sequence_num(-1)
        ret.set_cache_req_max_sequence_num(-1)
        ret.set_cache_req_wait_for_confirm(False)
        ret.set_num_temp_queue_endpoints(0)
        ret.set_num_temp_topic_endpoints(0)
        ret.set_want_context_per_cient(False)
        ret.set_burst_duration(0)
        ret.set_inter_burst_duration(0)
        ret.set_want_provisioned_endpoint(False)
        ret.set_pe_access_type(EndpointAccessType.ACCESSTYPE_API_DEFAULT)
        ret.set_pe_maxmsg_size(-1)
        ret.set_pe_quota_mb(-1)
        ret.set_pe_max_msg_redelivery(-1)
        ret.set_pe_permission("")
        ret.set_pe_respect_ttl(False)
        ret.set_want_stop_on_error(False)
        ret.set_want_queue_browser(False)
        ret.set_qb_delete_frequency(-1)
        ret.set_qb_delete_buffer(-1)
        ret.set_pub_send_vect_size(0)
        ret.set_ad_pub_window_size(50)
        ret.set_want_msg_dump(False)
        ret.set_msg_dump_dir("")
        ret.set_client_mode(ClientModeType.SINK)
        ret.set_want_msg_reflect(False)
        ret.set_want_reply_topic(False)
        ret.set_want_reply_temporary_queue(False)
        ret.set_reply_to_postfix("")
        ret.set_reply_to_string("")
        ret.set_pub_stream_id_offset(-1)
        ret.set_pub_order_offset(0)
        ret.set_subscription_client_name("")
        ret.set_want_cut_through_persistence(False)
        ret.set_want_user_prop_tool_data(False)
        ret.set_msg_rate_is_max(False)
        ret.set_ignore_exists_errors(True)
        ret.set_want_client_ack(False)
        ret.set_want_client_ack_thread(False)
        ret.set_client_ack_skip_num(0)
        ret.set_client_ack_flush_queue(False)
        ret.set_client_ack_queue_reverse(False)
        ret.set_client_ack_random_depth(0)
        ret.set_sub_msg_queue_depth(0)
        ret.set_want_sync_receive(False)
        ret.set_want_no_local(False)
        ret.set_want_quiet(False)
        ret.set_ad_ack_immediately_interval(0)
        ret.set_ssl_connection_downgrade_to("")
        ret.set_ssl_validate_certificate(False)
        ret.set_ssl_validate_certificate_date(False)
        ret.set_ssl_validate_certificate_hostname(False)
        ret.set_ssl_cipher_suites("")
        ret.set_ssl_trust_store("")
        ret.set_ssl_trust_store_dir("")
        ret.set_ssl_trust_store_password("")
        ret.set_ssl_trust_store_format("")
        ret.set_ssl_trusted_common_name_list("")
        ret.set_ssl_key_store("")
        ret.set_ssl_key_store_password("")
        ret.set_ssl_key_store_password("")
        ret.set_ssl_private_key_alias("")
        ret.set_ssl_private_key_password(None)
        ret.set_ssl_client_certificate_file("")
        ret.set_ssl_client_private_key_file("")
        ret.set_ssl_client_private_key_file_password(None)
        ret.set_authentication_scheme(AuthenticationScheme.BASIC)
        ret.set_ad_want_active_flow_indication(False)
        ret.set_ad_want_pub_ack_order_check(True)
        ret.set_want_replication_reconnect(False)
        ret.set_application_server_command("")
        ret.set_num_pub_threads(1)
        ret.set_want_per_message_details(False)
        ret.set_num_message_producer(1)
        ret.set_number_of_msgs_to_pub_on_rcv(0)
        ret.set_por_topic_list("")
        ret.set_por_queue_list("")
        ret.set_por_message_type("direct")
        ret.set_por_attachment_size_list(list())
        ret.set_por_reflect_message(False)
        ret.set_correlation_id("")
        ret.set_message_id("")
        ret.set_want_message_priority_stats(False)
        ret.set_want_message_priority_order_checking(False)
        ret.set_want_replay(False)
        ret.set_want_replay_from_date("")
        ret.set_reconnect_tries(-1)
        ret.set_reconnect_retry_interval_in_msecs(3000)
        ret.set_want_single_pub_type(False)
        ret.set_ad_transaction_size(0)
        ret.set_ad_transaction_idle_time(0)
        ret.set_ad_transaction_rollback_interval(-1)
        ret.set_skip_final_commit(False)
        ret.set_trace_exporter_batch_size(5)
        ret.set_trace_exporter_schedule_delay(100)
        ret.set_trace_exporter_queue_size(2048)
        return ret
