import argparse
import os
import platform
import sys
from multiprocessing import Manager
from xmlrpc.server import SimpleXMLRPCRequestHandler
from xmlrpc.server import SimpleXMLRPCServer

import pkg_resources

from pypubsubtools.core.perf_client_proxy import PerfClientProxy
from pypubsubtools.utils.log import get_logger, init_logging, set_python_solace_log_level

logger = get_logger()


# Can be changed to restrict to a particular path.
class RequestHandler(SimpleXMLRPCRequestHandler):
    """Do nothing for now"""


class PubSubServer:
    # Return code for successful calls
    SUCCESS_RETURN = {"rc": 0}

    def __init__(self):
        self._ip = 'localhost'
        self._port = 8000
        self._shutdown_queue = Manager().Queue(1)
        self._server = None
        # They are stored by the ID and PerfClientProxy object
        self._clients = dict()

    def create_client(self, unique_id: str, mode: str):
        logger.warning(f"PubSubServer manager.CreateClient {unique_id}")
        table = dict()
        if unique_id in self._clients:
            table["rc"] = 1
            table["errorInfo"] = f"{unique_id} already exists...need UNIQUE ID"
        else:
            proxy = PerfClientProxy.create(unique_id, mode)
            self._clients[unique_id] = proxy
            table = dict(self.SUCCESS_RETURN)
            table["result"] = f"ClientId: {unique_id}: Type: {mode} - Created"
            return table

    def _dispatch(self, method, params):
        logger.warning(f"{method}(...), param dump follows\n{params}")
        uid, method_name = method.split('.')
        if uid in self._clients:
            proxy_obj = self._clients[uid]
            if hasattr(proxy_obj, method_name):
                result = getattr(proxy_obj, method_name)(*params)
                return result
            else:
                logger.error(f"METHOD DOES NOT EXIST: {method_name}")
                return {'rc': 400, "errorInfo": "Method name does not exist."}
        else:
            logger.error(f"UID DOES NOT EXIST: {uid}")
            return {'rc': 400, "errorInfo": "Unique id does not exist."}

    def delete_client(self, unique_id: str):
        logger.warning(f"PubSubServer manager.DeleteClient {unique_id}")
        proxy = self._clients.pop(unique_id)
        proxy.destroy()
        del proxy
        proxy = None
        return self.SUCCESS_RETURN

    def get_clients(self):
        logger.debug("PubSubServer manager.GetClients")
        table = dict(self.SUCCESS_RETURN)
        table["clients"] = list(self._clients.keys())
        return table

    def insert_log(self, level: str, msg):
        logger.debug("PubSubServer manager.InsertLog")
        if level.lower() == "debug":
            logger.debug(msg)
        elif level.lower() == "info":
            logger.info(msg)
        elif level.lower() == "warn":
            logger.warning(msg)
        elif level.lower() == "error":
            logger.error(msg)
        else:
            logger.info(f"Level {level} not understood: {msg}")
        return self.SUCCESS_RETURN

    def get_version_info(self):
        logger.debug("PubSubServer manager.GetRuntime")
        info = os.uname()
        arch = platform.architecture()
        ver = f"{sys.version_info.major}.{sys.version_info.minor}"
        variant = f"{info.sysname}_{'_'.join(arch)}_{info.version}_PY_{ver}"

        table = dict(self.SUCCESS_RETURN)
        table["toolsVer"] = "1"
        table["sdkVer"] = "1"
        table["sdkType"] = "PYTHON"
        table["sdkVariant"] = variant
        return table

    def check_memory(self):
        logger.debug("PubSubServer manager.GetRuntime")
        table = dict(self.SUCCESS_RETURN)
        table["errorInfo"] = "Nothing to check currently."
        return table

    def get_runtime(self):
        logger.debug("PubSubServer manager.GetRuntime")
        return self.SUCCESS_RETURN

    def update_log_level(self, level: str):
        table = dict(self.SUCCESS_RETURN)
        if level.lower() == "debug":
            logger.setLevel(level)
        elif level.lower() == "info":
            logger.setLevel(level)
        elif level.lower() == "notice":
            logger.setLevel(level)
        elif level.lower() == "warn":
            logger.setLevel(level)
        elif level.lower() == "error":
            logger.setLevel(level)
        else:
            logger.error(f"Level {level} not understood")
            table["rc"] = 1
            table["errorInfo"] = f"Level not understood: {level}"
        try:
            set_python_solace_log_level(level)
        except Exception as e:
            logger.error(f"Could not update API logger")
            table["rc"] = 1
            table["errorInfo"] = f"Could not update API loggers to level: {level} {str(e)}"

        return table

    # noinspection PyTypeChecker
    def start_server(self, ip, port):
        self._ip = ip
        self._port = port
        # Create server
        self._server = SimpleXMLRPCServer((self._ip, self._port), requestHandler=RequestHandler, logRequests=False)
        self._server.timeout = 1
        self._server.register_introspection_functions()
        # Manager Methods
        self._server.register_function(self.create_client, "manager.CreateClient")
        self._server.register_function(self.delete_client, "manager.DeleteClient")
        self._server.register_function(self.get_clients, "manager.GetClients")
        self._server.register_function(self.insert_log, "manager.InsertLog")
        self._server.register_function(self.get_version_info, "manager.GetVersionInfo")
        self._server.register_function(self.check_memory, "manager.CheckMemory")
        self._server.register_function(self.get_runtime, "manager.GetRuntime")
        self._server.register_function(self.update_log_level, "manager.UpdateLogLevel")
        # System Methods
        self._server.register_function(self.stop_server, "system.shutdown")
        # Client Methods (Dispatch)
        self._server.register_instance(self, allow_dotted_names=True)

        # Run the server until something is put on the queue
        while self._shutdown_queue.qsize() < 1:
            self._server.handle_request()
        self._server.server_close()

    def stop_server(self, comment):
        logger.warning(f"Termination requested.  Reason: {comment}")
        self._shutdown_queue.put_nowait("SOMETHING")
        return True


def run():
    parser = argparse.ArgumentParser(prog="PyPubSubTools", description='Run PubsSubTools server for python API')
    parser.add_argument('--host', type=str,
                        help='Host address to run XMLRPC server from')
    parser.add_argument('--port', type=int,
                        help='Port to start XMLRPC server on')
    parser.add_argument('--log', type=str, default='WARN',
                        help='Log level to be sent to terminal')

    args = parser.parse_args()
    if not (args.host or args.port):
        parser.error("Host and Port arguments are required")

    init_logging(args.log, c_log_level=args.log)

    try:
        tool_version = pkg_resources.require("sdkperf_py")[0].version
        api_version = pkg_resources.require("solace-pubsubplus")[0].version
        otel_version = pkg_resources.require("pubsubplus-opentelemetry-integration")[0].version
        print("Manager: Version info: ")
        print("PubSubTools version: " + tool_version)
        print("API version: " + api_version)
        print(f"OTEL Version: {otel_version}")
    except Exception as e:
        print(f"Exception printing version info {e}")

    s = PubSubServer()
    s.start_server(args.host, args.port)


if __name__ == "__main__":
    run()
