from abc import ABC, abstractmethod
from typing import Union

from solace.messaging.publisher.outbound_message import OutboundMessage
from solace.messaging.utils.life_cycle_control import TerminationNotificationListener, LifecycleControl, \
    TerminationEvent
from solace.messaging.utils.manageable import Metric

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.channel_state import ChannelState
from sdkperf.config.ep_config_properties import EpConfigProperties
from sdkperf.config.flow_status import FlowStatus
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.config.message_object_type import MessageObjectType
from sdkperf.core.last_message_details import LastMessageDetails
from sdkperf.core.msg_rep import MessageRep
from sdkperf.core.perf_stats import PerfStats

logger = get_logger()


class AbstractClient(ABC):
    def __init__(self):
        self.channel_state = ChannelState.CLIENT_STATE_UNKNOWN
        self.service = None

    # noinspection PyAttributeOutsideInit
    def init(self, props: RuntimeProperties, client_id: int) -> None:
        self._rx_props = RuntimeProperties()
        self._rx_props.__dict__.update(props.__dict__)
        self._tx_props = RuntimeProperties()

        from sdkperf.core.client_factory import ClientFactory
        self._client_id_str = ClientFactory.generate_client_id_str(props, client_id)
        self._client_id = client_id

        self._num_pubs_per_session = getattr(props, RuntimeProperties.NUM_PUBS_PER_SESSION, 0)
        self._termination_listener = TerminationListener(self)
        if self._num_pubs_per_session < 1:
            self._num_pubs_per_session = 1

        self.stats = PerfStats(self._client_id_str,
                               getattr(self._rx_props, RuntimeProperties.LATENCY_BUCKETS, 0),
                               getattr(self._rx_props, RuntimeProperties.LATENCY_GRANULARITY, 0),
                               getattr(self._rx_props, RuntimeProperties.LATENCY_WARMUP_IN_SECS, 0))
        self._msg_count = 0
        self.stop_on_error = getattr(self._rx_props, RuntimeProperties.WANT_STOP_ON_ERROR, False)
        self.last_message_details = LastMessageDetails()
        self.last_messages_details_size = getattr(self._rx_props, RuntimeProperties.LAST_MESSAGE_QUEUE_DEPTH, 0)
        self.last_messages_details = list()

    def destroy(self):
        pass

    def get_id_str(self) -> str:
        return self._client_id_str

    def get_props(self) -> RuntimeProperties:
        return self._rx_props

    def set_collection(self, cc: 'AbstractClientCollection'):
        self._client_collection = cc

    def get_owner(self) -> 'AbstractClientCollection':
        return self._client_collection

    def get_last_message_details(self) -> LastMessageDetails:
        return self.last_message_details

    def get_last_messages_details(self, num):
        if num <= 1:
            return [self.last_message_details]

        msgs_to_pull = min(num, len(self.last_messages_details))
        details = list()
        offset = len(self.last_messages_details) - msgs_to_pull if msgs_to_pull < len(self.last_messages_details) else 0
        for msg in self.last_messages_details[offset:]:
            details.append(msg)
        return details

    def get_num_termination_registered(self):
        return self._termination_listener.get_num_registered()

    @abstractmethod
    def get_last_acked_msg_id(self) -> int:
        pass

    @abstractmethod
    def set_last_acked_msg_id(self, msg_id: int):
        pass

    @abstractmethod
    def update_auth_on_reconnect(self, access_token: str, id_token: str):
        pass

    @abstractmethod
    def get_last_exception(self) -> Exception:
        pass

    @abstractmethod
    def set_last_exception(self, exception: Exception):
        pass

    @abstractmethod
    def clear_last_exception(self):
        pass

    @abstractmethod
    def get_rx_stats(self) -> PerfStats:
        pass

    @abstractmethod
    def get_metric(self, metric: Metric) -> Union[int, None]:
        pass

    @abstractmethod
    def clear_metrics(self):
        pass

    @abstractmethod
    def connect(self):
        pass

    def is_connected(self) -> bool:
        return self.channel_state == ChannelState.CLIENT_STATE_CONNECTED or \
               self.channel_state == ChannelState.CLIENT_STATE_RECONNECTING

    @abstractmethod
    def disconnect(self, grace_period: int):
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def open_transacted_session(self, transaction_props: RuntimeProperties):
        pass

    @abstractmethod
    def close_transacted_session(self, session_name: str):
        pass

    @abstractmethod
    def close_all_transacted_sessions(self):
        pass

    @abstractmethod
    def commit_transaction(self, session_name: str, want_rollback: bool):
        pass

    @abstractmethod
    def commit_transaction_on_current_pub(self, session_name: str, want_rollback: bool):
        pass

    @abstractmethod
    def get_flow_status(self, ep_name: str) -> FlowStatus:
        pass

    @abstractmethod
    def is_ready_to_publish(self, delivery_mode: GenericMessageDeliveryMode) -> bool:
        pass

    @abstractmethod
    def subscription_update(self, ep_props: EpConfigProperties):
        pass

    @abstractmethod
    def queue_update(self, ep_props: EpConfigProperties):
        pass

    @abstractmethod
    def temp_endpoint_update(self, ep_props: EpConfigProperties):
        pass

    @abstractmethod
    def request_reply(self, topic: str, request: str):
        pass

    @abstractmethod
    def map_topics(self, queue: str, topics: list, is_adding: bool):
        pass

    def endpoint_provisioning(self, ep_props: EpConfigProperties):
        raise Exception("ENDPOINT PROVISIONING NOT SUPPORTED IN PYTHON AFW")

    @abstractmethod
    def publish_msg(self, message_rep: Union[MessageRep,bytearray], destination: str,
                    pub_session_index: int, additional_properties: dict = None,
                    mode: GenericMessageDeliveryMode = GenericMessageDeliveryMode.DIRECT,
                    msg_obj_type: MessageObjectType = MessageObjectType.NONE,
                    want_await: bool = False, want_request_reply: bool = False, transacted_session_name: str = ""):
        pass

    @abstractmethod
    def publish_send_multiple(self):
        pass

    @abstractmethod
    def cache_request(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool) -> None:
        pass

    @abstractmethod
    def get_client_name(self) -> str:
        pass


class TerminationListener(TerminationNotificationListener):
    def __init__(self, client: AbstractClient):
        self._num_registered = 0
        self._client = client

    def get_num_registered(self):
        return self._num_registered

    def register_for_listen(self, obj_to_register: LifecycleControl):
        if getattr(self._client.get_props(), RuntimeProperties.WANT_TERMINATION_LISTENERS, True):
            obj_to_register.set_termination_notification_listener(self)
            self._num_registered += 1

    def on_termination(self, event: TerminationEvent):
        logger.error(f"on_termination() called: {event.message}")
        self._client.stats.inc_stat("numTerminationEvent")

