import time
from abc import ABC, abstractmethod

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.channel_state import ChannelState
from sdkperf.config.constants import Constants
from sdkperf.config.ep_config_properties import EpConfigProperties
from sdkperf.config.subscriber_destinations_type import SubscriberDestinationsType
from sdkperf.core.client_factory import ClientFactory
from sdkperf.core.perf_stats import PerfStats

logger = get_logger()


class AbstractClientCollection(ABC):
    def __init__(self, props: RuntimeProperties):
        self.init_abstract_client_collection(props)

    # noinspection PyAttributeOutsideInit
    def init_abstract_client_collection(self, props):
        self._rprops = RuntimeProperties()
        self._rprops.__dict__.update(props.__dict__)
        self._num_clients = getattr(self._rprops, RuntimeProperties.NUM_CLIENTS, 1)
        self._num_contexts = getattr(self._rprops, RuntimeProperties.NUM_CONTEXT_THREADS, 1)

        if self._num_contexts > self._num_clients:
            self._num_contexts = self._num_clients

        self._clients = list()
        self._client_id = 0
        for i in range(int(self._num_clients)):
            self._client_id += 1
            logger.debug(f"Creating client with id: {self._client_id}")
            self._clients.append(ClientFactory.create_client(self._rprops, self._client_id, self))

    def destroy(self):
        try:
            self.stop_publishing()
        except Exception as e:
            logger.error(f"Exception in destroy: {str(e)}")

        try:
            if self.is_connected():
                self.disconnect()
        except Exception as e:
            logger.error(f"Exception in destroy: {str(e)}")

        for client in self._clients:
            try:
                client.destroy()
            finally:
                del client
                client = None
        self._clients = list()

    def update_auth_on_reconnect(self, client_index: int, access_token: str, id_token: str):
        if client_index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.update_auth_on_reconnect(access_token, id_token)
        else:
            return self._clients[client_index].update_auth_on_reconnect(access_token, id_token)

    def get_last_exception(self, index):
        if index == Constants.ALL_CLIENT_INDEX:
            if len(self._clients) > 1:
                raise Exception("PerfClientCollection: Cannot request last error on all clients "
                                "if there is more than 1 client in object")
            return self._clients[0].get_last_exception()
        else:
            return self._clients[index].get_last_exception()

    def clear_last_exception(self, index):
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.clear_last_exception()
        else:
            return self._clients[index].clear_last_exception()

    def get_last_acked_message_id(self, index):
        ret = list()
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                ret.append(client.get_last_acked_msg_id())
        else:
            ret.append(self.get_client(index).get_last_acked_msg_id())
        return ret

    def get_num_termination_registered(self, index):
        ret = 0
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                ret += client.get_num_termination_registered()
        else:
            ret += self.get_client(index).get_num_termination_registered()
        return ret

    def get_client_list(self):
        return self._clients

    def get_client(self, index: int):
        return self._clients[index]

    def get_last_message_details(self, client_index):
        if client_index == Constants.ALL_CLIENT_INDEX:
            raise Exception("Cannot get Last Message Details for all clients, must specify specific client.")
        else:
            return self._clients[client_index].get_last_message_details()

    def get_last_messages_details(self, client_index, msgs_to_get):
        if client_index == Constants.ALL_CLIENT_INDEX:
            raise Exception("Cannot get Last Messages Details for all clients, must specify specific client.")
        else:
            return self._clients[client_index].get_last_messages_details(msgs_to_get)

    @abstractmethod
    def get_sdk_stat(self, metric, index):
        pass

    @abstractmethod
    def clear_stats(self, index):
        pass

    @abstractmethod
    def publish_message(self, pubprops: RuntimeProperties, index):
        pass

    @abstractmethod
    def start_publishing(self, pubprops: RuntimeProperties):
        pass

    @abstractmethod
    def stop_publishing(self):
        pass

    @abstractmethod
    def is_publishing(self):
        pass

    @abstractmethod
    def start_cache_requesting(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool,
                               num_msgs: int, cache_rate: int) -> None:
        pass

    @abstractmethod
    def stop_cache_requesting(self) -> None:
        pass

    @abstractmethod
    def is_cache_requesting(self) -> bool:
        pass

    @abstractmethod
    def cache_request(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool,
                      client_index: int) -> None:
        pass

    def get_client_name(self, index: int):
        ret = list()
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                ret.append(client.get_client_name())
        else:
            ret.append(self._clients[index].get_client_name())
        return ret

    def get_channel_state(self):
        ret_state = ChannelState.CLIENT_STATE_CONNECTED
        for client in self._clients:
            client_state = client.channel_state
            if client_state == ChannelState.CLIENT_STATE_CONNECTING or \
                    client_state == ChannelState.CLIENT_STATE_RECONNECTING:
                ret_state = client_state
                continue
            if client_state == ChannelState.CLIENT_STATE_DISCONNECTED:
                ret_state = client_state
                break

        return ret_state

    def set_client_name(self, index: int):
        pass

    def is_connected(self):
        for client in self._clients:
            if not client.is_connected():
                return False
        return True

    def connect(self):
        for client in self._clients:
            logger.debug(f"Connecting client {client.get_client_name()}")
            client.connect()

        guard = 1200
        while not self.is_connected() and guard > 0:
            guard -= 1
            time.sleep(0.1)

        if not self.is_connected():
            logger.error("Unable to start all clients")
            raise Exception("Unable to start all clients")

    def disconnect(self, grace_period=0):
        disconnected = False

        for client in self._clients:
            try:
                client.disconnect(grace_period)
            except Exception:
                pass

        guard = 1000
        while self.is_connected() and guard > 0:
            guard -= 1
            time.sleep(0.2)

        if self.is_connected():
            logger.error("Unable to disconnect all clients")
            raise Exception("Unable to disconnect all clients")

    def start(self):
        if not self.is_connected():
            logger.info("Called start, clients are not connected. Returning.")
            return

        for client in self._clients:
            client.start()

    def stop(self):
        for client in self._clients:
            client.stop()

    def open_transacted_session(self, client_index: int, transaction_props: RuntimeProperties):
        if not self.is_connected():
            logger.info("Called OpenTransactedSession, clients are not connected. Returning.")
            return ""
        if client_index < 0 or client_index >= len(self._clients):
            raise Exception(f"OpenTransactedSession: Client Index out of range. Range (0-{len(self._clients) - 1}), "
                            f"Index used: {client_index}")

        session_name = ""
        try:
            session_name = self._clients[client_index].open_transacted_session(transaction_props)
        except Exception as e:
            logger.warning(e)
            raise Exception("Unable to OpenTransactedSession", e)

        return session_name

    def close_transacted_session(self,  client_index: int, session_name: str):
        if client_index < 0 or client_index >= len(self._clients):
            raise Exception(f"CloseTransactedSession: Client Index out of range. Range (0-{len(self._clients) - 1}), "
                            f"Index used: {client_index}")

        try:
            self._clients[client_index].close_transacted_session(session_name)
        except Exception as e:
            logger.warning(e)
            raise Exception("Unable to CloseTransactedSession", e)

    def close_all_transacted_sessions(self, client_index: int):
        if client_index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.close_all_transacted_sessions()
        else:
            self._clients[client_index].close_all_transacted_sessions()

    def commit_transaction(self, client_index: int, session_name: str, want_rollback: bool):
        if client_index < 0 or client_index >= len(self._clients):
            raise Exception(f"CommitTransaction: Client Index out of range. Range (0-{len(self._clients) - 1}), "
                            f"Index used: {client_index}")

        try:
            self._clients[client_index].commit_transaction(session_name, want_rollback)
        except Exception as e:
            logger.warning(e)
            raise Exception("Unable to CommitTransaction", e)

    def get_flow_status(self, ep_name: str, client_idx: int):
        if client_idx == Constants.ALL_CLIENT_INDEX:
            raise Exception(
                "AbstractClientCollection: getFlowStatus only available on specific client index. Not ALL_CLIENT_INDEX")
        else:
            return self._clients[client_idx].get_flow_status(ep_name)

    def subscription_update(self, epProps: EpConfigProperties, clientidx: int):
        if clientidx == Constants.ALL_CLIENT_INDEX:
            if epProps.distribute_topics:
                topic_list = epProps.subscriptions
                client_topic_list = list()
                topic_index = 0
                for topic in topic_list:
                    if topic_list < len(self._clients):
                        topics = list()
                        topics.append(topic)
                        client_topic_list.append(topics)
                    else:
                        client_topic_list[topic_index % len(self._clients)].append(topic)
                    topic_index += 1

                client_index = 0
                for client in self._clients:
                    epProps.subscriptions = client_topic_list[client_index % len(client_topic_list)]
                    client.subscription_update(epProps)
                    client_index += 1
            else:
                for client in self._clients:
                    client.subscription_update(epProps)
        else:
            self._clients[clientidx].subscription_update(epProps)

    def temp_endpoint_update(self, ep_props: EpConfigProperties, client_idx: int):
        ret_list = []
        if client_idx == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                temp_list = client.temp_endpoint_update(ep_props)
                ret_list.extend(temp_list)
        else:
            ret_list = self._clients[client_idx].temp_endpoint_update(ep_props)
        if ep_props.type != SubscriberDestinationsType.TOPIC and \
                ep_props.subscriptions is not None and \
                len(ep_props.subscriptions) > 0:
            for endpoint in ret_list:
                self.map_topics(endpoint, ep_props.subscriptions, True, client_idx)
        return ret_list

    def unbind_all_temp_endpoints(self, clientidx: int):
        if clientidx == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.unbind_all_temp_endpoints()
        else:
            self._clients[clientidx].unbind_all_temp_endpoints()

    def queue_update(self, epProps: EpConfigProperties, clientidx: int):
        if clientidx == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.queue_update(epProps)
        else:
            self._clients[clientidx].queue_update(epProps)

    def endpoint_provision(self, ep_props: EpConfigProperties, client_idx: int):
        # provisioning is done in queue_update for python, calling endpoint_provision directs the call to queue_update
        self.queue_update(ep_props, client_idx)


    def map_topics(self, queue: str, topics: list, is_adding: bool, client_index: int):
        if client_index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.map_topics(queue, topics, is_adding)
        else:
            self._clients[client_index].map_topics(queue, topics, is_adding)

    def request_reply(self, topic: str, request: str, client_index: int):
        return self._clients[client_index].request_reply(topic, request)

    def get_rx_stats(self, client_index):
        if len(self._clients) == 1:
            return self._clients[0].get_rx_stats()
        elif client_index == Constants.ALL_CLIENT_INDEX:
            ret = PerfStats("aggregate",
                            getattr(self._rprops, RuntimeProperties.LATENCY_BUCKETS, 0),
                            getattr(self._rprops, RuntimeProperties.LATENCY_GRANULARITY, 0),
                            getattr(self._rprops, RuntimeProperties.LATENCY_WARMUP_IN_SECS, 0))
            for client in self._clients:
                ret.aggregate(client.get_rx_stats())
            return ret
        else:
            return self._clients[client_index].get_rx_stats()
