import time
from threading import Thread

from pypubsubtools.utils.log import get_logger
from sdkperf.core.abstract_client import AbstractClient

logger = get_logger()


class CacheThread:
    def __init__(self, client: AbstractClient):
        self.client = client
        self._cache_thread = None
        self.start_time = 0
        self.end_time = 0

    def is_done_cache_requesting(self):
        if self._cache_thread is None:
            return False
        return self._cache_thread.is_done_cache_requesting

    def is_cache_requesting(self):
        if self._cache_thread is None:
            return False
        return self._cache_thread.is_cache_requesting

    def stop_cache_requesting(self):
        if self._cache_thread is not None:
            self._cache_thread.shutdown = True
            self._cache_thread.join(1)

    def start_cache_requesting(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool,
                               num_msgs: int, cache_rate: int, want_stop_on_error: bool = False):
        if self._cache_thread is not None:
            if self._cache_thread.is_alive():
                raise Exception("Attempted to start a cache thread when one was already running.")

        self._cache_thread = CacheThreadObj(self, topics, want_subscribe, live_data_action, wait_for_confirm, num_msgs,
                                            cache_rate, want_stop_on_error)
        self._cache_thread.daemon = True
        self._cache_thread.start()


class CacheThreadObj(Thread):
    def __init__(self, parent: CacheThread, topics: list, want_subscribe: bool, live_data_action: str,
                 wait_for_confirm: bool, num_msgs: int, cache_rate: int, want_stop_on_error: bool):
        super().__init__()
        self.parent = parent
        self.topics = topics
        self.want_subscribe = want_subscribe
        self.live_data_action = live_data_action
        self.wait_for_confirm = wait_for_confirm
        self.num_cache_requests = num_msgs
        self.cache_rate = cache_rate
        self.want_stop_on_error = want_stop_on_error

        self.shutdown = False
        self.is_cache_requesting = False
        self.is_done_cache_requesting = False

    def run(self):
        cache_requests_sent = 0
        rate_check_threshold = 1000
        single_cache_period_in_ns = 0
        if self.cache_rate > 0:
            single_cache_period_in_ns = 1000000000 / self.cache_rate
        self.shutdown = False
        self.is_done_cache_requesting = False

        start_time_us = time.time_ns() / 1000
        cache_start_time_in_ns = start_time_us * 1000
        self.is_cache_requesting = True

        self.parent.start_time = cache_start_time_in_ns
        while cache_requests_sent < self.num_cache_requests and not self.shutdown:
            if single_cache_period_in_ns > 0:
                if (self.cache_rate > rate_check_threshold and cache_requests_sent % rate_check_threshold == 0) \
                        or self.cache_rate <= rate_check_threshold:
                    calc_time_since_start_in_ns = single_cache_period_in_ns * cache_requests_sent
                    next_time_in_ns = cache_start_time_in_ns + calc_time_since_start_in_ns

                    time_waited_in_ns = next_time_in_ns - time.time_ns()
                    if time_waited_in_ns > 0:
                        time.sleep(time_waited_in_ns / 1000000000)
            try:
                temp_topic_list = [self.topics[cache_requests_sent % len(self.topics)]]
                self.parent.client.cache_request(temp_topic_list, self.want_subscribe, self.live_data_action,
                                                 self.wait_for_confirm)
            except Exception as e:
                logger.error(f"CLIENT {self.parent.client.get_id_str()}: cannot cache request due to {e}")
                if self.want_stop_on_error:
                    self.parent.stop_cache_requesting()
                else:
                    guard = 60
                    while not self.parent.client.is_connected() and guard != 0:
                        time.sleep(1)
                        guard -= 1
                    if guard == 0:
                        logger.error(f"CLIENT {self.parent.client.get_id_str()}: Channel failed to return to up state")
                        self.parent.stop_cache_requesting()
                    else:
                        logger.warn(f"CLIENT {self.parent.client.get_id_str()}: Cache requesting resuming following "
                                    f"reconnect")
            cache_requests_sent += 1

        self.parent.end_time = time.time_ns()
        self.is_cache_requesting = False
        self.is_done_cache_requesting = True

