import copy
import time
import uuid
from typing import Union

from opentelemetry.trace import SpanKind
from solace.messaging.config.solace_properties.message_properties import PRIORITY, SEQUENCE_NUMBER, \
    APPLICATION_MESSAGE_ID, APPLICATION_MESSAGE_TYPE, HTTP_CONTENT_TYPE, HTTP_CONTENT_ENCODING

from solace.messaging.config.message_acknowledgement_configuration import Outcome
from solace.messaging.receiver.inbound_message import InboundMessage, ReplicationGroupMessageId, CacheStatus
from solace.messaging.receiver.message_receiver import MessageHandler
from solace.messaging.receiver.persistent_message_receiver import PersistentMessageReceiver
from solace.messaging.receiver.request_reply_message_receiver import RequestMessageHandler, Replier
from solace.messaging.receiver.transactional_message_receiver import TransactionalMessageReceiver
from solace.messaging.resources.topic_subscription import TopicSubscription
from solace_otel.messaging.trace.propagation import InboundMessageCarrier, InboundMessageGetter

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.constants import Constants
from sdkperf.config.flow_status import FlowStatus
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.core.abstract_client import AbstractClient
from sdkperf.core.python_client_transacted_session import PythonClientTransactedSession
from sdkperf.core.python_wrapped_message import PythonWrappedMessage
from sdkperf.core.tool_info import ToolData
from sdkperf.core.tracing_utils import TracingUtils

logger = get_logger()


class MessageListener:
    def __init__(self, parent: AbstractClient, destination: Union[str, None], is_topic: bool,
                 topic: Union[TopicSubscription, None], is_non_durable: bool,
                 transacted_session: Union[PythonClientTransactedSession, None], want_msg_priority_order_checking: bool,
                 sleep_time: int = 0, flow_status: FlowStatus = None, settlement_pattern=None):
        self.parent = parent

        self.destination = destination
        self.is_topic = is_topic
        self.topic = topic
        self.is_non_durable = is_non_durable
        self.transacted_session = transacted_session
        self.want_msg_priority_order_checking = want_msg_priority_order_checking
        self.settlement_pattern_list = list()
        self.settlement_pattern_index = 0
        settlement_pattern_string = settlement_pattern if settlement_pattern is not None else getattr(self.parent.get_props(), RuntimeProperties.SETTLEMENT_PATTERN_STRING, None)
        if settlement_pattern_string is not None:
            for char in settlement_pattern_string:
                if char.lower() == 'a':
                    self.settlement_pattern_list.append(Outcome.ACCEPTED)
                elif char.lower() == 'r':
                    self.settlement_pattern_list.append(Outcome.REJECTED)
                elif char.lower() == 'f':
                    self.settlement_pattern_list.append(Outcome.FAILED)
                else:
                    logger.warn(f"Invalid character used in settlement pattern: {char}")
                    

        self._msg_dump = getattr(self.parent.get_props(), RuntimeProperties.WANT_MSG_DUMP, False)
        self._order_check = getattr(self.parent.get_props(), RuntimeProperties.WANT_ORDER_CHECK, False)
        self._want_cache_stats = getattr(self.parent.get_props(), RuntimeProperties.CACHE_WANT_STATS, False)
        self._crc_check = getattr(self.parent.get_props(), RuntimeProperties.WANT_PAYLOAD_CHECK, False)
        self._lat_check = getattr(self.parent.get_props(), RuntimeProperties.WANT_LATENCY, False)
        self._details = getattr(self.parent.get_props(), RuntimeProperties.WANT_PER_MESSAGE_DETAILS, False)
        self._user_prop_tool_data = getattr(self.parent.get_props(), RuntimeProperties.WANT_USER_PROP_TOOL_DATA, False)

        self.uid = uuid.uuid1()
        self.consumer = None
        self.is_started = False
        self.done_receive = False
        self._subscriptions = list()
        self.sleep_time = sleep_time
        if flow_status is None:
            self.flow_status = FlowStatus()
        else:
            self.flow_status = flow_status

    def get_flow_status(self):
        return self.flow_status

    def add_subscription(self, sub: str):
        self._subscriptions.append(sub)

    def remove_subscription(self, sub: str):
        try:
            idx = self._subscriptions.index(sub)
            self._subscriptions.pop(idx)
            return True
        except ValueError:
            return False

    def get_subscriptions(self):
        return list(self._subscriptions)

    def process_message(self, message):
        """ Message receive callback """
        time_recv = time.time() * Constants.SEC_TO_NANOSEC
        delivery_mode = GenericMessageDeliveryMode.DIRECT
        if isinstance(self.consumer, PersistentMessageReceiver):
            if len(self.settlement_pattern_list) > 0:
                self.consumer.settle(message, self.settlement_pattern_list[self.settlement_pattern_index%len(self.settlement_pattern_list)])
                self.settlement_pattern_index += 1
            else:
                self.consumer.ack(message)
            self.parent.set_last_acked_msg_id(message.get_sequence_number())
            delivery_mode = GenericMessageDeliveryMode.PERSISTENT

        if isinstance(self.consumer, TransactionalMessageReceiver):
            self.parent.set_last_acked_msg_id(message.get_sequence_number())
            delivery_mode = GenericMessageDeliveryMode.PERSISTENT

        if getattr(self.parent.get_props(), RuntimeProperties.WANT_TRACE_COPY_RECEIVE_CONTEXT, False) \
                and TracingUtils.is_init:
            receiver_context_carrier = InboundMessageCarrier(message)
            extracted_ctx = TracingUtils.text_map.extract(carrier=receiver_context_carrier, getter=InboundMessageGetter())
            with TracingUtils.tracer.start_as_current_span(f"{message.get_destination_name()} sdkperf_py_process",
                                                           kind=SpanKind.CONSUMER,
                                                           context=extracted_ctx) as receiver_context_span:
                # Nothing to do in this span, we just want to start and end it
                pass

        if self._details:
            try:
                self._set_last_message(PythonWrappedMessage(message, delivery_mode))
            except Exception as e:
                logger.warn(f"Exception setting last message: {e}")

        topic = message.get_destination_name()
        try:
            payload_as_bytes = message.get_payload_as_bytes()
            payload_as_string = message.get_payload_as_string()
        except Exception:
            payload_as_bytes = None
            payload_as_string = None
        correlation_id = message.get_correlation_id()

        if getattr(self.parent.get_props(), RuntimeProperties.WANT_STRUCT_MSG_CHECK, False):
            check = False
            for prop in self.parent.stats.sdm_props:
                if prop.check_message(message):
                    check = True
                    break
            if not check:
                logger.warn(f"Message failed struct check: {message}")
            self.parent.stats.inc_struct_check(check)

        if getattr(self.parent.get_props(), RuntimeProperties.WANT_MSG_OBJ_TYPE_CHECK, False):
            got_correct = False
            msg_type = message.get_application_message_type()

            if msg_type is not None:
                msg_type = msg_type.lower()
                if msg_type == 'b':
                    message_content = message.get_payload_as_bytes()
                    if message_content == bytearray([0x31, 0x32, 0x33]):
                        got_correct = True
                elif msg_type == 't':
                    message_content = message.get_payload_as_string()
                    if message_content == "TEXT_MESSAGE":
                        got_correct = True
                elif msg_type == 's':
                    message_content = message.get_payload_as_list()
                    if message_content[0] == "STREAM_MESSAGE":
                        got_correct = True
                elif msg_type == 'm':
                    message_content = message.get_payload_as_dictionary()
                    if message_content["MAP"] == "MESSAGE":
                        got_correct = True

                if got_correct:
                    self.parent.stats.inc_stat("numMsgsObjTypeOk")
                else:
                    self.parent.stats.inc_stat("numMsgsObjTypeFail")

        if payload_as_bytes is not None:
            self.parent.stats.update_rx_bytes(len(payload_as_bytes))
        else:
            self.parent.stats.update_rx_bytes(0)

        tool_data = ToolData()
        if payload_as_bytes is not None and not self._user_prop_tool_data and tool_data.decode(payload_as_bytes):
            if self._order_check and tool_data.has_message_identifier():
                self.parent.stats.inc_stat("numOrderCheckedMsgs")
                redelivered = message.is_redelivered()
                if redelivered:
                    self.parent.stats.inc_stat('numMsgRedelivered')
                self.parent.stats.check_order(tool_data.get_stream_id(), tool_data.get_message_identifier(),
                                              tool_data.get_republished(), redelivered)
            if self._crc_check and tool_data.has_bin_attach_hash() and payload_as_bytes is not None:
                self.parent.stats.check_crc(tool_data.get_bin_attach_hash(), payload_as_bytes[tool_data.get_encoded_size():])
            if self._lat_check and tool_data.has_latency():
                self.parent.stats.inc_latency(time_recv, time_recv - float(tool_data.get_latency()))
            else:
                self.parent.stats.update_time_recvd(time_recv)
        else:
            if self._order_check and message.has_property(ToolData.PROP_MESSAGE_ID):
                self.parent.stats.inc_stat("numOrderCheckedMsgs")
                redelivered = message.is_redelivered()
                if redelivered:
                    self.parent.stats.inc_stat('numMsgRedelivered')
                stream_id = message.get_property(ToolData.PROP_STREAM_ID)
                msg_id = message.get_property(ToolData.PROP_MESSAGE_ID)
                republished = message.get_property(ToolData.PROP_REPUBLISHED)
                self.parent.stats.check_order(stream_id if stream_id is not None else -1,
                                              msg_id if msg_id is not None else -1,
                                              republished if republished is not None else False,
                                              redelivered)
            if self._crc_check and message.has_property(ToolData.PROP_BIN_PAYLOAD_HASH) and \
                    payload_as_bytes is not None:
                self.parent.stats.check_crc(message.get_property(ToolData.PROP_BIN_PAYLOAD_HASH), payload_as_bytes)
            if self._lat_check and message.has_property(ToolData.PROP_LATENCY):
                self.parent.stats.inc_latency(time_recv,
                                              int(time_recv - float(message.get_property(ToolData.PROP_LATENCY))))
            else:
                self.parent.stats.update_time_recvd(time_recv)

        if message.get_priority() is not None and getattr(
                self.parent.get_props(), RuntimeProperties.WANT_MESSAGE_PRIORITY_STATS, False):
            self.parent.stats.message_priority_stats_inc(message.get_priority(), True)
        
        if getattr(self.parent.get_props(), RuntimeProperties.WANT_DI_STATS, False):
            if message.get_message_discard_notification() is not None and \
                    message.get_message_discard_notification().has_broker_discard_indication():
                self.parent.stats.inc_stat('numMsgDISet')
            if message.get_message_discard_notification() is not None and \
                    message.get_message_discard_notification().has_internal_discard_indication():
                self.parent.stats.inc_stat('numMsgBpDiSet')

        if self._want_cache_stats:
            cache_status = message.get_cache_status()
            if cache_status == CacheStatus.CACHED:
                self.parent.stats.inc_stat("numCacheMsgsRecv")
            elif cache_status == CacheStatus.SUSPECT:
                # A suspect messages is still a cached message, increment both stats
                self.parent.stats.inc_stat("numCacheMsgsRecv")
                self.parent.stats.inc_stat("numSuspectRecv")
            elif cache_status == CacheStatus.LIVE:
                self.parent.stats.inc_stat("numLiveMsgsRecv")
            # We don't do anything with this at the moment but call it just to see if anything breaks
            message.get_cache_request_id()

        try:
            if self._order_check and \
                    message.get_replication_group_message_id() is not None and \
                    not message.is_redelivered():
                replication_group_msg_id: ReplicationGroupMessageId = message.get_replication_group_message_id()
                id_index: int = -1
                inc_ooo: bool = False
                inc_suid_change: bool = True

                if self.parent.stats.get_replication_group_message_ooo_detection_list(self) is not None:
                    for rgmid in self.parent.stats.get_replication_group_message_ooo_detection_list(self):
                        id_index += 1
                        try:
                            if rgmid.compare(replication_group_msg_id) > 0:
                                inc_ooo = True
                            inc_suid_change = False
                            break
                        except Exception:
                            pass

                self.parent.stats.update_replication_group_message_ooo_detection(
                    self,
                    replication_group_msg_id,
                    id_index,
                    inc_ooo,
                    inc_suid_change
                )
        except Exception:
            pass

        if self._msg_dump:
            print("^^^^^^^^^^^^^^^^^^ Start Message ^^^^^^^^^^^^^^^^^^^^^^^^^^^")
            print(message)
            print("^^^^^^^^^^^^^^^^^^ End Message ^^^^^^^^^^^^^^^^^^^^^^^^^^^")

        if self.sleep_time > 0:
            time.sleep(self.sleep_time)

        if self.transacted_session is not None:
            self.transacted_session.on_message()

    def _set_last_message(self, message):
        self.parent.last_message_details.clear()

        inbound_msg = message.get_message()

        self.parent.last_message_details.set_destination(message.get_destination_name())
        self.parent.last_message_details.set_destination_type(None)
        self.parent.last_message_details.set_http_content_type(
            inbound_msg.get_rest_interoperability_support().get_http_content_type())
        self.parent.last_message_details.set_http_content_encoding(
            inbound_msg.get_rest_interoperability_support().get_http_content_encoding())
        self.parent.last_message_details.set_correlation_id(inbound_msg.get_correlation_id())
        self.parent.last_message_details.set_expiration(inbound_msg.get_expiration())
        self.parent.last_message_details.set_priority(inbound_msg.get_priority())
        self.parent.last_message_details.set_application_message_id(inbound_msg.get_application_message_id())
        self.parent.last_message_details.set_application_message_type(inbound_msg.get_application_message_type())
        self.parent.last_message_details.set_sender_timestamp(inbound_msg.get_sender_timestamp())
        self.parent.last_message_details.set_redelivered(inbound_msg.is_redelivered())
        self.parent.last_message_details.set_cos(inbound_msg.get_class_of_service())

        try:
            creation_context = inbound_msg.get_creation_trace_context()
            self.parent.last_message_details.set_creation_trace_id(creation_context[0].hex() if creation_context[0] is not None else None)
            self.parent.last_message_details.set_creation_span_id(creation_context[1].hex() if creation_context[1] is not None else None)
            self.parent.last_message_details.set_creation_sampled(creation_context[2])
            self.parent.last_message_details.set_creation_trace_state(creation_context[3])
        except Exception:
            pass

        try:
            transport_context = inbound_msg.get_transport_trace_context()
            self.parent.last_message_details.set_transport_trace_id(transport_context[0].hex() if transport_context[0] is not None else None)
            self.parent.last_message_details.set_transport_span_id(transport_context[1].hex() if transport_context[1] is not None else None)
            self.parent.last_message_details.set_transport_sampled(transport_context[2])
            self.parent.last_message_details.set_transport_trace_state(transport_context[3])
        except Exception:
            pass

        try:
            baggage = inbound_msg.get_baggage()
            self.parent.last_message_details.set_baggage(baggage)
        except Exception:
            pass

        try:
            binary_payload_size = len(inbound_msg.get_payload_as_bytes())
            self.parent.last_message_details.set_binary_payload_size(binary_payload_size)
        except Exception:
            pass

        if self.parent.last_messages_details_size > 0:
            self.parent.last_messages_details.append(copy.deepcopy(self.parent.last_message_details))
            if len(self.parent.last_messages_details) > self.parent.last_messages_details_size:
                self.parent.last_messages_details = self.parent.last_messages_details[1:]


class MessageHandlerImpl(MessageHandler, MessageListener):
    def __init__(self, parent: AbstractClient, destination: Union[str, None], is_topic: bool,
                 topic: Union[TopicSubscription, None], is_non_durable: bool,
                 transacted_session: Union[PythonClientTransactedSession, None], want_msg_priority_order_checking: bool,
                 sleep_time: int = 0, flow_status: FlowStatus = None, settlement_pattern=None):
        super().__init__(parent, destination, is_topic, topic, is_non_durable, transacted_session,
                         want_msg_priority_order_checking, sleep_time, flow_status, settlement_pattern)

    def on_message(self, message: InboundMessage):
        self.process_message(message)


class RequestMessageHandlerImpl(RequestMessageHandler, MessageListener):
    def __init__(self, parent: AbstractClient, destination: Union[str, None], is_topic: bool,
                 topic: Union[TopicSubscription, None], is_non_durable: bool, want_msg_priority_order_checking: bool,
                 sleep_time: int = 0, flow_status: FlowStatus = None):
        super().__init__(parent, destination, is_topic, topic, is_non_durable, None, want_msg_priority_order_checking,
                         sleep_time, flow_status)

    def on_message(self, message: InboundMessage, replier: Replier):
        self.process_message(message)

        if replier is not None:
            props = message.get_properties()
            if props is None:
                props = {}
            if message.get_priority() is not None:
                props[PRIORITY] = message.get_priority()
            if message.get_sequence_number() is not None:
                props[SEQUENCE_NUMBER] = message.get_sequence_number()
            if message.get_application_message_id() is not None:
                props[APPLICATION_MESSAGE_ID] = message.get_application_message_id()
            if message.get_application_message_type() is not None:
                props[APPLICATION_MESSAGE_TYPE] = message.get_application_message_type()
            if message.get_rest_interoperability_support() is not None and \
                    message.get_rest_interoperability_support().get_http_content_type() is not None:
                props[HTTP_CONTENT_TYPE] = message.get_rest_interoperability_support().get_http_content_type()
            if message.get_rest_interoperability_support() is not None and \
                    message.get_rest_interoperability_support().get_http_content_encoding() is not None:
                props[HTTP_CONTENT_ENCODING] = message.get_rest_interoperability_support().get_http_content_encoding()
            outbound_message = self.parent.service.message_builder().build(message.get_payload_as_bytes(),
                                                                           props)
            replier.reply(outbound_message)
