import binascii
import math
from threading import Lock

from solace.messaging.receiver.inbound_message import ReplicationGroupMessageId

from pypubsubtools.utils.log import get_logger
from sdkperf.config.constants import Constants
from sdkperf.core.single_stream_order_checking_data import SingleStreamOrderCheckingData

logger = get_logger()
lock = Lock()


class PerfStats:
    """time.perf_counter_ns()"""

    def __init__(self, sub_name: str, latency_num_buckets: int, latency_granularity_factor: int,
                 latency_warm_up_secs: float):
        self._receive_start_time_in_nanos = 0
        self._receive_end_time_in_nanos = 0
        self._prev_msg_id = -1
        self._lat_buckets = [0] * latency_num_buckets
        self._lat_granularity = latency_granularity_factor
        self._lat_warmup_in_sec = latency_warm_up_secs
        self._ooo_data = dict()
        self._sub_name = sub_name
        self.stats = {
            "oooMsgs": 0,
            "numOrderCheckedMsgs": 0,
            "crcOk": 0,
            "crcFail": 0,
            "binCrcOk": 0,
            "binCrcFail": 0,
            "structCheckOk": 0,
            "structCheckFail": 0,
            "minLat": Constants.SEC_TO_NANOSEC,  # This is just a real big number that latency will never be (I hope)
            "maxLat": 0,
            "totalLat": 0,
            "latMsgs": 0,
            "rxMsgs": 0,
            "rxBytes": 0,
            'adPubSuccess': 0,
            'adPubFail': 0,
            'directPubFail': 0,
            'numAckEvents': 0,
            'numMsgDISet': 0,
            'numMsgBpDiSet': 0,
            'numMsgRedelivered': 0,
            'numReplicationGroupMessageIdDuplicates': 0,
            'numReplicationGroupMessageIdSpoolMessageIdChanges': 0,
            'numTerminationEvent': 0,
            'numCacheReqIncompNoData': 0,
            'numCacheReqOk': 0,
            'numCacheReqIncompSuspect': 0,
            'numCacheReqIncompTimeout': 0,
            'numCacheReqError': 0,
            'numCacheRespRecv': 0,
            'numLiveMsgsRecv': 0,
            'numCacheMsgsRecv': 0,
            'numSuspectRecv': 0,
            'numMsgsObjTypeOk': 0,
            'numMsgsObjTypeFail': 0,
        }
        self.rx_priority = dict()
        self.tx_priority = dict()
        self.sdm_props = None
        self._replication_group_message_ooo_detection = dict()

    def set_sdm_props(self, sdm_props: list):
        self.sdm_props = sdm_props

    def get_latency_buckets(self):
        return self._lat_buckets

    def aggregate(self, input_stats):
        if self._receive_start_time_in_nanos > input_stats._receive_start_time_in_nanos:
            self._receive_start_time_in_nanos = input_stats._receive_start_time_in_nanos
        if self._receive_end_time_in_nanos < input_stats._receive_end_time_in_nanos:
            self._receive_end_time_in_nanos = input_stats._receive_end_time_in_nanos

        for stat in input_stats.stats:
            if stat == "latMin":
                if input_stats.get_stat(stat) < self.get_stat(stat):
                    self.stats[stat] = input_stats.get_stat(stat)
            elif stat == "latMax":
                if input_stats.get_stat(stat) > self.get_stat(stat):
                    self.stats[stat] = input_stats.get_stat(stat)
            else:
                self.stats[stat] += input_stats.get_stat(stat)

        if len(self._lat_buckets) != len(input_stats._lat_buckets):
            logger.error("Aggregating perf stats whose latency buckets are not equal in length")

        for i in range(len(input_stats._lat_buckets)):
            self._lat_buckets[i] += input_stats._lat_buckets[i]

        for k, v in input_stats._ooo_data.items():
            if k in self._ooo_data:
                self._ooo_data[k].aggregate(v)
            else:
                self._ooo_data[k] = v

        for k, v in input_stats.rx_priority.items():
            try:
                self.rx_priority[k] += v
            except KeyError:
                self.rx_priority[k] = v
        for k, v in input_stats.tx_priority.items():
            try:
                self.tx_priority[k] += v
            except KeyError:
                self.tx_priority[k] = v

        return True

    def message_priority_stats_inc(self, priority: int, rx: bool = True):
        if priority is None:
            return
        try:
            if rx:
                stat = self.rx_priority[priority]
            else:
                stat = self.tx_priority[priority]
            stat += 1
        except KeyError:
            stat = 1

        if rx:
            self.rx_priority[priority] = stat
        else:
            self.tx_priority[priority] = stat

    def get_stat(self, stat_type):
        val = 0
        if stat_type == "latAvg":
            if self.stats["latMsgs"] != 0:
                val = self.stats["totalLat"] / self.stats["latMsgs"]
        elif stat_type == "minLat":
            if self.stats[stat_type] != Constants.SEC_TO_NANOSEC:
                val = self.stats[stat_type]
        else:
            if stat_type in self.stats:
                val = self.stats[stat_type]
        return val

    def reset_stats(self):
        self._receive_start_time_in_nanos = 0
        self._receive_end_time_in_nanos = 0
        self._prev_msg_id = -1
        self._ooo_data = dict()

        for i in range(len(self._lat_buckets)):
            self._lat_buckets[i] = 0

        for key in self.stats:
            if key == "minLat":
                self.stats["minLat"] = Constants.SEC_TO_NANOSEC
            else:
                self.stats[key] = 0

    def update_time_recvd(self, time_recvd: int):
        self._receive_end_time_in_nanos = time_recvd
        if self._receive_start_time_in_nanos == 0:
            self._receive_start_time_in_nanos = time_recvd

    def get_thru_put(self, stat):
        if self._receive_start_time_in_nanos >= self._receive_end_time_in_nanos:
            return 0
        else:
            return stat * Constants.SEC_TO_NANOSEC / (
                    self._receive_end_time_in_nanos - self._receive_start_time_in_nanos)

    def get_lat_thru_put(self, stat):
        start_time = self._receive_start_time_in_nanos + (self._lat_warmup_in_sec * 1000000000)
        if start_time >= self._receive_end_time_in_nanos:
            return 0
        else:
            return stat * Constants.SEC_TO_NANOSEC / (self._receive_end_time_in_nanos - start_time)

    def update_rx_bytes(self, size):
        self.stats["rxMsgs"] += 1
        self.stats["rxBytes"] += size

    def check_order(self, stream_id, msg_id, republished, redelivered):
        with lock:
            if msg_id > 0:
                if stream_id in self._ooo_data:
                    self.check_stream_order(self._ooo_data[stream_id], msg_id, republished, redelivered)
                else:
                    order_data = SingleStreamOrderCheckingData(prev_msg_id=msg_id, prev_committed_msg_id=msg_id-1)
                    self._ooo_data[stream_id] = order_data

    def check_stream_order(self, order_data: SingleStreamOrderCheckingData, msg_id, republished, redelivered):
        prev_msg_id = order_data.prev_msg_id
        if msg_id != (prev_msg_id + 1):
            # Message is out of order
            if msg_id < prev_msg_id:
                # Remove form lat msg list
                # If it wasn't in the lost msg list, it is a duplicate
                ooo_msg_id = msg_id
                if ooo_msg_id not in order_data.lost_msg_ids:
                    if redelivered:
                        order_data.redelivered_dup_msg_ids.append(ooo_msg_id)
                    elif republished:
                        order_data.republished_dup_msg_ids.append(ooo_msg_id)
                    else:
                        order_data.dup_msg_ids.append(ooo_msg_id)
                        # A redelivered duplicate message is never considered out of order.
                        # The redelivered flag is a get out of jail card.  These
                        # messages can arrive at any time (based on the api contract)
                        order_data.ooo_msg_ids[ooo_msg_id] = prev_msg_id
                else:
                    order_data.lost_msg_ids.remove(ooo_msg_id)
                    # However a message that was lost and now found
                    # is always an out-of-order message even if
                    # redelivered.
                    order_data.ooo_msg_ids[ooo_msg_id] = prev_msg_id
                # Don't increment expected msg

            if msg_id > (prev_msg_id + 1):
                self.update_lost_msgs(order_data, prev_msg_id, msg_id)
                order_data.prev_msg_id = msg_id

            if msg_id == prev_msg_id:
                if redelivered:
                    order_data.redelivered_dup_msg_ids.append(msg_id)
                elif republished:
                    order_data.republished_dup_msg_ids.append(msg_id)
                else:
                    order_data.dup_msg_ids.append(msg_id)
                # Don't increment expected msg
        else:
            # In order message received so update previous message id
            order_data.prev_msg_id += 1

    def update_lost_msgs(self, order_data: SingleStreamOrderCheckingData, prev_msg_id, end_id):
        if end_id - prev_msg_id > 10000:
            logger.warn(f"STATS {self._sub_name} Error during loss check. > 10000 msgs lost. "
                        f"Current msg ID: {prev_msg_id} Next msg ID received: {end_id}")
            order_data.lost_msg_ids.append(prev_msg_id+1)
            return
        for msg_id in range(prev_msg_id+1, end_id):
            order_data.lost_msg_ids.append(msg_id)

    def check_crc(self, crc, msg):
        # SOL-68082: The property values from NGJava messages are always in string format, so we need to convert to 
        # an integer explicitly. Once this bug is fixed, remove this condition.
        if isinstance(crc, str):
            crc = int(crc)
        if (binascii.crc32(msg) & 0xffffffff) == (crc & 0xffffffff):
            self.stats["binCrcOk"] += 1
            self.stats["crcOk"] += 1
        else:
            self.stats["binCrcFail"] += 1
            self.stats["crcFail"] += 1

    def get_latency_bucket_size(self):
        return 1 << self._lat_granularity

    def inc_latency(self, time_recvd, latency):
        self.update_time_recvd(time_recvd)
        # Skip latency calculation during warmup
        if time_recvd < self._receive_start_time_in_nanos + self._lat_warmup_in_sec * Constants.SEC_TO_NANOSEC:
            return

        usec_lat = latency / 1000

        if usec_lat < self.stats["minLat"]:
            self.stats["minLat"] = usec_lat
        if usec_lat > self.stats["maxLat"]:
            self.stats["maxLat"] = usec_lat

        self.stats["totalLat"] += usec_lat
        self.stats["latMsgs"] += 1

        bucket_idx = int(usec_lat) >> self._lat_granularity
        if bucket_idx >= len(self._lat_buckets):
            bucket_idx = len(self._lat_buckets) - 1
        if bucket_idx < 0:
            logger.error(f"NEGATIVE LATENCY {usec_lat}")
            bucket_idx = 0

        self._lat_buckets[bucket_idx] += 1

    def get_percentile_latency_in_usec(self, perc):
        percent_num_sample = self.get_stat("latMsgs") * perc / 100
        num_sample = 0
        lat_usec = 0.0
        is_done = False
        bucket_size_usec = 1 << self._lat_granularity
        bucket_end_usec = bucket_size_usec

        bucket_loop = 0
        while bucket_loop < len(self._lat_buckets) and not is_done:
            if self._lat_buckets[bucket_loop] > 0:
                num_sample += self._lat_buckets[bucket_loop]
                if num_sample >= percent_num_sample and bucket_loop != (len(self._lat_buckets) - 1):
                    is_done = True
                    lat_usec = bucket_end_usec
            bucket_end_usec += bucket_size_usec
            bucket_loop += 1
        return lat_usec

    def get_latency_std_dev_in_usec(self):
        num_samples = self.get_stat("latMsgs")
        if num_samples == 0:
            return 0
        std_dev_sum = 0
        bucket_size_usec = 1 << self._lat_granularity
        bucket_end_usec = bucket_size_usec
        avg_lat = self.get_stat("latAvg")

        bucket_loop = 0
        while bucket_loop < len(self._lat_buckets):
            if self._lat_buckets[bucket_loop] > 0:
                temp_sum = math.pow(bucket_end_usec - avg_lat, 2)
                std_dev_sum += temp_sum * self._lat_buckets[bucket_loop]
            bucket_end_usec += bucket_size_usec
            bucket_loop += 1
        return math.sqrt(std_dev_sum / num_samples)

    def inc_struct_check(self, ok):
        if ok:
            self.stats['structCheckOk'] += 1
            self.stats['crcOk'] += 1
        else:
            self.stats['structCheckFail'] += 1
            self.stats['crcFail'] += 1

    def inc_stat(self, stat_name, inc=1):
        try:
            self.stats[stat_name] += inc
        except Exception as e:
            logger.warn(f"Stat {stat_name} does not exist: {e}")

    def update_replication_group_message_ooo_detection(self, key: ReplicationGroupMessageId, id_to_add: str, index: int,
                                                       inc_ooo: bool, inc_suid_change: bool):
        if len(self._replication_group_message_ooo_detection) == 0 or \
                self._replication_group_message_ooo_detection.get(key) is None:
            self._replication_group_message_ooo_detection[key] = []
            self._replication_group_message_ooo_detection.get(key).append(id_to_add)
            return
        elif not(inc_ooo or inc_suid_change):
            self._replication_group_message_ooo_detection.get(key).set(index, id_to_add)
        if inc_ooo:
            self.inc_stat('numReplicationGroupMessageIdDuplicates')
        if inc_suid_change:
            self._replication_group_message_ooo_detection.get(key).add(id_to_add)
            self.inc_stat('numReplicationGroupMessageIdSpoolMessageIdChanges')

    def get_replication_group_message_ooo_detection_list(self, key) -> list:
        return self._replication_group_message_ooo_detection.get(key)

    def get_name(self):
        return self._sub_name

    def __str__(self):
        return f"OOO: {self.stats['oooMsgs']}\nCRC OK: {self.stats['crcOk']}, CRC FAIL: {self.stats['crcFail']}\n" \
               f"Lat Min: {self.get_stat('minLat')}, Max: {self.get_stat('maxLat')}, Total: {self.get_stat('totalLat')}"
