import datetime
import queue
import threading
import time
from typing import Union

from opentelemetry import baggage
from opentelemetry.trace import SpanKind
from solace.messaging.builder.direct_message_publisher_builder import DirectMessagePublisherBuilder
from solace.messaging.builder.direct_message_receiver_builder import DirectMessageReceiverBuilder
from solace.messaging.builder.persistent_message_publisher_builder import PersistentMessagePublisherBuilder
from solace.messaging.builder.persistent_message_receiver_builder import PersistentMessageReceiverBuilder
from solace.messaging.builder.request_reply_message_publisher_builder import RequestReplyMessagePublisherBuilder
from solace.messaging.config.authentication_strategy import ClientCertificateAuthentication, Kerberos, \
    BasicUserNamePassword, OAuth2
from solace.messaging.config.missing_resources_creation_configuration import MissingResourcesCreationStrategy
from solace.messaging.config.receiver_activation_passivation_configuration import ReceiverStateChangeListener, \
    ReceiverState
from solace.messaging.config.replay_strategy import ReplayStrategy
from solace.messaging.config.retry_strategy import RetryStrategy
from solace.messaging.config.solace_properties import authentication_properties
from solace.messaging.config.solace_properties.service_properties import VPN_NAME, PAYLOAD_COMPRESSION_LEVEL
from solace.messaging.config.solace_properties.transport_layer_properties import HOST, KEEP_ALIVE_INTERVAL
from solace.messaging.config.transport_security_strategy import TLS
from solace.messaging.config.message_acknowledgement_configuration import Outcome
from solace.messaging.errors.pubsubplus_client_error import PubSubTimeoutError
from solace.messaging.messaging_service import MessagingService, ServiceInterruptionListener, ReconnectionListener, \
    ReconnectionAttemptListener, ServiceEvent, MessagingServiceClientBuilder, TransactionalMessagingService
from solace.messaging.publisher.direct_message_publisher import PublishFailureListener, FailedPublishEvent, \
    DirectMessagePublisher
from solace.messaging.publisher.outbound_message import OutboundMessage
from solace.messaging.publisher.outbound_message import OutboundMessageBuilder
from solace.messaging.publisher.persistent_message_publisher import MessagePublishReceiptListener, PublishReceipt, \
    PersistentMessagePublisher
from solace.messaging.publisher.request_reply_message_publisher import RequestReplyMessagePublisher
from solace.messaging.publisher.transactional_message_publisher import TransactionalMessagePublisher
from solace.messaging.receiver.direct_message_receiver import DirectMessageReceiver
from solace.messaging.receiver.inbound_message import InboundMessage
from solace.messaging.receiver.request_reply_message_receiver import RequestReplyMessageReceiver
from solace.messaging.receiver.transactional_message_receiver import TransactionalMessageReceiver
from solace.messaging.resources.cached_message_subscription_request import CachedMessageSubscriptionRequest
from solace.messaging.resources.queue import Queue
from solace.messaging.resources.topic import Topic
from solace.messaging.resources.topic_subscription import TopicSubscription
from solace.messaging.utils.cache_request_outcome import CacheRequestOutcome
from solace.messaging.utils.cache_request_outcome_listener import CacheRequestOutcomeListener
from solace_otel.messaging.trace.propagation import OutboundMessageCarrier, OutboundMessageSetter, OutboundMessageGetter


from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.authentication_scheme import AuthenticationScheme
from sdkperf.config.channel_state import ChannelState
from sdkperf.config.client_mode_type import ClientModeType
from sdkperf.config.constants import Constants
from sdkperf.config.endpoint_access_type import EndpointAccessType
from sdkperf.config.ep_config_properties import EpConfigProperties
from sdkperf.config.flow_status import FlowStatus, ActiveFlowIndicationType
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.config.message_object_type import MessageObjectType
from sdkperf.config.publisher_back_pressure_type import PublisherBackPressureType
from sdkperf.config.subscriber_back_pressure_type import SubscriberBackPressureType
from sdkperf.config.subscriber_destinations_type import SubscriberDestinationsType
from sdkperf.core.abstract_client import AbstractClient
from sdkperf.core.message_listener import MessageHandlerImpl, RequestMessageHandlerImpl, MessageListener
from sdkperf.core.msg_rep import MessageRep
from sdkperf.core.perf_stats import PerfStats
from sdkperf.core.python_client_transacted_session import PythonClientTransactedSession
from sdkperf.core.tracing_utils import TracingUtils

logger = get_logger()


class PythonClient(AbstractClient, ServiceInterruptionListener,
                   ReconnectionAttemptListener, ReconnectionListener,
                   CacheRequestOutcomeListener):

    def __init__(self):
        super().__init__()

    # noinspection PyAttributeOutsideInit
    def init(self, props: RuntimeProperties, client_id: int):
        super().init(props, client_id)
        builder: MessagingServiceClientBuilder = MessagingService.builder()
        self._apply_solace_properties(builder)
        if hasattr(self._rx_props, RuntimeProperties.CLIENT_NAME_PREFIX):
            self.service: MessagingService = builder.build(application_id=self._client_id_str)
        else:
            self.service: MessagingService = builder.build()

        self._default_consumer = None
        self._default_direct_producer = None
        self._default_persistent_producer = None
        self._default_request_reply_producer = None
        self._consumers_list = list()
        self._transacted_sessions = dict()
        # Note: PRODUCER LIST HAS DIFFERENT BEHAVIOR THAN USUAL
        # Due to the fact that the new api has two different types of publishers we have to make sure we can
        # accommodate whatever kind of publishing the user throws at us.  To that end when we create new
        # publishers we will now create one of each and then store them as a tuple in this list. When publish
        # is called later the client will pick the correct publisher based on the message type.
        # THE ORDER MUST ALWAYS BE (DIRECT, PERSISTENT, RR)
        self._producer_list = list()
        self._producer_list_index = 0
        self._backpressure_index = 0
        if len(getattr(self._rx_props, RuntimeProperties.SDM_PROP_LIST, [])) > 0:
            self.stats.set_sdm_props(getattr(self._rx_props, RuntimeProperties.SDM_PROP_LIST))

        self._messages_list = [None] * self._num_pubs_per_session
        self.sub_memory = list()
        self.request_reply_receivers = list()
        self.last_exception = None

        self._reply_message_queue = None
        self._reply_message_listener = None
        self._reply_message_handler_thread = None

        self._cache_condition = threading.Condition()
        self._cache_request_received = False
        self._request_id = 0

    def get_last_acked_msg_id(self) -> int:
        return self._last_acked_msg_id

    def set_last_acked_msg_id(self, msg_id: int):
        self._last_acked_msg_id = msg_id

    def update_auth_on_reconnect(self, access_token: str, id_token: str):
        try:
            self.service.update_property(authentication_properties.SCHEME_OAUTH2_ACCESS_TOKEN, access_token)
        except Exception as e:
            self.set_last_exception(e)

        try:
            self.service.update_property(authentication_properties.SCHEME_OAUTH2_OIDC_ID_TOKEN, id_token)
        except Exception as e:
            self.set_last_exception(e)

    def get_last_exception(self) -> Exception:
        return self.last_exception

    def set_last_exception(self, exception: Exception):
        self.last_exception = exception

    def clear_last_exception(self):
        self.last_exception = None

    def get_rx_stats(self) -> PerfStats:
        return self.stats

    def get_metric(self, metric) -> Union[int, None]:
        if self.service is None or self.channel_state == ChannelState.CLIENT_STATE_UNKNOWN or self.channel_state == ChannelState.CLIENT_STATE_DISCONNECTED:
            return None
        return self.service.metrics().get_value(metric)

    def clear_metrics(self) -> bool:
        if self.service is None or self.channel_state == ChannelState.CLIENT_STATE_UNKNOWN or self.channel_state == ChannelState.CLIENT_STATE_DISCONNECTED:
            return False
        self.service.metrics().reset()
        return True

    def connect(self):
        if self.service is None:
            raise Exception("Trying to connect None service, this is usually from calling connect after disconnect")

        if self.is_connected():
            return

        try:
            if getattr(self._rx_props, RuntimeProperties.WANT_NON_BLOCKING_CONNECT, False):
                f = self.service.connect_async()
                res = f.result()
                logger.info(f"Result from connect async: {res}")
            else:
                self.service.connect()
            self.service.add_reconnection_listener(self)
            self.service.add_reconnection_attempt_listener(self)
            self.service.add_service_interruption_listener(self)

            if getattr(self._rx_props, RuntimeProperties.WANT_TRACE_CONTEXT_PROP, False):
                TracingUtils.init_tracing(
                    getattr(self._rx_props, RuntimeProperties.TRACE_EXPORTER_CONNECTION_STRING, ""),
                    getattr(self._rx_props, RuntimeProperties.TRACE_EXPORTER_BATCH_SIZE, 5),
                    getattr(self._rx_props, RuntimeProperties.TRACE_EXPORTER_QUEUE_SIZE, 2048),
                    getattr(self._rx_props, RuntimeProperties.TRACE_EXPORTER_SCHEDULE_DELAY, 100))

            want_only_direct = getattr(self._rx_props, RuntimeProperties.WANT_SINGLE_PUB_TYPE, False) and getattr(
                self._rx_props, RuntimeProperties.PUB_MESSAGE_TYPE,
                GenericMessageDeliveryMode.DIRECT) == GenericMessageDeliveryMode.DIRECT
            want_only_persistent = getattr(self._rx_props, RuntimeProperties.WANT_SINGLE_PUB_TYPE, False) and getattr(
                self._rx_props, RuntimeProperties.PUB_MESSAGE_TYPE,
                GenericMessageDeliveryMode.DIRECT) == GenericMessageDeliveryMode.PERSISTENT
            want_only_request_reply = getattr(self._rx_props, RuntimeProperties.WANT_SINGLE_PUB_TYPE, False) and \
                                      getattr(self._rx_props, RuntimeProperties.WANT_REPLY_TOPIC, False)
            not_want_any = not (want_only_direct or want_only_persistent or want_only_request_reply)
            if want_only_direct or not_want_any:
                direct_pub_builder: DirectMessagePublisherBuilder = \
                    self.service.create_direct_message_publisher_builder()
                self._set_publisher_backpressure(direct_pub_builder)
                self._default_direct_producer: DirectMessagePublisher = direct_pub_builder.build()
                self._termination_listener.register_for_listen(self._default_direct_producer)
                self._default_direct_producer.start()
                if getattr(self._rx_props, RuntimeProperties.WANT_PRODUCER_LISTENERS, True):
                    self._default_direct_producer.set_publish_failure_listener(PublishFailureListenerImpl(self.stats))
            else:
                self._default_direct_producer = None

            if want_only_persistent or not_want_any:
                persistent_pub_builder: PersistentMessagePublisherBuilder = \
                    self.service.create_persistent_message_publisher_builder()
                self._set_publisher_backpressure(persistent_pub_builder)
                self._default_persistent_producer: PersistentMessagePublisher = persistent_pub_builder.build()
                self._termination_listener.register_for_listen(self._default_persistent_producer)
                self._default_persistent_producer.start()
                if getattr(self._rx_props, RuntimeProperties.WANT_PRODUCER_LISTENERS, True):
                    self._default_persistent_producer.set_message_publish_receipt_listener(
                        MessagePublishReceiptListenerImpl(self.stats, self))
            else:
                self._default_persistent_producer = None

            if want_only_request_reply or not_want_any:
                rr_pub_builder: RequestReplyMessagePublisherBuilder = \
                    self.service.request_reply().create_request_reply_message_publisher_builder()
                self._default_request_reply_producer: RequestReplyMessagePublisher = rr_pub_builder.build()
                self._termination_listener.register_for_listen(self._default_request_reply_producer)
                self._default_request_reply_producer.start()
            else:
                self._default_request_reply_producer = None

            self._backpressure_index += 1
            num_msg_producers = getattr(self._rx_props, RuntimeProperties.NUM_MESSAGE_PRODUCER, 1)
            if num_msg_producers > 1:
                self._producer_list = list()
                self._producer_list.append((self._default_direct_producer, self._default_persistent_producer,
                                            self._default_request_reply_producer))

                for i in range(num_msg_producers - 1):
                    direct_prod: Union[DirectMessagePublisher, None] = None
                    persistent_prod: Union[PersistentMessagePublisher, None] = None
                    request_reply_producer: Union[RequestReplyMessagePublisher, None] = None
                    if want_only_direct or not_want_any:
                        direct_pub_builder = self.service.create_direct_message_publisher_builder()
                        self._set_publisher_backpressure(direct_pub_builder)
                        direct_prod: DirectMessagePublisher = direct_pub_builder.build()
                        self._termination_listener.register_for_listen(direct_prod)
                        direct_prod.start()
                        if getattr(self._rx_props, RuntimeProperties.WANT_PRODUCER_LISTENERS, True):
                            direct_prod.set_publish_failure_listener(PublishFailureListenerImpl(self.stats))

                    if want_only_persistent or not_want_any:
                        persistent_pub_builder = self.service.create_persistent_message_publisher_builder()
                        self._set_publisher_backpressure(persistent_pub_builder)
                        persistent_prod: PersistentMessagePublisher = persistent_pub_builder.build()
                        self._termination_listener.register_for_listen(persistent_prod)
                        persistent_prod.start()
                        if getattr(self._rx_props, RuntimeProperties.WANT_PRODUCER_LISTENERS, True):
                            persistent_prod.set_message_publish_receipt_listener(
                                MessagePublishReceiptListenerImpl(self.stats, self))

                    if want_only_request_reply or not_want_any:
                        rr_pub_builder: RequestReplyMessagePublisherBuilder = self.service.request_reply() \
                            .create_request_reply_message_publisher_builder()
                        request_reply_producer = rr_pub_builder.build()
                        self._termination_listener.register_for_listen(request_reply_producer)
                        request_reply_producer.start()

                    self._producer_list.append((direct_prod, persistent_prod, request_reply_producer))
                    self._backpressure_index += 1

            self._connect_default_flow()
        except Exception as e:
            self.set_last_exception(e)
            raise e
        self.channel_state = ChannelState.CLIENT_STATE_CONNECTED

    def disconnect(self, grace_period):
        self.channel_state = ChannelState.CLIENT_STATE_DISCONNECTED
        if not getattr(self._rx_props, RuntimeProperties.WANT_UNGRACEFUL_TERMINATE, False):
            # Stop any producers
            if getattr(self._rx_props, RuntimeProperties.NUM_MESSAGE_PRODUCER, 1) > 1:
                for dir_prod, pers_prod, rr_prod in self._producer_list:
                    if dir_prod is not None and not dir_prod.is_terminated():
                        try:
                            dir_prod.terminate(grace_period)
                        except Exception as e:
                            logger.error(f"Terminate direct producer failed: {e}")
                        finally:
                            del dir_prod
                            dir_prod = None

                    if pers_prod is not None and not pers_prod.is_terminated():
                        try:
                            pers_prod.terminate(grace_period)
                        except Exception as e:
                            logger.error(f"Terminate persistent producer failed: {e}")
                        finally:
                            del pers_prod
                            pers_prod = None

                    if rr_prod is not None and not rr_prod.is_terminated():
                        try:
                            rr_prod.terminate(grace_period)
                        except Exception as e:
                            logger.error(f"Terminate request/reply producer failed: {e}")
                        finally:
                            del rr_prod
                            rr_prod = None
                self._producer_list.clear()
            else:
                if self._default_direct_producer is not None and not self._default_direct_producer.is_terminated():
                    try:
                        self._default_direct_producer.terminate(grace_period)
                    except Exception as e:
                        logger.error(f"Terminate default direct producer failed: {e}")
                    finally:
                        del self._default_direct_producer
                        self._default_direct_producer = None

                if self._default_persistent_producer is not None and not self._default_persistent_producer.is_terminated():
                    try:
                        self._default_persistent_producer.terminate(grace_period)
                    except Exception as e:
                        logger.error(f"Terminate default persistent producer failed: {e}")
                    finally:
                        del self._default_persistent_producer
                        self._default_persistent_producer = None

                if self._default_request_reply_producer is not None and \
                        not self._default_request_reply_producer.is_terminated():
                    try:
                        self._default_request_reply_producer.terminate(grace_period)
                    except Exception as e:
                        logger.error(f"Terminate default request/reply producer failed: {e}")
                    finally:
                        del self._default_request_reply_producer
                        self._default_request_reply_producer = None

            if self._reply_message_handler_thread is not None:
                try:
                    self._stop_replay_message_handler_thread()
                except Exception as e:
                    logger.error(f"Stopping reply message handler failed: {e}")
                finally:
                    self._reply_message_handler_thread = None

            # Used for AD subscribers
            for consumer in self._consumers_list:
                if consumer.consumer is not None and not consumer.consumer.is_terminated():
                    try:
                        consumer.consumer.terminate(grace_period)
                    except Exception as e:
                        logger.error(f"Terminate persistent consumer failed: {e}")
                    finally:
                        del consumer.consumer
                        consumer.consumer = None
                        del consumer
                        consumer = None

            for receiver in self.request_reply_receivers:
                if receiver.consumer is not None and not receiver.consumer.is_terminated():
                    try:
                        receiver.consumer.terminate(grace_period)
                    except Exception as e:
                        logger.error(f"Terminate request/reply consumer failed: {e}")
                    finally:
                        del receiver.consumer
                        receiver.consumer = None
                        del receiver
                        receiver = None

            # Clear out the default consumer
            if self._default_consumer is not None:
                try:
                    if self._default_consumer.consumer is not None and not self._default_consumer.consumer.is_terminated():
                        self._default_consumer.consumer.terminate(grace_period)
                except Exception as e:
                    logger.error(f"Terminate direct consumer failed: {e}")
                finally:
                    if self._default_consumer.consumer is not None:
                        del self._default_consumer.consumer
                        self._default_consumer.consumer = None
                    del self._default_consumer
                    self._default_consumer = None

            # Terminate the transacted pubs (and idle thread if running)
            if len(self._transacted_sessions) > 0:
                for name, session in self._transacted_sessions.items():
                    session.close_pub()
                    session.disconnect()
                self._transacted_sessions.clear()

        # Disconnect the service
        if not getattr(self._rx_props, RuntimeProperties.WANT_UNGRACEFUL_DISCONNECT, False):
            if getattr(self._rx_props, RuntimeProperties.WANT_NON_BLOCKING_CONNECT, False):
                f = self.service.disconnect_async()
                res = f.result()
                logger.info(f"Result from disconnect async: {res}")
            else:
                self.service.disconnect()

    def start(self):
        # Used for AD subscribing (binding to queues)
        for consumer in self._consumers_list:
            if consumer.consumer is not None:
                if not consumer.is_started:
                    consumer.consumer.resume()
                    consumer.is_started = True
                if not consumer.done_receive:
                    consumer.consumer.receive_async(consumer)
                    consumer.done_receive = True

        # Start receiving messages on the default consumer
        self._connect_default_flow()

    def stop(self):
        # Used for AD subscribing (binding to queues)
        for consumer in self._consumers_list:
            if consumer.consumer is not None:
                try:
                    consumer.consumer.pause()
                finally:
                    consumer.is_started = False

        # Stop receiving messages on the default consumer
        if self._default_consumer is not None:
            try:
                self._default_consumer.consumer.terminate(0)
            finally:
                self.sub_memory = self._default_consumer.get_subscriptions()
                self._default_consumer = None

    def open_transacted_session(self, transaction_props: RuntimeProperties):
        session = None
        try:
            session: TransactionalMessagingService = self.service.create_transactional_service_builder() \
                .build().connect()
        except Exception as e:
            self.set_last_exception(e)
            raise e
        prod: TransactionalMessagePublisher = session.create_transactional_message_publisher_builder().build()
        name = session.transactional_service_info().get_transactional_service_id()
        self._transacted_sessions[name] = PythonClientTransactedSession(self, session, prod, transaction_props)
        return name

    def close_transacted_session(self, session_name: str):
        session: PythonClientTransactedSession = self._transacted_sessions.pop(session_name, None)
        if session is None:
            raise Exception(f"CLIENT {self.get_id_str()} : Transacted Session not found ({session_name})")

        session.close_pub()
        session.disconnect()

    def close_all_transacted_sessions(self):
        for name, session in self._transacted_sessions.items():
            session.close_pub()
            session.disconnect()

        self._transacted_sessions.clear()

    def commit_transaction(self, session_name: str, want_rollback: bool):
        session: PythonClientTransactedSession = self._transacted_sessions.get(session_name, None)
        if session is None:
            raise Exception(f"CLIENT {self.get_id_str()} : Transacted Session not found ({session_name})")
        for consumer in self._consumers_list:
            if consumer.transacted_session == session:
                consumer.consumer.pause()
        try:
            session.commit(want_rollback)
        except Exception as e:
            self.set_last_exception(e)
            raise e
        finally:
            for consumer in self._consumers_list:
                if consumer.transacted_session == session and consumer.is_started is True:
                    consumer.consumer.resume()

    def commit_transaction_on_current_pub(self, session_name: str, want_rollback: bool):
        session: PythonClientTransactedSession = self._transacted_sessions.get(session_name, None)
        if session is None:
            raise Exception(f"CLIENT {self.get_id_str()} : Transacted Session not found ({session_name})")
        for consumer in self._consumers_list:
            if consumer.transacted_session == session:
                consumer.consumer.pause()
        try:
            if want_rollback:
                session.get_session().rollback()
            else:
                session.get_session().commit()
        except Exception as e:
            self.set_last_exception(e)
            raise e
        finally:
            for consumer in self._consumers_list:
                if consumer.transacted_session == session and consumer.is_started is True:
                    consumer.consumer.resume()

    def get_flow_status(self, ep_name: str) -> FlowStatus:
        status = FlowStatus()
        for cons in self._consumers_list:
            if cons.destination == ep_name:
                status = cons.get_flow_status()
                break
        return status

    def is_ready_to_publish(self, delivery_mode: GenericMessageDeliveryMode) -> bool:
        if delivery_mode == GenericMessageDeliveryMode.PERSISTENT:
            return self._default_persistent_producer.is_ready()
        elif delivery_mode == GenericMessageDeliveryMode.DIRECT:
            return self._default_direct_producer.is_ready()
        else:
            return False

    def subscription_update(self, ep_props: EpConfigProperties):
        # Make sure the default consumer is set up
        if ep_props.is_adding:
            self._connect_default_flow()

        if ep_props.type == SubscriberDestinationsType.NULL:
            ep_props.type = SubscriberDestinationsType.TOPIC

        # If we are removing the default queue from the default consumer then
        # we can just stop it now
        subs = ep_props.subscriptions
        if Constants.SUB_DEFAULT_QUEUE_NAME in subs:
            if not ep_props.is_adding and self._default_consumer is not None:
                try:
                    self._default_consumer.consumer.terminate(0)
                finally:
                    self._default_consumer.consumer = None
                    self._default_consumer = None
            return

        if getattr(self._rx_props, RuntimeProperties.CLIENT_MODE, ClientModeType.SINK) != ClientModeType.REPLY:
            # This list holds tuples of (Future, topic, is add) that are used
            # After the add/remove async they are used to update the consumer list
            futures = list()

            # Add/remove all of the subscriptions to the default consumer
            # Async for all but the last one to block until all added
            for i in range(len(ep_props.subscriptions)):
                want_async = True
                if i == (len(ep_props.subscriptions) - 1):
                    want_async = False

                topic = ep_props.subscriptions[i]
                subscription = TopicSubscription.of(topic)

                if self._default_consumer is not None:
                    try:
                        if ep_props.is_adding:
                            # Add it to the actual consumer
                            if want_async:
                                futures.append(
                                    (self._default_consumer.consumer.add_subscription_async(subscription), topic, True))
                            else:
                                self._default_consumer.consumer.add_subscription(subscription)
                                self._default_consumer.add_subscription(topic)
                        else:
                            # Remove it from the actual consumer
                            if want_async:
                                futures.append(
                                    (self._default_consumer.consumer.remove_subscription_async(subscription), topic,
                                     False))
                            else:
                                self._default_consumer.consumer.remove_subscription(subscription)
                                self._default_consumer.remove_subscription(topic)
                    except Exception as e:
                        logger.error(f"Caught error in subscription_update {e}")
                        self.set_last_exception(e)

            for future, topic, is_add in futures:
                try:
                    future.result()
                    if is_add:
                        self._default_consumer.add_subscription(topic)
                    else:
                        self._default_consumer.remove_subscription(topic)
                except Exception as e:
                    logger.error(f"Caught error from future in subscription_update: {e}")
                    self.set_last_exception(e)
        else:  # Doing request/reply
            for i in range(len(ep_props.subscriptions)):
                topic = ep_props.subscriptions[i]
                subscription = TopicSubscription.of(topic)
                try:
                    if ep_props.is_adding:
                        new_consumer: RequestReplyMessageReceiver = self.service.request_reply(). \
                            create_request_reply_message_receiver_builder().build(subscription)
                        self._termination_listener.register_for_listen(new_consumer)
                        new_handler = RequestMessageHandlerImpl(self, None, True, subscription, False, False,
                                                                getattr(self._rx_props,
                                                                        RuntimeProperties.RECEIVER_SLEEP_TIME, 0))
                        new_handler.consumer = new_consumer
                        new_handler.consumer.start()
                        new_handler.is_started = True
                        new_handler.consumer.receive_async(new_handler)
                        new_handler.done_receive = True
                        self.request_reply_receivers.append(new_handler)
                    else:
                        i = 0
                        for receiver in self.request_reply_receivers:
                            if subscription == receiver.topic:
                                to_remove = self.request_reply_receivers.pop(i)
                                to_remove.terminate(0)
                                break
                            i += 1
                except Exception as e:
                    logger.error(f"Caught error in request reply receiver of subscription_update {e}")
                    self.set_last_exception(e)

    def queue_update(self, ep_props: EpConfigProperties):
        queues = ep_props.ep_names
        selectors = ep_props.selectors
        settlement_patterns = ep_props.settlement_patterns
        required_settlements = ep_props.required_settlements

        i_queue_size = len(queues)
        i_selector_size = len(selectors)
        i_settlement_patterns_size = len(settlement_patterns)
        i_required_settlements_size = len(required_settlements)
        i = 0

        if i_queue_size == 0:
            raise Exception(f"CLIENT {self.get_id_str()} Empty queue list provided. Aborting queue update")
        if i_selector_size != i_queue_size and i_selector_size != 0:
            raise Exception(f"CLIENT {self.get_id_str()} Queue list must be the same size as selector list")
        if i_settlement_patterns_size >= 1 and i_queue_size != i_settlement_patterns_size:
            raise Exception(f"CLIENT {self.get_id_str()} Queue list must be the same size as settlement patterns list or settlement pattern list must be size 1 to apply to all queues")
        if i_required_settlements_size >= 1 and i_queue_size != i_required_settlements_size:
            raise Exception(f"CLIENT {self.get_id_str()} Queue list must be the same size as required settlements list or required settlements list must be size 1 to apply to all queues")

        want_msg_priority_order_checking = ep_props.want_message_priority_order_checking
        try:
            if ep_props.is_adding:
                for queue_name in queues:
                    if ep_props.access_type == EndpointAccessType.ACCESSTYPE_NONEXCLUSIVE:
                        q = Queue.durable_non_exclusive_queue(queue_name)
                    else:
                        q = Queue.durable_exclusive_queue(queue_name)

                    transacted_session: PythonClientTransactedSession = None
                    if ep_props.transacted_session_name != "":
                        transacted_session = self._transacted_sessions.get(ep_props.transacted_session_name, None)

                    if transacted_session is not None:
                        builder = transacted_session.get_session().create_transactional_message_receiver_builder()
                    else:
                        builder = self.service.create_persistent_message_receiver_builder()
                    status = FlowStatus()
                    if i_selector_size > 0:
                        builder = builder.with_message_selector(selectors[i])
                    if ep_props.want_active_flow_indication:
                        activation_listener = ActivationListener(status)
                        builder.with_activation_passivation_support(activation_listener)
                    if ep_props.want_provision:
                        builder.with_missing_resources_creation_strategy(
                            MissingResourcesCreationStrategy.CREATE_ON_START)
                        
                    # The settlement pattern will be passed to the listener when receiving messages
                    settlement_pattern_string = None
                    if i_settlement_patterns_size > 0:
                        settlement_pattern_string = settlement_patterns[i % i_settlement_patterns_size]
                    else:
                        settlement_pattern_string = getattr(self._rx_props, RuntimeProperties.SETTLEMENT_PATTERN_STRING, None)
                    
                    # The required settlements will be used now on the builder to set up what it can handle
                    required_settlements_string = None
                    if i_required_settlements_size > 0:
                        required_settlements_string = required_settlements[i % i_required_settlements_size]
                    else:
                        required_settlements_string = getattr(self._rx_props, RuntimeProperties.REQUIRED_SETTLEMENTS_STRING, None)
                    if required_settlements_string is not None:
                        required_settlements_list = list()
                        if 'a' in required_settlements_string.lower():
                            required_settlements_list.append(Outcome.ACCEPTED)
                        if 'r' in required_settlements_string.lower():
                            required_settlements_list.append(Outcome.REJECTED)
                        if 'f' in required_settlements_string.lower():
                            required_settlements_list.append(Outcome.FAILED)
                        if len(required_settlements_list) > 0:
                            builder.with_required_message_outcome_support(*required_settlements_list)

                    PythonClient._handle_replay(builder, ep_props)

                    consumer = builder.build(q)
                    self._termination_listener.register_for_listen(consumer)
                    listener = MessageHandlerImpl(self, queue_name, False, None, False,
                                                  transacted_session, want_msg_priority_order_checking,
                                                  getattr(self._rx_props, RuntimeProperties.RECEIVER_SLEEP_TIME, 0),
                                                  status, settlement_pattern_string)

                    listener.consumer = consumer
                    if transacted_session is not None:
                        listener.consumer.receive_async(listener)
                        listener.done_receive = True
                    listener.consumer.start()
                    listener.is_started = True
                    i += 1

                    self._consumers_list.append(listener)
            else:
                match_found = False
                for queue_name in queues:
                    for x in range(len(self._consumers_list)):
                        # This is here because we are modifying the list on the fly and can reach a sate
                        # where we access a element that has been deleted
                        if x < len(self._consumers_list):
                            if self._consumers_list[x].destination == queue_name and \
                                    self._consumers_list[x].is_topic is False:
                                try:
                                    if not self._consumers_list[x].consumer.is_terminated():
                                        self._consumers_list[x].consumer.terminate(0)
                                except Exception as e:
                                    self.set_last_exception(e)
                                    logger.error(f"Got exception terminating queue consumer {str(e)}")
                                finally:
                                    self._consumers_list[x].consumer.is_started = False
                                self._consumers_list.pop(x)
                                match_found = True
                if not match_found:
                    logger.warning(f"CLIENT {self.get_id_str()}: Did not find matching in topic remove")

        except Exception as e:
            self.set_last_exception(e)
            logger.error(f"Exception in queue update {str(e)}")
            raise e

    def temp_endpoint_update(self, ep_props: EpConfigProperties):
        num_temp_endpoints = ep_props.num_temp_endpoints
        selectors = ep_props.selectors
        topics = ep_props.subscriptions
        settlement_patterns = ep_props.settlement_patterns
        required_settlements = ep_props.required_settlements

        i_selector_size = len(selectors)
        i_settlement_patterns_size = len(settlement_patterns)
        i_required_settlements_size = len(required_settlements)

        if 0 < num_temp_endpoints != i_selector_size > 0 and i_selector_size != 1:
            logger.error(
                f"CLIENT {self.get_id_str()}: Size of selector list and number of temporary endpoints are not equal")
            return
        if i_settlement_patterns_size >= 1 and num_temp_endpoints != i_settlement_patterns_size:
            logger.error(f"CLIENT {self.get_id_str()} Num temp endpoints must be the same size as settlement patterns list or settlement pattern list must be size 1 to apply to all temp endpoints")
            return
        if i_required_settlements_size >= 1 and num_temp_endpoints != i_required_settlements_size:
            logger.error(f"CLIENT {self.get_id_str()} Num temp endpoints must be the same size as required settlements list or required settlements list must be size 1 to apply to all temp endpoints")
            return
        
        want_msg_priority_order_checking = ep_props.want_message_priority_order_checking
        ret_list = []
        try:
            for i in range(num_temp_endpoints):
                q = Queue.non_durable_exclusive_queue()

                transacted_session: PythonClientTransactedSession = None
                if ep_props.transacted_session_name != "":
                    transacted_session = self._transacted_sessions.get(ep_props.transacted_session_name, None)

                if transacted_session is not None:
                    builder = transacted_session.get_session().create_transactional_message_receiver_builder()
                else:
                    builder = self.service.create_persistent_message_receiver_builder()
                status = FlowStatus()
                topic = None
                if topics is not None and len(topics) != 0:
                    topic = topics[i % len(topics)]
                if i_selector_size > 0:
                    builder = builder.with_message_selector(selectors[i])
                if ep_props.want_active_flow_indication:
                    activation_listener = ActivationListener(status)
                    builder.with_activation_passivation_support(activation_listener)
                    
                # The settlement pattern will be passed to the listener when receiving messages
                settlement_pattern_string = None
                if i_settlement_patterns_size > 0:
                    settlement_pattern_string = settlement_patterns[i % i_settlement_patterns_size]
                else:
                    settlement_pattern_string = getattr(self._rx_props, RuntimeProperties.SETTLEMENT_PATTERN_STRING, None)
                
                # The required settlements will be used now on the builder to set up what it can handle
                required_settlements_string = None
                if i_required_settlements_size > 0:
                    required_settlements_string = required_settlements[i % i_required_settlements_size]
                else:
                    required_settlements_string = getattr(self._rx_props, RuntimeProperties.REQUIRED_SETTLEMENTS_STRING, None)
                if required_settlements_string is not None:
                    required_settlements_list = list()
                    if 'a' in required_settlements_string.lower():
                        required_settlements_list.append(Outcome.ACCEPTED)
                    if 'r' in required_settlements_string.lower():
                        required_settlements_list.append(Outcome.REJECTED)
                    if 'f' in required_settlements_string.lower():
                        required_settlements_list.append(Outcome.FAILED)
                    if len(required_settlements_list) > 0:
                        builder.with_required_message_outcome_support(*required_settlements_list)

                PythonClient._handle_replay(builder, ep_props)

                consumer = builder.build(q)
                self._termination_listener.register_for_listen(consumer)
                listener = MessageHandlerImpl(self, "", False, None, True, transacted_session,
                                              want_msg_priority_order_checking,
                                              getattr(self._rx_props, RuntimeProperties.RECEIVER_SLEEP_TIME, 0), status, settlement_pattern_string)

                listener.consumer = consumer
                if transacted_session is not None:
                    listener.consumer.receive_async(listener)
                    listener.done_receive = True
                listener.consumer.start()
                if topic is not None:
                    listener.consumer.add_subscription(TopicSubscription.of(topic))
                    listener.add_subscription(topic)
                listener.is_started = True
                queue_name = listener.consumer.receiver_info().get_resource_info().get_name()
                listener.destination = queue_name
                ret_list.append(queue_name)

                self._consumers_list.append(listener)
        except Exception as e:
            self.set_last_exception(e)
            logger.error(f"Exception in temp endpoint update {str(e)}")
            raise e
        return ret_list

    def unbind_all_temp_endpoints(self):
        for consumer in self._consumers_list:
            if consumer.is_non_durable:
                if consumer.consumer is not None:
                    try:
                        consumer.consumer.terminate(0)
                    except Exception:
                        pass
                self._consumers_list.remove(consumer)

    def request_reply(self, topic: str, request: str):
        destination = Topic.of(topic)
        msg: OutboundMessage = self.service.message_builder().build(bytearray(request.encode('utf-8')))

        requester: RequestReplyMessagePublisher = self.service.request_reply(). \
            create_request_reply_message_publisher_builder().build()
        self._termination_listener.register_for_listen(requester)
        requester.start()

        reply_msg = requester.publish_await_response(request_message=msg,
                                                     request_destination=destination,
                                                     reply_timeout=Constants.DEFAULT_TIMEOUT_IN_MSEC)
        reply_str = ""
        if reply_msg is not None:
            if reply_msg.get_payload_as_string() is not None:
                reply_str = reply_msg.get_payload_as_string()
            elif reply_msg.get_payload_as_bytes() is not None:
                try:
                    reply_str = reply_msg.get_payload_as_bytes().decode('utf-8')
                except Exception:
                    reply_str = ""
        return reply_str

    def map_topics(self, queue_name: str, topics: list, is_adding: bool):
        i_topic_size = len(topics)
        if i_topic_size == 0:
            logger.error(f"CLIENT {self.get_id_str()}: Empty topics list provided. Aborting topic mapping")
            return

        for topic in topics:
            for consumer in self._consumers_list:
                if consumer.destination == queue_name:
                    if is_adding:
                        consumer.consumer.add_subscription(TopicSubscription.of(topic))
                        consumer.add_subscription(topic)
                    else:
                        consumer.consumer.remove_subscription(TopicSubscription.of(topic))
                        consumer.remove_subscription(topic)

    @staticmethod
    def _handle_reply_messages(reply_message_queue: queue.Queue, message_listener: MessageListener):
        while True:
            item_from_queue = reply_message_queue.get()
            if item_from_queue is None:
                reply_message_queue.task_done()
                break

            if isinstance(item_from_queue, InboundMessage):
                message = item_from_queue
            else:
                if item_from_queue.exception() is not None:
                    logger.warn(f"Got exception from request reply publish: {item_from_queue.exception()}")
                    reply_message_queue.task_done()
                    continue
                message = item_from_queue.result()

            message_listener.process_message(message)
            reply_message_queue.task_done()
        return

    def _start_reply_message_handler_thread(self):
        self._reply_message_queue = queue.Queue()
        self._reply_message_listener = RequestMessageHandlerImpl(self, None, False, None, False, False,
                                                                 getattr(self._rx_props,
                                                                         RuntimeProperties.RECEIVER_SLEEP_TIME, 0))
        self._reply_message_handler_thread = threading.Thread(target=PythonClient._handle_reply_messages,
                                                              args=(self._reply_message_queue,
                                                                    self._reply_message_listener))
        self._reply_message_handler_thread.start()

    def _stop_replay_message_handler_thread(self):
        self._reply_message_queue.put_nowait(None)
        self._reply_message_handler_thread.join()
        self._reply_message_handler_thread = None

    def publish_msg(self, message_rep: Union[MessageRep,bytearray], destination: str,
                    pub_session_index: int, additional_properties: dict = None,
                    mode: GenericMessageDeliveryMode = GenericMessageDeliveryMode.DIRECT,
                    msg_obj_type: MessageObjectType = MessageObjectType.NONE, 
                    want_await: bool = False, want_request_reply: bool = False, transacted_session_name: str = ""):
        use_rr = want_request_reply or getattr(self._rx_props, RuntimeProperties.WANT_REPLY_TOPIC, False)
        rr_wait_time = Constants.DEFAULT_TIMEOUT_IN_MSEC
        if getattr(self._rx_props, RuntimeProperties.REQUEST_REPLY_WAIT_TIME, 0) > 0:
            rr_wait_time = getattr(self._rx_props, RuntimeProperties.REQUEST_REPLY_WAIT_TIME)

        if (msg_obj_type != MessageObjectType.NONE):
            msg_type = "x"
            if msg_obj_type == MessageObjectType.BYTES:
                message_content = bytearray([0x31, 0x32, 0x33])
                msg_type = "b"
            elif msg_obj_type == MessageObjectType.TEXT:
                message_content = "TEXT_MESSAGE"
                msg_type = "t"
            elif msg_obj_type == MessageObjectType.STREAM:
                message_content = ["STREAM_MESSAGE"]
                msg_type = "s"
            elif msg_obj_type == MessageObjectType.MAP:
                message_content = {"MAP": "MESSAGE"}
                msg_type = "m"

            builder: OutboundMessageBuilder = self.service.message_builder()
            builder.with_application_message_type(msg_type)
            message_rep = builder.build(message_content)

        want_any_tracing = False
        if (getattr(self._tx_props, RuntimeProperties.WANT_TRACE_CREATION_CONTEXT, False)
            or getattr(self._tx_props, RuntimeProperties.WANT_TRACE_TRANSPORT_CONTEXT, False)
            or getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "") != "") \
                and TracingUtils.is_init:
            want_any_tracing = True

        # If we're calling this from PublishMsg it's just a bytearray
        # The message rep is only from the pubThread
        if isinstance(message_rep, MessageRep):
            message = message_rep.get_message_obj(want_any_tracing)
        else:
            message = message_rep

        if getattr(self._tx_props, RuntimeProperties.WANT_TRACE_CREATION_CONTEXT, False) and TracingUtils.is_init:
            with TracingUtils.tracer.start_as_current_span(f"{destination} sdkperf_py_create", kind=SpanKind.PRODUCER) \
                    as creation_context_span:
                baggage_ctx = None
                if getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "") != "":
                    for bag_item in TracingUtils.get_baggage(
                            getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "")):
                        baggage_ctx = baggage.set_baggage(bag_item[0], bag_item[1], baggage_ctx)
                creation_context_carrier = OutboundMessageCarrier(message)
                TracingUtils.text_map.inject(carrier=creation_context_carrier,
                                             setter=OutboundMessageSetter(),
                                             context=baggage_ctx)
        elif not getattr(self._tx_props, RuntimeProperties.WANT_TRACE_CREATION_CONTEXT, False) \
                and getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "") != "" and TracingUtils.is_init:
            baggage_ctx = None
            if getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "") != "":
                for bag_item in TracingUtils.get_baggage(
                        getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "")):
                    baggage_ctx = baggage.set_baggage(bag_item[0], bag_item[1], baggage_ctx)
            creation_context_carrier = OutboundMessageCarrier(message)
            TracingUtils.text_map.inject(carrier=creation_context_carrier,
                                         setter=OutboundMessageSetter(),
                                         context=baggage_ctx)
        if getattr(self._tx_props, RuntimeProperties.TRACE_BAGGAGE, "") != "":
            baggage.clear()

        if getattr(self._tx_props, RuntimeProperties.WANT_TRACE_TRANSPORT_CONTEXT, False) and TracingUtils.is_init:
            transport_context_carrier = OutboundMessageCarrier(message)
            extracted_ctx = TracingUtils.text_map.extract(carrier=transport_context_carrier,
                                                          getter=OutboundMessageGetter())
            with TracingUtils.tracer.start_as_current_span(f"{destination} sdkperf_py_send", kind=SpanKind.PRODUCER,
                                                           context=extracted_ctx) as transport_context_span:
                TracingUtils.text_map.inject(carrier=transport_context_carrier, setter=OutboundMessageSetter())

        if transacted_session_name != "":
            transacted_session: PythonClientTransactedSession = self._transacted_sessions.get(transacted_session_name, None)
            if transacted_session is None:
                raise Exception(f"CLIENT {self.get_client_name()}: "
                                f"Transacted Session session not found ({transacted_session_name})")
            try:
                transacted_session.get_producer().publish(destination=Topic.of(destination),
                                                          message=message,
                                                          additional_message_properties=additional_properties)
            except Exception as e:
                self.set_last_exception(e)
                raise e
        elif getattr(self._rx_props, RuntimeProperties.NUM_MESSAGE_PRODUCER, 1) > 1:
            prod_tuple = self._producer_list[self._producer_list_index % len(self._producer_list)]
            if use_rr:
                if self._reply_message_handler_thread is None:
                    self._start_reply_message_handler_thread()
                prod = prod_tuple[2]
            elif mode == GenericMessageDeliveryMode.DIRECT:
                prod = prod_tuple[0]
                if want_await:
                    raise Exception(f"Cannot use await ack with direct messaging")
            elif mode == GenericMessageDeliveryMode.PERSISTENT:
                prod = prod_tuple[1]
            else:
                raise Exception(
                    f"Invalid publish type {mode} to pub index {self._producer_list_index % len(self._producer_list)}")
            try:
                result = None
                if want_await:
                    if use_rr:
                        result = prod.publish_await_response(
                            request_destination=Topic.of(destination), request_message=message,
                            reply_timeout=rr_wait_time,
                            additional_message_properties=additional_properties)
                    else:
                        prod.publish_await_acknowledgement(
                            destination=Topic.of(destination), message=message,
                            time_out=Constants.DEFAULT_TIMEOUT_IN_MSEC,
                            additional_message_properties=additional_properties)
                else:
                    if use_rr:
                        result = prod.publish(request_destination=Topic.of(destination), request_message=message,
                                              reply_timeout=rr_wait_time,
                                              additional_message_properties=additional_properties)
                    else:
                        result = prod.publish(destination=Topic.of(destination), message=message,
                                              additional_message_properties=additional_properties)
                self._producer_list_index += 1
                if result is not None and self._reply_message_handler_thread is not None:
                    self._reply_message_queue.put_nowait(result)
            except Exception as e:
                self.set_last_exception(e)
                raise e
        else:
            if use_rr:
                if self._reply_message_handler_thread is None:
                    self._start_reply_message_handler_thread()
                if want_await:
                    result = self._default_request_reply_producer.publish_await_response(
                        request_destination=Topic.of(destination), request_message=message,
                        reply_timeout=rr_wait_time,
                        additional_message_properties=additional_properties)
                else:
                    result = self._default_request_reply_producer.publish(
                        request_destination=Topic.of(destination), request_message=message,
                        reply_timeout=rr_wait_time,
                        additional_message_properties=additional_properties)
                if result is not None and self._reply_message_handler_thread is not None:
                    self._reply_message_queue.put_nowait(result)
            elif mode == GenericMessageDeliveryMode.DIRECT:
                try:
                    self._default_direct_producer.publish(destination=Topic.of(destination), message=message,
                                                          additional_message_properties=additional_properties)
                except Exception as e:
                    self.set_last_exception(e)
                    raise e
            elif mode == GenericMessageDeliveryMode.PERSISTENT:
                try:
                    if want_await:
                        self._default_persistent_producer.publish_await_acknowledgement(
                            destination=Topic.of(destination), message=message,
                            additional_message_properties=additional_properties)
                    else:
                        self._default_persistent_producer.publish(destination=Topic.of(destination), message=message,
                                                                  additional_message_properties=additional_properties)
                except Exception as e:
                    self.set_last_exception(e)
                    raise e
            else:
                raise Exception(f"Invalid publish type {mode}")

    def publish_send_multiple(self):
        raise Exception("Publish Send Multiple is not supported by the python API")

    def cache_request(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool) -> None:
        cache_name = getattr(self._rx_props, RuntimeProperties.CACHE_PROP_NAME, "")
        max_msgs_per_topic = getattr(self._rx_props, RuntimeProperties.CACHE_PROP_MAX_MSGS_PER_TOPIC, 0)
        max_age = getattr(self._rx_props, RuntimeProperties.CACHE_PROP_MAX_AGE, 0)
        timeout = getattr(self._rx_props, RuntimeProperties.CACHE_PROP_TIMEOUT_IN_MSEC, 10000)

        for topic in topics:
            topic_subscription = TopicSubscription.of(topic)
            request_type: CachedMessageSubscriptionRequest
            is_cached_only = not want_subscribe
            if is_cached_only:
                request_type = CachedMessageSubscriptionRequest.cached_only(cache_name, topic_subscription,
                                                                            timeout, max_msgs_per_topic, max_age)
            elif live_data_action == "FULFILL":
                request_type = CachedMessageSubscriptionRequest.live_cancels_cached(cache_name, topic_subscription,
                                                                                    timeout, max_msgs_per_topic,
                                                                                    max_age)
            elif live_data_action == "QUEUE":
                request_type = CachedMessageSubscriptionRequest.cached_first(cache_name, topic_subscription,
                                                                             timeout, max_msgs_per_topic, max_age)
            elif live_data_action == "FLOW_THRU":
                request_type = CachedMessageSubscriptionRequest.as_available(cache_name, topic_subscription,
                                                                             timeout, max_msgs_per_topic, max_age)
            else:
                raise Exception("Live data action not understood")

            self._cache_request_received = False
            self._default_consumer.consumer.request_cached(request_type, self._request_id, self)
            if wait_for_confirm:
                with self._cache_condition:
                    self._cache_condition.wait_for(self.has_cache_request_been_received)

            self._request_id += 1

    def has_cache_request_been_received(self) -> bool:
        return self._cache_request_received

    def get_client_name(self) -> str:
        if self.service is not None:
            return self.service.get_application_id()
        else:
            return self._client_id_str

    def __str__(self):
        return self._client_id_str

    def _set_publisher_backpressure(self, builder):
        backpressure_list = getattr(self._rx_props, RuntimeProperties.PUBLISHER_BACKPRESSURE_TYPE,
                                    [PublisherBackPressureType.default()])
        backpressure_type = backpressure_list[self._backpressure_index % len(backpressure_list)]
        queue_size = getattr(self._rx_props, RuntimeProperties.PUBLISHER_BACKPRESSURE_BUFFER_SIZE, 10000)
        if backpressure_type == PublisherBackPressureType.ELASTIC:
            builder.on_back_pressure_elastic()
        elif backpressure_type == PublisherBackPressureType.WAIT:
            builder.on_back_pressure_wait(queue_size)
        elif backpressure_type == PublisherBackPressureType.REJECT:
            builder.on_back_pressure_reject(queue_size)
        elif backpressure_type == PublisherBackPressureType.NO:
            builder.on_back_pressure_reject(0)
        else:
            pass

    def _set_subscriber_backpressure(self, builder: DirectMessageReceiverBuilder):
        backpressure_type = getattr(self._rx_props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_TYPE,
                                    SubscriberBackPressureType.default())
        buffer_size = getattr(self._rx_props, RuntimeProperties.SUBSCRIBER_BACKPRESSURE_BUFFER_SIZE, 10000)
        if backpressure_type == SubscriberBackPressureType.ELASTIC:
            builder.on_back_pressure_elastic()
        elif backpressure_type == SubscriberBackPressureType.LATEST:
            builder.on_back_pressure_drop_latest(buffer_size)
        elif backpressure_type == SubscriberBackPressureType.OLDEST:
            builder.on_back_pressure_drop_oldest(buffer_size)
        else:
            pass

    def _apply_solace_properties(self, builder: MessagingServiceClientBuilder):
        broker_props = dict()
        broker_props[HOST] = getattr(self._rx_props, RuntimeProperties.CLIENT_IP_ADDR)
        broker_props[VPN_NAME] = getattr(self._rx_props, RuntimeProperties.CLIENT_VPN, "default")

        keep_alive_interval = getattr(self._rx_props, RuntimeProperties.KEEPALIVE_INTERVAL_MSEC, None)
        if keep_alive_interval is not None:
            if isinstance(keep_alive_interval, int):
                broker_props[KEEP_ALIVE_INTERVAL] = keep_alive_interval
            else:
                logger.warning("Values passed in for -keepAliveInterval should be integers")

        auth_scheme = getattr(self._rx_props, RuntimeProperties.AUTHENTICATION_SCHEME, AuthenticationScheme.BASIC)
        username = getattr(self._rx_props, RuntimeProperties.CLIENT_USERNAME, "")

        if auth_scheme == AuthenticationScheme.GSS_KRB:
            if getattr(self._rx_props, RuntimeProperties.KRB_SERVICE_NAME, None) is None or \
                    getattr(self._rx_props, RuntimeProperties.KRB_SERVICE_NAME) == "":
                auth = Kerberos.default()
            else:
                auth = Kerberos.of(getattr(self._rx_props, RuntimeProperties.KRB_SERVICE_NAME))
            if username != "":
                auth.with_user_name(username)
        elif auth_scheme == AuthenticationScheme.CLIENT_CERTIFICATE:
            auth = ClientCertificateAuthentication.of(
                certificate_file=getattr(self._rx_props, RuntimeProperties.SSL_CLIENT_CERTIFICATE_FILE),
                key_file=getattr(self._rx_props, RuntimeProperties.SSL_CLIENT_PRIVATE_KEY_FILE),
                key_password=getattr(self._rx_props, RuntimeProperties.SSL_CLIENT_PRIVATE_KEY_FILE_PASSWORD, None)
            )
            if username != "":
                auth.with_user_name(username)
        elif auth_scheme == AuthenticationScheme.OAUTH2:
            access_token = getattr(self._rx_props, RuntimeProperties.OAUTH_ACCESS_TOKEN, None)
            id_token = getattr(self._rx_props, RuntimeProperties.OAUTH_ID_TOKEN, None)
            issuer_id = getattr(self._rx_props, RuntimeProperties.OAUTH_ISSUER_ID, None)

            if access_token == "":
                access_token = None
            if id_token == "":
                id_token = None
            if issuer_id == "":
                issuer_id = None

            auth = OAuth2.of(access_token,
                             id_token,
                             issuer_id)
        else:
            if username == "":
                username = self._generate_client_username()
            auth = BasicUserNamePassword.of(username=username,
                                            password=getattr(self._rx_props, RuntimeProperties.CLIENT_PASSWORD, ""))

        if getattr(self._rx_props, RuntimeProperties.SSL_VALIDATE_CERTIFICATE, False):
            tls = TLS().with_certificate_validation(
                ignore_expiration=not getattr(self._rx_props, RuntimeProperties.SSL_VALIDATE_CERTIFICATE_DATE),
                validate_server_name=getattr(self._rx_props, RuntimeProperties.SSL_VALIDATE_CERTIFICATE_HOSTNAME, True),
                trust_store_file_path=getattr(self._rx_props, RuntimeProperties.SSL_TRUST_STORE_DIR, None),
                trusted_common_name_list=getattr(self._rx_props, RuntimeProperties.SSL_TRUSTED_COMMON_NAME_LIST, None)
            )
        else:
            tls = TLS().without_certificate_validation()

        if getattr(self._rx_props, RuntimeProperties.SSL_CIPHER_SUITES, "") != "":
            tls.with_cipher_suites(getattr(self._rx_props, RuntimeProperties.SSL_CIPHER_SUITES))
        if len(getattr(self._rx_props, RuntimeProperties.SSL_EXLCUDED_PROTOCOLS, [])) > 0:
            tls.with_excluded_protocols(*getattr(self._rx_props, RuntimeProperties.SSL_EXLCUDED_PROTOCOLS))
        if getattr(self._rx_props, RuntimeProperties.SSL_CONNECTION_DOWNGRADE_TO, "") != "":
            tls.downgradable()

        if getattr(self._rx_props, RuntimeProperties.PAYLOAD_COMPRESSION_LEVEL, None) is not None:
            broker_props[PAYLOAD_COMPRESSION_LEVEL] = getattr(self._rx_props, RuntimeProperties.PAYLOAD_COMPRESSION_LEVEL)            

        builder.from_properties(broker_props)
        builder.with_authentication_strategy(auth)
        builder.with_transport_security_strategy(tls)

        if getattr(self._rx_props, RuntimeProperties.CLIENT_COMPRESSION_LEVEL, None) is not None:
            builder.with_message_compression(getattr(self._rx_props, RuntimeProperties.CLIENT_COMPRESSION_LEVEL))

        if getattr(self._rx_props, RuntimeProperties.RECONNECT_ATTEMPTS, None) is not None and \
                getattr(self._rx_props, RuntimeProperties.RECONNECT_INTERVAL_MSEC, None) is not None:
            if getattr(self._rx_props, RuntimeProperties.RECONNECT_ATTEMPTS) == -1:
                builder.with_reconnection_retry_strategy(
                    RetryStrategy.forever_retry(getattr(self._rx_props, RuntimeProperties.RECONNECT_INTERVAL_MSEC)))
            else:
                builder.with_reconnection_retry_strategy(
                    RetryStrategy.parametrized_retry(getattr(self._rx_props, RuntimeProperties.RECONNECT_ATTEMPTS),
                                                     getattr(self._rx_props,
                                                             RuntimeProperties.RECONNECT_INTERVAL_MSEC)))

        if getattr(self._rx_props, RuntimeProperties.EXTRA_PROP_DICT, None) is not None:
            builder.from_properties(getattr(self._rx_props, RuntimeProperties.EXTRA_PROP_DICT))

    def _connect_default_flow(self):
        if self._default_consumer is not None:
            return False

        slow_receiver: int = getattr(self._rx_props, RuntimeProperties.RECEIVER_SLEEP_TIME, 0)

        # create receiver and add requested backpressure
        direct_builder: DirectMessageReceiverBuilder = self.service.create_direct_message_receiver_builder()
        self._set_subscriber_backpressure(direct_builder)
        consumer: DirectMessageReceiver = direct_builder.build()
        self._termination_listener.register_for_listen(consumer)

        self._default_consumer = MessageHandlerImpl(self, "", False, None, False, None, False, slow_receiver)

        self._default_consumer.consumer = consumer
        self._default_consumer.consumer.start()
        self._default_consumer.is_started = True

        # Add back any subs that got lost from the last stop
        if self.sub_memory is not None and len(self.sub_memory) > 0:
            for sub in self.sub_memory:
                try:
                    self._default_consumer.consumer.add_subscription(TopicSubscription.of(sub))
                    self._default_consumer.add_subscription(sub)
                except Exception as e:
                    self.set_last_exception(e)
                    logger.error(f"Error adding subscription {sub} {e}")
            self.sub_memory = list()

        if getattr(self._rx_props, RuntimeProperties.WANT_SYNC_RECEIVE, False):
            self._default_consumer.consumer.receive_message()
            if slow_receiver > 0:
                time.sleep(slow_receiver)
        else:
            self._default_consumer.consumer.receive_async(self._default_consumer)
        self._default_consumer.done_receive = True
        return True

    def _reconnect_to_flows(self):
        """ Nothing to do here with python solace API """
        pass

    def _generate_client_username(self) -> str:
        client_str = ""
        if getattr(self._rx_props, RuntimeProperties.CLIENT_USERNAME, "") == "":
            if getattr(self._rx_props, RuntimeProperties.CLIENT_NAME_PREFIX, "") == "":
                client_str += "perf_client"
            else:
                client_str += getattr(self._rx_props, RuntimeProperties.CLIENT_NAME_PREFIX)
            client_str += f"{self._client_id:06}"
        else:
            client_str += getattr(self._rx_props, RuntimeProperties.CLIENT_USERNAME)
        return client_str

    @staticmethod
    def _handle_replay(builder: Union[PersistentMessageReceiverBuilder, TransactionalMessageReceiver],
                       ep_props: EpConfigProperties):
        if ep_props.want_replay:
            if ep_props.want_replay_from_date is not None and ep_props.want_replay_from_date != "":
                logger.info(f"Requesting replay from date: {ep_props.want_replay_from_date}")
                date = ep_props.want_replay_from_date
                if ":" == date[-3] and ("+" == date[-6] or "-" == date[-6]):
                    date = date[:-3] + date[-2:]
                datetime_obj = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S%z')
                builder.with_message_replay(ReplayStrategy.time_based(datetime_obj))
            elif ep_props.want_replay_from_msg_id is not None and ep_props.want_replay_from_msg_id != "":
                logger.info(f"Requesting replay from msg id: {ep_props.want_replay_from_msg_id}")
                msg_id = ep_props.want_replay_from_msg_id
                builder.with_message_replay(ReplayStrategy.replication_group_message_id_based(msg_id))
            else:
                logger.info("Requesting replay from beginning")
                builder.with_message_replay(ReplayStrategy.all_messages())

    def on_reconnected(self, service_event: ServiceEvent):
        logger.debug(f"{self._client_id_str} on_reconnected() called. "
                     f"Cause: {service_event.get_cause()}, Message: {service_event.get_message()}")
        self.channel_state = ChannelState.CLIENT_STATE_CONNECTED

    def on_reconnecting(self, event: ServiceEvent):
        logger.debug(f"{self._client_id_str} on_reconnecting() called. "
                     f"Cause: {event.get_cause()}, Message: {event.get_message()}")
        self.channel_state = ChannelState.CLIENT_STATE_RECONNECTING

    def on_service_interrupted(self, event: ServiceEvent):
        logger.debug(f"{self._client_id_str} on_service_interrupted() called. "
                     f"Cause: {event.get_cause()}, Message: {event.get_message()}")
        if event.get_cause() is not None:
            self.set_last_exception(event.get_cause())
        self.channel_state = ChannelState.CLIENT_STATE_DISCONNECTED

    def on_completion(self, result: CacheRequestOutcome, correlation_id: int, exception: Union[Exception, None]):
        self.stats.inc_stat("numCacheRespRecv")
        if result == CacheRequestOutcome.OK:
            self.stats.inc_stat("numCacheReqOk")
        elif result == CacheRequestOutcome.NO_DATA:
            self.stats.inc_stat("numCacheReqIncompNoData")
        elif result == CacheRequestOutcome.SUSPECT_DATA:
            self.stats.inc_stat("numCacheReqIncompSuspect")
        elif result == CacheRequestOutcome.FAILED:
            if exception is None:
                logger.error("Got CacheRequestOutcome.FAILED but exception is not set")
            else:
                logger.warn(f"Got CacheRequestOutcome.FAILED with exception: {exception}")
                # To keep parity with CCSMP, a timeout shouldn't count as an error even though it triggers failure here
                if type(exception) is PubSubTimeoutError:
                    self.stats.inc_stat("numCacheReqIncompTimeout")
                else:
                    self.stats.inc_stat("numCacheReqError")

        if result != CacheRequestOutcome.FAILED and exception is not None:
            logger.error(f"Got {result} but exception is set: {exception}")

        with self._cache_condition:
            self._cache_request_received = True
            self._cache_condition.notify()
        logger.debug(f"Cache Request Completed {result}, {correlation_id}, {exception}")


class MessagePublishReceiptListenerImpl(MessagePublishReceiptListener):

    def __init__(self, stats, parent: AbstractClient):
        self.stats = stats
        self.parent = parent

    def on_publish_receipt(self, publish_receipt: PublishReceipt):
        if publish_receipt.exception is None:
            self.stats.inc_stat('adPubSuccess')
            self.stats.inc_stat('numAckEvents')
        else:
            self.stats.inc_stat('adPubFail')
            self.parent.set_last_exception(publish_receipt.exception)
            logger.warn(f"Got AD publish failure: {publish_receipt.exception}")


class PublishFailureListenerImpl(PublishFailureListener):

    def __init__(self, stats):
        self.stats = stats

    def on_failed_publish(self, failed_publish_event: FailedPublishEvent):
        self.stats.inc_stat('directPubFail')
        logger.warn(f"Got Direct publish failure: {failed_publish_event.get_exception()}")


class ActivationListener(ReceiverStateChangeListener):
    def __init__(self, status):
        super()
        self.flow_status = status

    def on_change(self, old_state, new_state, change_time_stamp):
        if new_state == ReceiverState.ACTIVE:
            self.flow_status.is_up = True
            self.flow_status.afi_state = ActiveFlowIndicationType.ACTIVE
        elif new_state == ReceiverState.PASSIVE:
            self.flow_status.is_up = False
            self.flow_status.afi_state = ActiveFlowIndicationType.INACTIVE
        else:
            self.flow_status.is_up = False
            self.flow_status.afi_state = ActiveFlowIndicationType.UNKNOWN
