import time
from threading import Thread

from solace.messaging.messaging_service import TransactionalMessagingService
from solace.messaging.publisher.transactional_message_publisher import TransactionalMessagePublisher

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.core.abstract_client import AbstractClient

logger = get_logger()


class PythonClientTransactedSession:
    def __init__(self, client: AbstractClient, transacted_session: TransactionalMessagingService,
                 producer: TransactionalMessagePublisher, props: RuntimeProperties):
        super().__init__()
        self._client = client
        self._transacted_session = transacted_session
        self._producer = producer
        self._props = RuntimeProperties()
        self._props.__dict__.update(props.__dict__)

        self._transaction_size = getattr(self._props, RuntimeProperties.AD_TRANSACTION_SIZE, 0)
        self._rollback_interval = getattr(self._props, RuntimeProperties.AD_TRANSACTION_ROLLBACK_INTERVAL, -1)
        self._num_msgs_recv = 0
        self._prev_num_msgs_recv = 0
        self._num_msgs_on_commit = 0
        self._num_commits = 0

    def get_producer(self) -> TransactionalMessagePublisher:
        if not self._producer.is_running():
            self._producer.start()
        return self._producer

    def get_session(self) -> TransactionalMessagingService:
        return self._transacted_session

    def close_pub(self):
        try:
            self._producer.terminate(0)
        except Exception as e:
            logger.error(f"Terminating transacted producer; _client={self._client.get_client_name()}, exception={e}")

    def disconnect(self):
        try:
            self._transacted_session.disconnect()
        except Exception as e:
            logger.error(f"Disconnecting transacted service; _client={self._client.get_client_name()}, exception={e}")

    def on_message(self):
        self._num_msgs_recv += 1

        if self._transaction_size > 0 and self._num_msgs_recv > 0 and self._num_msgs_recv % self._transaction_size == 0:
            self._num_commits += 1
            if self._rollback_interval > 0 and self._num_commits % self._rollback_interval == 0:
                self.commit(rollback=True)
            else:
                self.commit()

    def commit(self, rollback: bool = False):
        self._num_msgs_on_commit = self._num_msgs_recv
        if rollback:
            self._transacted_session.rollback()
        else:
            self._transacted_session.commit()

