import binascii
import time

from solace.messaging.config.solace_properties.message_properties import PERSISTENT_DMQ_ELIGIBLE, ELIDING_ELIGIBLE, \
    PRIORITY, PERSISTENT_TIME_TO_LIVE, SEQUENCE_NUMBER, CLASS_OF_SERVICE, APPLICATION_MESSAGE_ID, HTTP_CONTENT_TYPE, \
    HTTP_CONTENT_ENCODING, PERSISTENT_ACK_IMMEDIATELY
from solace.messaging.utils.manageable import Metric

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.constants import Constants
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.core.abstract_client_collection import AbstractClientCollection
from sdkperf.core.cache_thread import CacheThread
from sdkperf.core.sdkperf_pub_thread import SdkPerfPubThread
from sdkperf.core.tool_info import ToolData

logger = get_logger()


class SdkPerfClientCollection(AbstractClientCollection):

    def __init__(self, props: RuntimeProperties):
        super().__init__(props)

        self._pub_threads = list()
        self._pubs_per_session = getattr(self._rprops, RuntimeProperties.NUM_PUBS_PER_SESSION, 1)

        self.stop_cache_required = False
        self._cache_threads = list()

        for i in range(len(self._clients)):
            client = self._clients[i]
            for p in range(self._pubs_per_session):
                pub_thread = SdkPerfPubThread(client, p)
                self._pub_threads.append(pub_thread)
            self._cache_threads.append(CacheThread(client))

    def _get_pub_start_time_in_ns(self):
        start_time = 0
        for pub in self._pub_threads:
            if pub.start_time != 0:
                if start_time == 0:
                    start_time = pub.start_time
                elif pub.start_time < start_time:
                    start_time = pub.start_time
        return start_time

    def _get_pub_end_time_in_ns(self):
        end_time = 0
        for pub in self._pub_threads:
            if pub.end_time != 0:
                if end_time == 0 or pub.end_time < end_time:
                    end_time = pub.end_time
        return end_time

    def get_tx_throughput(self):
        start = self._get_pub_start_time_in_ns()
        end = self._get_pub_end_time_in_ns()
        msg_pub = self.get_sdk_stat(Metric.TOTAL_MESSAGES_SENT, Constants.ALL_CLIENT_INDEX)
        diff = end - start
        if diff <= 0:
            return 0
        else:
            return (msg_pub * Constants.SEC_TO_NANOSEC) / diff

    def get_perf_stat(self, stat, index):
        val = 0
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                val += client.stats.get_stat(stat)
        else:
            client = self._clients[index]
            val += client.stats.get_stat(stat)
        return val

    def get_sdk_stat(self, metric, index):
        val = 0
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                x = client.get_metric(metric)
                val += x if x is not None else 0
        else:
            client = self._clients[index]
            x = client.get_metric(metric)
            val += x if x is not None else 0
        return val

    def clear_stats(self, index):
        if index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                try:
                    client.clear_metrics()
                    client.stats.reset_stats()
                except Exception as e:
                    logger.error(f"Could not clear stats for client {client.get_id_str()} {str(e)}")
        else:
            client = self._clients[index]
            try:
                client.clear_metrics()
                client.stats.reset_stats()
            except Exception as e:
                logger.error(f"Could not clear stats for client {client.get_id_str()} {str(e)}")

    def publish_message(self, pubprops: RuntimeProperties, index):
        if self.is_publishing():
            raise Exception("Already publishing, must call stopPublishing before publish message.")
        if getattr(pubprops, RuntimeProperties.STRUCT_DATA_MSGS_LIST, None) is not None and (
                getattr(pubprops, RuntimeProperties.WANT_ORDER_CHECK, False) or
                getattr(pubprops, RuntimeProperties.WANT_LATENCY, False) or
                getattr(pubprops, RuntimeProperties.WANT_PAYLOAD_CHECK, False) or
                getattr(pubprops, RuntimeProperties.WANT_LOSS_AND_DUPLICATE_DETECTION, False)):
            raise Exception(
                "Cannot use order check, loss/duplicate detection, latency or payload check with structured messages")

        destinations = list()
        topics = getattr(pubprops, RuntimeProperties.PUBLISH_TOPIC_LIST, None)
        queues = getattr(pubprops, RuntimeProperties.PUBLISH_QUEUE_LIST, None)
        if topics is not None:
            destinations += topics
        if queues is not None:
            destinations += queues
        doc_list = SdkPerfClientCollection._create_doc_list(pubprops)
        send_type = getattr(pubprops, RuntimeProperties.PUB_MESSAGE_TYPE, GenericMessageDeliveryMode.DIRECT)
        priority_gen = getattr(pubprops, RuntimeProperties.PUBLISH_MESSAGE_PRIORITY, None)
        transacted_session_name_list = getattr(pubprops, RuntimeProperties.TRANSACTED_SESSION_NAME_LIST, [])
        props = dict()
        if getattr(pubprops, RuntimeProperties.WANT_PAYLOAD_CHECK, False):
            # The crc32 algorithm will always return an unsigned integer in Python3+, but all the
            # other APIs use a signed int32. If the crc32 algorithm produces an integer which requires
            # more than 32 bits, we need to manually mimic overflow behaviour.
            crc_val = binascii.crc32(doc_list[0]) & 0xffffffff
            crc_val = crc_val | (-(crc_val & 0x80000000))
            props[ToolData.PROP_BIN_PAYLOAD_HASH] = crc_val

        if getattr(pubprops, RuntimeProperties.HTTP_CONTENT_ENCODING, None) is not None:
            props[HTTP_CONTENT_ENCODING] = getattr(pubprops, RuntimeProperties.HTTP_CONTENT_ENCODING)
        if getattr(pubprops, RuntimeProperties.HTTP_CONTENT_TYPE, None) is not None:
            props[HTTP_CONTENT_TYPE] = getattr(pubprops, RuntimeProperties.HTTP_CONTENT_TYPE)

        if getattr(pubprops, RuntimeProperties.MSG_DMQ_ELIGIBLE, None) is not None:
            props[PERSISTENT_DMQ_ELIGIBLE] = getattr(pubprops, RuntimeProperties.MSG_DMQ_ELIGIBLE, True)
        if getattr(pubprops, RuntimeProperties.MSG_ELIDING_ELIGIBLE, None) is not None:
            props[ELIDING_ELIGIBLE] = getattr(pubprops, RuntimeProperties.MSG_ELIDING_ELIGIBLE, False)
        if getattr(pubprops, RuntimeProperties.MESSAGE_ID, "") != "":
            props[APPLICATION_MESSAGE_ID] = int(getattr(pubprops, RuntimeProperties.MESSAGE_ID))
        if int(getattr(pubprops, RuntimeProperties.MSG_TTL, 0)) > 0:
            props[PERSISTENT_TIME_TO_LIVE] = int(getattr(pubprops, RuntimeProperties.MSG_TTL))
        if int(getattr(pubprops, RuntimeProperties.PUBLISH_COS, -1)) >= 0:
            props[CLASS_OF_SERVICE] = int(getattr(pubprops, RuntimeProperties.PUBLISH_COS)) - 1
        if int(getattr(pubprops, RuntimeProperties.AD_ACK_IMMEDIATELY_INTERVAL, 0)) > 0:
            props[PERSISTENT_ACK_IMMEDIATELY] = True
        if getattr(pubprops, RuntimeProperties.WANT_ORDER_CHECK, False):
            msg_id = int(1 + getattr(pubprops, RuntimeProperties.PUB_ORDER_OFFSET, 0))
            props[ToolData.PROP_MESSAGE_ID] = msg_id
            props[SEQUENCE_NUMBER] = msg_id

        if send_type == GenericMessageDeliveryMode.PERSISTENT:
            props[PERSISTENT_DMQ_ELIGIBLE] = getattr(pubprops, RuntimeProperties.MSG_DMQ_ELIGIBLE, True)
            props[ELIDING_ELIGIBLE] = getattr(pubprops, RuntimeProperties.MSG_ELIDING_ELIGIBLE, False)
            if priority_gen is not None:
                props[PRIORITY] = priority_gen.get_random_priority()
            if int(getattr(pubprops, RuntimeProperties.MSG_TTL, 0)) > 0:
                props[PERSISTENT_TIME_TO_LIVE] = int(getattr(pubprops, RuntimeProperties.MSG_TTL))

        if getattr(pubprops, RuntimeProperties.CUSTOM_PROPERTIES_LIST, None) is not None:
            props = {**props, **getattr(pubprops, RuntimeProperties.CUSTOM_PROPERTIES_LIST)}

        if index == Constants.ALL_CLIENT_INDEX:
            if len(transacted_session_name_list) > 0 and len(transacted_session_name_list) != len(self._clients):
                raise Exception("Error trying to publishing using transacted sessions.  "
                                "Session names list not equal to num clients")
            for i in range(len(self._clients)):
                transacted_session_name = ""
                if len(transacted_session_name_list) > 0:
                    transacted_session_name = transacted_session_name_list[i]
                self._clients[i].connect()
                self._clients[i].publish_msg(doc_list[0], destinations[0], 0, props if len(props) > 0 else None,
                                             send_type, transacted_session_name=transacted_session_name)
        else:
            if len(transacted_session_name_list) > 0 and len(transacted_session_name_list) != 1:
                raise Exception("Error trying to publishing using transacted sessions.  "
                                "Session names size 1 when specific client chosen")
            transacted_session_name = ""
            if len(transacted_session_name_list) > 0:
                transacted_session_name = transacted_session_name_list[0]
            self._clients[index].connect()
            self._clients[index].publish_msg(doc_list[0], destinations[0], 0, props if len(props) > 0 else None,
                                             send_type, transacted_session_name=transacted_session_name)

    def start_publishing(self, pubprops: RuntimeProperties):
        if self.is_publishing():
            raise Exception("Already publishing, must call stop_publishing before starting again")

        doc_list = SdkPerfClientCollection._create_doc_list(pubprops)
        num_total_msg = float(getattr(pubprops, RuntimeProperties.NUM_MSGS_TO_PUBLISH, 0))

        for client in self._clients:
            client._num_msg_pub = num_total_msg
            client._num_msg_pub_offset = 1
            client._is_start_pub = True

        num_msg_per_pub = num_total_msg / self._num_clients / self._pubs_per_session
        for pub in self._pub_threads:
            pub.start_publish(doc_list, num_msg_per_pub, pubprops)

    def stop_publishing(self):
        for pub in self._pub_threads:
            pub.stop_publish()

        guard = 1000
        while self.is_publishing() and guard > 0:
            guard -= 1
            time.sleep(0.2)

        if self.is_publishing():
            logger.error("Failed to stop all publishers")
            raise Exception("Failed to stop all publishers")

    def is_publishing(self):
        for pub in self._pub_threads:
            if pub is not None and pub.is_publishing():
                return True
        return False

    def start_cache_requesting(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool,
                               num_msgs: int, cache_rate: int) -> None:
        logger.info("Starting Cache Requesting")

        if self.is_cache_requesting():
            raise Exception("Already Cache Requesting. Must stop before calling start Cache Requesting")

        if self.stop_cache_required:
            raise Exception("Threads currently stopped. However, mut call stop cache requesting to be able to restart "
                            "requesting")

        try:
            for cache_thread in self._cache_threads:
                cache_thread.start_cache_requesting(topics, want_subscribe, live_data_action, wait_for_confirm, num_msgs, cache_rate)
        except Exception as e:
            raise e

        guard = 1800  # Wait up to 3 minutes
        while not self.is_cache_requesting() and not self.is_done_cache_requesting() and guard != 0:
            time.sleep(0.1)
            guard -= 1

        if guard == 0:
            raise Exception("Cache requesting did not start within 3 min.")

        logger.info("Started Cache Requesting")
        self.stop_cache_required = True

    def stop_cache_requesting(self) -> None:
        if not self.is_cache_requesting() and not self.stop_cache_required:
            logger.debug("Cache requesting already stopped. Returning")
            return

        logger.info("Stopping Cache Threads")

        self.stop_cache_required = False

        try:
            for cache_thread in self._cache_threads:
                cache_thread.stop_cache_requesting()
        except Exception as e:
            raise e

        guard = 1800  # Wait up to 3 minutes
        while self.is_cache_requesting() and guard != 0:
            time.sleep(0.1)
            guard -= 1

        if guard == 0:
            raise Exception("Cache requesting did not stop within 3 min.")

        logger.info("Stopped Cache Threads")

    def is_cache_requesting(self) -> bool:
        is_cacheing = False
        for cache_thread in self._cache_threads:
            if cache_thread.is_cache_requesting():
                is_cacheing = True
        return is_cacheing

    def is_done_cache_requesting(self) -> bool:
        is_done = True
        for cache_thread in self._cache_threads:
            if not cache_thread.is_done_cache_requesting():
                is_done = False
        return is_done

    def cache_request(self, topics: list, want_subscribe: bool, live_data_action: str, wait_for_confirm: bool,
                      client_index: int) -> None:
        if client_index == Constants.ALL_CLIENT_INDEX:
            for client in self._clients:
                client.cache_request(topics, want_subscribe, live_data_action, wait_for_confirm)
        else:
            self._clients[client_index].cache_request(topics, want_subscribe, live_data_action, wait_for_confirm)

    @staticmethod
    def _create_doc_list(props: RuntimeProperties):
        message_sizes = getattr(props, RuntimeProperties.ATTACHMENT_SIZE_LIST, [])
        file_strs = getattr(props, RuntimeProperties.ATTACHMENT_FILE_LIST, [])
        xml_strs = getattr(props, RuntimeProperties.XML_FILE_LIST, [])
        sdm_props = getattr(props, RuntimeProperties.SDM_PROP_LIST, [])

        # Use SDM to create our docs has some weird edge cases, this all_others_empty will help
        # determine if we are just doing SDM work or mixing the docs
        all_others_empty = len(getattr(props, RuntimeProperties.ATTACHMENT_SIZE_LIST, [])) == 0 and \
                           len(getattr(props, RuntimeProperties.ATTACHMENT_FILE_LIST, [])) == 0 and \
                           len(getattr(props, RuntimeProperties.XML_FILE_LIST, [])) == 0

        doc_list = list()
        for prop in sdm_props:
            if all_others_empty or prop.get_payload() != bytearray():
                doc_list.append(prop.get_payload())
        if len(doc_list) == 0:
            for size in message_sizes:
                doc_list.append(SdkPerfClientCollection._generate_doc(size))
            for byte_arr in file_strs:
                doc_list.append(byte_arr)
            for string in xml_strs:
                b = bytearray()
                b.extend(map(ord, str(string)))
                doc_list.append(b)

        if len(doc_list) == 0:
            doc_list = [bytearray(0)]

        return doc_list

    @staticmethod
    def _generate_doc(size):
        buf = ""
        for i in range(0, size // 8):
            buf += f":{i:07x}"
        remain = size - len(buf)
        for _ in range(remain):
            buf += "A"
        b = bytearray()
        b.extend(map(ord, str(buf)))
        return b
