import time
from threading import Thread

from solace.messaging.config.solace_constants.message_user_property_constants import QUEUE_PARTITION_KEY
from solace.messaging.config.solace_properties.message_properties import SEQUENCE_NUMBER, PRIORITY, \
    PERSISTENT_ACK_IMMEDIATELY
from solace.messaging.errors.pubsubplus_client_error import PublisherOverflowError
from solace.messaging.publisher.outbound_message import OutboundMessageBuilder

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.config.message_object_type import MessageObjectType
from sdkperf.core.abstract_client import AbstractClient
from sdkperf.core.msg_rep import MessageRep
from sdkperf.core.tool_info import ToolData

logger = get_logger()


class SdkPerfPubThread:
    def __init__(self, client: AbstractClient, pub_index: int):
        self.client = client
        self.pub_per_session_index = pub_index
        self.order_offset = 0
        self._pub_thread = None
        self.start_time = 0
        self.end_time = 0

    def is_publishing(self):
        if self._pub_thread is None:
            return False
        return self._pub_thread.is_publishing

    def stop_publish(self):
        if self._pub_thread is not None:
            self._pub_thread.shutdown = True
            self._pub_thread.join(1)

    def start_publish(self, doc_list, num_msg, pubprops):
        self._num_msg_pub = num_msg
        # We have to clear/reset the dictionaries on each start_publishing in case
        # there are props that are absent in the next call which wouldn't get removed
        self._perf_props = RuntimeProperties()
        self._perf_props.__dict__.update(pubprops.__dict__)
        self.client._tx_props = RuntimeProperties()
        self.client._tx_props.__dict__.update(pubprops.__dict__)

        if self._pub_thread is not None:
            if self._pub_thread.is_alive():
                raise Exception("Attempted to start a publisher thread when one was already running.")

        self._pub_thread = PubThreadObj(self, num_msg, doc_list, pubprops)
        self._pub_thread.daemon = True
        self._pub_thread.start()

    def get_topics(self):
        pub_topic_list = getattr(self._perf_props, RuntimeProperties.PUBLISH_TOPIC_LIST, None)
        if pub_topic_list is not None and len(pub_topic_list) > 0:
            return pub_topic_list
        else:
            return None

    def get_queues(self):
        pub_queue_list = getattr(self._perf_props, RuntimeProperties.PUBLISH_QUEUE_LIST, None)
        if pub_queue_list is not None and len(pub_queue_list) > 0:
            return pub_queue_list
        else:
            return None


class PubThreadObj(Thread):
    def __init__(self, parent: SdkPerfPubThread, num_msg_pub: int, doc_list: list, props: RuntimeProperties):
        super().__init__()
        self.parent = parent
        self.num_msg_pub = num_msg_pub
        self.doc_list = doc_list
        self.props = RuntimeProperties()
        self.props.__dict__.update(props.__dict__)

        self.shutdown = False
        self.is_publishing = False

    def run(self):
        self.is_publishing = True
        destinations = list()
        topics = self.parent.get_topics()
        queues = self.parent.get_queues()
        if topics is not None:
            destinations += topics
        if queues is not None:
            destinations += queues
        send_type = getattr(self.props, RuntimeProperties.PUB_MESSAGE_TYPE, GenericMessageDeliveryMode.DIRECT)
        msg_builder: OutboundMessageBuilder = self.parent.client.service.message_builder()
        session_name_list = getattr(self.props, RuntimeProperties.TRANSACTED_SESSION_NAME_LIST, [])
        transacted_session_name = ""
        if len(session_name_list) == 1:
            transacted_session_name = session_name_list[0]
        elif len(session_name_list) > 1:
            transacted_session_name = session_name_list[self.parent.pub_per_session_index % len(session_name_list)]

        transaction_size = getattr(self.props, RuntimeProperties.AD_TRANSACTION_SIZE, 0)
        rollback_interval = getattr(self.props, RuntimeProperties.AD_TRANSACTION_ROLLBACK_INTERVAL, -1)
        skip_final_commit = getattr(self.props, RuntimeProperties.SKIP_FINAL_COMMIT, False)
        num_commits = 0

        fast_path_msgs = list()
        has_sdm_props = False
        if len(getattr(self.props, RuntimeProperties.SDM_PROP_LIST, [])) > 0:
            has_sdm_props = True
            for sdm_prop in getattr(self.props, RuntimeProperties.SDM_PROP_LIST):
                fast_path_msgs.append(MessageRep(msg_builder, self.props, self.parent.client._client_id, None, sdm_prop))
        else:
            for content in self.doc_list:
                fast_path_msgs.append(MessageRep(msg_builder, self.props, self.parent.client._client_id, content, None))

        msg_published = 0
        starting_offset = self.parent.order_offset
        if starting_offset < 0:
            starting_offset = 0
        msg_rate = getattr(self.props, RuntimeProperties.PUBLISH_RATE_PER_PUB, 1)

        rate_check_threshold = 1000
        single_msg_period_in_ns = 0
        cycles_per_msg = 0
        if msg_rate > 0:
            single_msg_period_in_ns = 1000000000 / msg_rate
            cycles_per_msg = 1000000 / msg_rate
        time_waited_in_ns = 0

        next_time_in_us = 0
        psm_max_cycles_behind = 0
        burst_duration = getattr(self.props, RuntimeProperties.BURST_DURATION, 0.0)
        inter_burst_duration = getattr(self.props, RuntimeProperties.INTER_BURST_DURATION, 0.0)
        first_sampled = False

        want_smooth_pub_calc_latency = getattr(self.props, RuntimeProperties.WANT_SMOOTH_PUB_CALC_LAT, False)
        burst_mode = burst_duration > 0 and inter_burst_duration > 0
        msg_rate_is_max = getattr(self.props, RuntimeProperties.MSG_RATE_IS_MAX, False)

        stop_on_error = self.parent.client.stop_on_error
        ack_immediately_interval = getattr(self.props, RuntimeProperties.AD_ACK_IMMEDIATELY_INTERVAL, 0)
        priority_gen = getattr(self.props, RuntimeProperties.PUBLISH_MESSAGE_PRIORITY, None)
        want_await = getattr(self.props, RuntimeProperties.WANT_PUBLISH_AWAIT_ACK, False)
        request_reply = getattr(self.props, RuntimeProperties.WANT_REPLY_TOPIC, False)
        partition_key_list = getattr(self.props, RuntimeProperties.PARTITION_KEY_LIST, None)
        partition_key_index = 0
        msg_obj_type_list = getattr(self.props, RuntimeProperties.MSG_OBJ_TYPE_LIST, None)
        msg_obj_type_index = 0

        start_time_us = time.time_ns() / 1000
        next_pub_time_in_us = start_time_us
        burst_start_time_in_us = start_time_us
        pub_start_time_in_ns = start_time_us * 1000

        self.parent.start_time = pub_start_time_in_ns
        logger.debug(f"Starting publish to destinations {destinations}")
        while msg_published < self.num_msg_pub and not self.shutdown:
            if len(destinations) == 0 or (
                    len(destinations) != 1 and self.parent.pub_per_session_index > len(destinations) - 1):
                logger.error(f"ERROR IN PUB THREAD: Trying to publish to invalid destination {destinations}")
                self.is_publishing = False
                break
            if len(destinations) == 1:
                dest = destinations[0]
            else:
                dest = destinations[msg_published % len(destinations)]

            msg_obj_type = MessageObjectType.NONE
            if msg_obj_type_list is not None:
                msg_obj_type = msg_obj_type_list[msg_obj_type_index % len(msg_obj_type_list)]
                msg_obj_type_index += 1

            if want_smooth_pub_calc_latency:
                next_pub_time_in_us += cycles_per_msg

                current_time_in_us = time.time_ns() / 1000
                if next_time_in_us < current_time_in_us:
                    #TODO: Increment times behind stat
                    if (current_time_in_us - next_pub_time_in_us) > psm_max_cycles_behind:
                        psm_max_cycles_behind = current_time_in_us - next_pub_time_in_us
                while next_pub_time_in_us > current_time_in_us:
                    current_time_in_us = time.time_ns() / 1000
                lat_timestamp = next_pub_time_in_us * 1000
            else:
                if burst_mode:
                    if ((time.time_ns() / 1000) - burst_start_time_in_us) / 1000000 >= burst_duration:
                        time.sleep(inter_burst_duration)
                        pub_start_time_in_ns = time.time_ns() - (single_msg_period_in_ns * msg_published)
                        first_sampled = True
                if single_msg_period_in_ns > 0:
                    if (msg_rate > rate_check_threshold and msg_published % rate_check_threshold == 0) \
                            or msg_rate <= rate_check_threshold:
                        calc_time_since_start_in_ns = single_msg_period_in_ns * msg_published
                        next_time_in_ns = pub_start_time_in_ns + calc_time_since_start_in_ns

                        time_waited_in_ns = next_time_in_ns - time.time_ns()
                        if time_waited_in_ns > 0:
                            time.sleep(time_waited_in_ns / 1000000000)
                    if msg_rate_is_max and time_waited_in_ns < -1*200000000:
                        pub_start_time_in_ns = time.time_ns() - (single_msg_period_in_ns * msg_published)
                lat_timestamp = time.time_ns()

            pri = -1
            if priority_gen is not None:
                pri = priority_gen.get_random_priority()
            try:
                additional_props = dict()
                # We don't want to add extra props if using SDM as they could overlap
                if not has_sdm_props:
                    if getattr(self.props, RuntimeProperties.WANT_LATENCY, False):
                        additional_props[ToolData.PROP_LATENCY] = int(lat_timestamp)
                    if getattr(self.props, RuntimeProperties.WANT_ORDER_CHECK, False):
                        additional_props[ToolData.PROP_MESSAGE_ID] = int(msg_published + starting_offset)
                        additional_props[SEQUENCE_NUMBER] = int(msg_published + starting_offset)
                    if ack_immediately_interval > 0:
                        if (msg_published + 1) % ack_immediately_interval == 0:
                            additional_props[PERSISTENT_ACK_IMMEDIATELY] = True
                        else:
                            additional_props[PERSISTENT_ACK_IMMEDIATELY] = False
                    if pri != -1:
                        additional_props[PRIORITY] = pri
                    if partition_key_list is not None:
                        additional_props[QUEUE_PARTITION_KEY] = partition_key_list[partition_key_index % len(partition_key_list)]
                        partition_key_index += 1

                if burst_mode and first_sampled:
                    first_sampled = False
                    burst_start_time_in_us = time.time_ns() / 1000

                self.parent.client.publish_msg(fast_path_msgs[msg_published % len(fast_path_msgs)],
                                               dest,
                                               self.parent.pub_per_session_index,
                                               additional_props,
                                               send_type,
                                               msg_obj_type,
                                               want_await,
                                               request_reply,
                                               transacted_session_name)

                msg_published += 1
                self.parent.order_offset += 1

                if pri != -1:
                    self.parent.client.get_rx_stats().message_priority_stats_inc(pri, False)

            except PublisherOverflowError as poe:
                logger.debug(f"Caught publisher overflow error, {poe}")
                if not self.parent.client.is_ready_to_publish(send_type):
                    time.sleep(1)
                continue
            except Exception as e:
                if not self.shutdown:
                    logger.error(f"Exception caught while publishing {str(e)}")
                    time.sleep(1)
                    if stop_on_error:
                        self.is_publishing = False
                        raise e
                else:
                    logger.debug(f"Caught shutdown exception {e}")
                    self.is_publishing = False
                    break

            if transaction_size > 0 and msg_published % transaction_size == 0:
                num_commits += 1
                want_rollback = False
                if rollback_interval > 0 and num_commits % rollback_interval == 0:
                    want_rollback = True

                try:
                    self.parent.client.commit_transaction_on_current_pub(transacted_session_name, want_rollback)
                except Exception as e:
                    num_commits -= 1
                    msg_published -= transaction_size
                    logger.info(f"Pub received exception on commit: {e}")

        if transaction_size > 0 and not skip_final_commit:
            num_commits += 1
            want_rollback = False
            if rollback_interval > 0 and num_commits % rollback_interval == 0:
                want_rollback = True

            try:
                self.parent.client.commit_transaction_on_current_pub(transacted_session_name, want_rollback)
            except Exception as e:
                logger.info(f"Pub received exception on final commit: {e}")

        logger.debug(f"Done publish, {msg_published} messages published")
        self.is_publishing = False
        self.parent.end_time = time.time_ns()
        return
