import re
from typing import Union

from solace.messaging.publisher.outbound_message import OutboundMessageBuilder, OutboundMessage
from solace.messaging.receiver.inbound_message import InboundMessage

from pypubsubtools.utils.log import get_logger

MSG_REGEX = re.compile(r'^msg:\s(\w*?)=(.*)$')
HDR_REGEX = re.compile(r'^hdr:\s(\w*?)=(.*)$')
MAP_REGEX = re.compile(r'^\s*?(\w*?)\s*?name=(.*?)\s*?value=(.*)\s*?$')
STREAM_REGEX = re.compile(r'^\s*?(\w*?)\s*?value=(.*)\s*?$')

MSG_PROP_KEY_TYPE = "type"
MSG_PROP_KEY_LOCATION = "useBinaryAttachment"

MSG_PROP_VALUE_STREAM = "stream"
MSG_PROP_VALUE_MAP = "map"

HDR_PROP_KEY_REPLY_TO = "replyTo"
HDR_PROP_KEY_CORRELATION_ID = "correlationId"
HDR_PROP_KEY_MESSAGE_ID = "messageId"
HDR_PROP_KEY_MESSAGE_TYPE = "messageType"

HDR_PROP_KEY_TIMESTAMP = "timestamp"
HDR_PROP_KEY_TOPICSEQUENCE = "topicSequence"
HDR_PROP_KEY_USERDATA = "userData"

C_FIELD_TYPE_BOOL = "bool"
C_FIELD_TYPE_STRING = "string"
C_FIELD_TYPE_FLOAT = "float"
C_FIELD_TYPE_DOUBLE = "double"

C_SHORT_INT_MAX = 32767
C_INT_MAX = 2147483648
C_LONG_LONG_INT_MAX = 9223372036854775807
C_FLOATING_POINT_SUBNORMAL_BOUND = 1.401e-45
C_FLOATING_POINT_NORMAL_BOUND = 1.175e-38

logger = get_logger()


class SdmProperties:
    def __init__(self):
        self._reset_all()

    def _reset_all(self):
        self.is_map_msg = False
        self.use_binary_attachment = False
        self.container = list()

        self.reply_to = None
        self.correlation_id = None
        self.message_id = None
        self.message_type = None
        self.timestamp = None
        self.topic_sequence = None
        self.user_data = None

    def parse_input_file(self, file_name: str):
        with open(file_name, 'r') as sdt_file:
            self.parse_input_lines(sdt_file.readlines())

    def parse_input_lines(self, lines: list):
        for line in lines:
            # Skip over empty lines and comments
            if line == "" or line.isspace() or line[0] == "#":
                continue
            if "msg:" in line:
                matches = MSG_REGEX.findall(line)[0]
                key = matches[0]
                value = matches[1]
                if key == MSG_PROP_KEY_TYPE:
                    if value == MSG_PROP_VALUE_MAP:
                        self.is_map_msg = True
                        self.container = dict()
                elif key == MSG_PROP_KEY_LOCATION:
                    if "0" not in value:
                        self.use_binary_attachment = True
                else:
                    self._reset_all()
                    raise Exception(f"Unknown msg key {key}")
            elif "hdr:" in line:
                # These properties are currently not checked in process_message()
                matches = HDR_REGEX.findall(line)[0]
                if matches[0] == HDR_PROP_KEY_REPLY_TO:
                    self.reply_to = matches[1]
                elif matches[0] == HDR_PROP_KEY_CORRELATION_ID:
                    self.correlation_id = matches[1]
                elif matches[0] == HDR_PROP_KEY_MESSAGE_ID:
                    self.message_id = matches[1]
                elif matches[0] == HDR_PROP_KEY_MESSAGE_TYPE:
                    self.message_type = matches[1]
                elif matches[0] == HDR_PROP_KEY_TIMESTAMP:
                    self.timestamp = int(matches[1])
                elif matches[0] == HDR_PROP_KEY_TOPICSEQUENCE:
                    self.topic_sequence = int(matches[1])
                elif matches[0] == HDR_PROP_KEY_USERDATA:
                    self.user_data = matches[1]
                else:
                    self._reset_all()
                    raise Exception(f"Unknown hdr key {matches[0]}")
            else:
                if self.is_map_msg:
                    matches = MAP_REGEX.findall(line)[0]
                    name = matches[1]
                    value = matches[2]
                else:
                    matches = STREAM_REGEX.findall(line)[0]
                    value = matches[1]
                val_type = matches[0]

                if val_type == C_FIELD_TYPE_BOOL:
                    temp_bool = True
                    if "0" in value:
                        temp_bool = False
                    if self.is_map_msg:
                        self.container[name] = temp_bool
                    else:
                        self.container.append(temp_bool)
                elif "int" in val_type:  # Catches all the int types which are the same in python
                    if self.is_map_msg:
                        self.container[name] = int(value)
                    else:
                        self.container.append(int(value))
                elif val_type == C_FIELD_TYPE_STRING:
                    if self.is_map_msg:
                        self.container[name] = value.strip()
                    else:
                        self.container.append(value.strip())
                elif val_type == C_FIELD_TYPE_FLOAT or val_type == C_FIELD_TYPE_DOUBLE:
                    if self.is_map_msg:
                        self.container[name] = float(value)
                    else:
                        self.container.append(float(value))
                else:
                    self._reset_all()
                    raise Exception(f"Unknown SDM type {val_type}")

    def get_payload(self) -> Union[dict, list, bytearray]:
        if self.use_binary_attachment:
            return self.container
        else:
            return bytearray()

    def get_properties(self) -> Union[dict, list, None]:
        if self.use_binary_attachment:
            return {}
        else:
            return self.container

    def build_msg(self, builder: OutboundMessageBuilder) -> OutboundMessage:
        # Cannot set any of the other fields here in python
        if self.message_id is not None:
            builder.with_application_message_id(self.message_id)
        return builder.build(self.get_payload(), self.get_properties())

    def check_message(self, msg: InboundMessage):
        if self.is_map_msg:
            if self.use_binary_attachment:
                if msg.get_payload_as_list() is not None and len(msg.get_payload_as_list()) > 0:
                    logger.debug("SDM Check: Found unexpected Attachment Stream header")
                    return False
                if msg.get_payload_as_dictionary() is None:
                    logger.debug("SDM Check: Did not find expected attachment map")
                    return False
            else:
                if msg.get_payload_as_dictionary() is not None and len(msg.get_payload_as_dictionary()) > 0:
                    logger.debug("SDM Check: Found unexpected Attachment Map header")
                    return False
                if msg.get_payload_as_list() is not None and len(msg.get_payload_as_list()) > 0:
                    logger.debug("SDM Check: Found unexpected Attachment Stream header")
                    return False
        else:
            if self.use_binary_attachment:
                if msg.get_payload_as_dictionary() is not None and len(msg.get_payload_as_dictionary()) > 0:
                    logger.debug("SDM Check: Found unexpected Attachment Map header")
                    return False
                if msg.get_payload_as_list() is None:
                    logger.debug("SDM Check: Did not find expected attachment stream")
                    return False
            else:
                logger.debug("SDM Check: Invalid config.  Cannot create stream in user props")
                return False

        if not isinstance(self.get_payload(), bytearray):
            if self.is_map_msg:
                if not isinstance(self.container, dict):
                    logger.debug("SDM Check: Attachment map is different than expected")
                    return False
            else:
                if not isinstance(self.container, list):
                    logger.debug("SDM Check: Attachment stream is different than expected")
                    return False

        if (msg.get_properties() is None or len(msg.get_properties()) == 0) and len(self.get_properties()) > 0:
            logger.debug("SDM Check: Did not find expected user property map")
            return False
        elif msg.get_properties() is not None and len(msg.get_properties()) > 0 and len(self.get_properties()) == 0:
            logger.debug("SDM Check: Found unexpected user property map")
            return False
        elif msg.get_properties() is not None and len(msg.get_properties()) > 0 and len(self.get_properties()) > 0:
            msg_props = msg.get_properties()
            self_props = self.get_properties()
            for k in msg_props:
                if k in self_props:
                    init_val = self_props[k]
                    if isinstance(msg_props[k], float):
                        # In C++, if a float value cannot be represented, the behavior is undefined. We do
                        # not deal with this case here. However, if the value can be represented, but cannot 
                        # be represented exactly, the result is the nearest higher or the nearest lower value
                        # This often happens with subnormal floats.
                        if self_props[k] != msg_props[k]:
                            if C_FLOATING_POINT_SUBNORMAL_BOUND < self_props[k] < C_FLOATING_POINT_NORMAL_BOUND:
                                self_props[k] = C_FLOATING_POINT_SUBNORMAL_BOUND
                            elif -C_FLOATING_POINT_SUBNORMAL_BOUND > self_props[k] > -C_FLOATING_POINT_NORMAL_BOUND:
                                self_props[k] = -C_FLOATING_POINT_SUBNORMAL_BOUND
                        # C++ also skews some other float values, so we just round them all to 2 decimal places
                        temp_str = str(msg_props[k])
                        if "e" in temp_str:
                            # The scientific notation case
                            msg_props[k] = "{:0.2e}".format(msg_props[k])
                            self_props[k] = "{:0.2e}".format(self_props[k])
                        else:
                            msg_props[k] = round(msg_props[k], 2)
                            self_props[k] = round(self_props[k], 2)
                    elif isinstance(msg_props[k], int) and self_props[k] > msg_props[k]:
                        # In C++, if we try and set an int type to a value greater than its maximum, it will
                        # do an implicit integer conversion (value % 2^n), where n is the width of the
                        # integer type (16, 32, or 64). We need to adjust the Python value accordingly.
                        bits = 0
                        if self_props[k] > C_LONG_LONG_INT_MAX:
                            bits = 64
                        elif self_props[k] > C_INT_MAX:
                            bits = 32
                        elif self_props[k] > C_SHORT_INT_MAX:
                            bits = 16
                        if bits != 0:
                            self_props[k] = self_props[k] % 2**bits
                    if msg_props[k] != self_props[k]:
                        # The value of the inbound message is not the same as our parsed file
                        return False
                    self_props[k] = init_val
                else:
                    return False
        return True
