from pypubsubtools.utils.log import get_logger

logger = get_logger()


# File called tool_data does not get recognized by pycharm, using tool_info for now
class ToolData:
    MAGIC_WORD = bytearray([0x50, 0x1A, 0xCE, 0x01])

    PROP_XML_PAYLOAD_HASH = "x"
    PROP_BIN_PAYLOAD_HASH = "b"
    PROP_STREAM_ID = "s"
    PROP_MESSAGE_ID = "m"
    PROP_LATENCY = "l"
    PROP_REPUBLISHED = "r"

    XML_PAYLOAD_BIT = 0
    STREAM_ID_BIT = 1
    # rffuBit2 = 2; THIS WAS UNUSED IN JAVA, LEAVING IT HERE TO EXPLAIN THE GAP
    BIN_ATTACH_BIT = 3
    MESSAGE_ID_BIT = 4
    LATENCY_BIT = 5
    REPUBLISHED_BIT = 6

    PRESENCE_DATA_OFFSET = 4
    XML_PAYLOAD_SIZE = 4
    BIN_ATTACH_SIZE = 4
    STREAM_ID_SIZE = 4
    MESSAGE_ID_SIZE = 8
    LATENCY_SIZE = 8

    def __init__(self):
        self._has_latency = None
        self._latency = None
        self._xml_payload_hash = None
        self._bin_attach_hash = None
        self._message_identifier = None
        self._stream_id = None
        self._decoded_size = None
        self._republished = None

        self.reset()

    def reset(self):
        self._has_latency = False
        self._latency = 0
        self._xml_payload_hash = -1
        self._bin_attach_hash = -1
        self._message_identifier = -1
        self._stream_id = -1
        self._decoded_size = 0
        self._republished = False

    def encode(self, dest: bytearray):
        if self.has_data():
            if self.get_encoded_size() > len(dest):
                logger.error("Encode destination too small, returning")
                return

            tool_data_end = 0
            presence_data = 0

            # First put in our magic number
            dest[tool_data_end:tool_data_end + len(self.MAGIC_WORD)] = self.MAGIC_WORD
            tool_data_end += len(self.MAGIC_WORD)

            # Leave space for presence stuff, it will be inserted at the end
            tool_data_end += self.PRESENCE_DATA_OFFSET

            # add xml has if preset
            if self.has_xml_payload_hash():
                presence_data = self.set_bits(presence_data, 1, 1, self.XML_PAYLOAD_BIT)
                self.int_to_four_bytes(self._xml_payload_hash, dest, tool_data_end)
                tool_data_end += 4

            if self.has_stream_id():
                presence_data = self.set_bits(presence_data, 1, 1, self.STREAM_ID_BIT)
                self.int_to_four_bytes(self._stream_id, dest, tool_data_end)
                tool_data_end += 4

            if self.has_bin_attach_hash():
                presence_data = self.set_bits(presence_data, 1, 1, self.BIN_ATTACH_BIT)
                self.int_to_four_bytes(self._bin_attach_hash, dest, tool_data_end)
                tool_data_end += 4

            if self.has_message_identifier():
                presence_data = self.set_bits(presence_data, 1, 1, self.MESSAGE_ID_BIT)
                self.long_to_eight_bytes(self._message_identifier, dest, tool_data_end)
                tool_data_end += 8

            if self.has_latency():
                presence_data = self.set_bits(presence_data, 1, 1, self.LATENCY_BIT)
                self.long_to_eight_bytes(self._latency, dest, tool_data_end)
                tool_data_end += 8

            if self.get_republished():
                presence_data = self.set_bits(presence_data, 1, 1, self.REPUBLISHED_BIT)

            # Add presence bits in where we saved space at the top
            self.int_to_four_bytes(presence_data, dest, self.PRESENCE_DATA_OFFSET)

    def decode(self, msg_bytes: bytearray):
        self.reset()

        if len(msg_bytes) > (len(self.MAGIC_WORD) + self.PRESENCE_DATA_OFFSET) and \
                ToolData.four_bytes_to_int(msg_bytes) == ToolData.four_bytes_to_int(self.MAGIC_WORD):
            self._decoded_size = 0

            presence_data = ToolData.four_bytes_to_int(msg_bytes, self.PRESENCE_DATA_OFFSET)
            self._decoded_size += len(self.MAGIC_WORD) + self.PRESENCE_DATA_OFFSET

            if presence_data & (- presence_data) == 1 << self.XML_PAYLOAD_BIT:
                self._xml_payload_hash = ToolData.four_bytes_to_int(msg_bytes, self._decoded_size)
                self._decoded_size += self.XML_PAYLOAD_SIZE
                presence_data = self.set_bits(presence_data, 0, 1, self.XML_PAYLOAD_BIT)

            if presence_data & (- presence_data) == 1 << self.STREAM_ID_BIT:
                self._stream_id = ToolData.four_bytes_to_int(msg_bytes, self._decoded_size)
                self._decoded_size += self.STREAM_ID_SIZE
                presence_data = self.set_bits(presence_data, 0, 1, self.STREAM_ID_BIT)

            if presence_data & (- presence_data) == 1 << self.BIN_ATTACH_BIT:
                self._bin_attach_hash = ToolData.four_bytes_to_int(msg_bytes, self._decoded_size)
                self._decoded_size += self.BIN_ATTACH_SIZE
                presence_data = self.set_bits(presence_data, 0, 1, self.BIN_ATTACH_BIT)

            if presence_data & (- presence_data) == 1 << self.MESSAGE_ID_BIT:
                self._message_identifier = ToolData.eight_bytes_to_long(msg_bytes, self._decoded_size)
                self._decoded_size += self.MESSAGE_ID_SIZE
                presence_data = self.set_bits(presence_data, 0, 1, self.MESSAGE_ID_BIT)

            if presence_data & (- presence_data) == 1 << self.LATENCY_BIT:
                self._latency = ToolData.eight_bytes_to_long(msg_bytes, self._decoded_size)
                self._has_latency = True
                self._decoded_size += self.LATENCY_SIZE
                presence_data = self.set_bits(presence_data, 0, 1, self.LATENCY_BIT)

            if presence_data & (- presence_data) == 1 << self.REPUBLISHED_BIT:
                self._republished = True
                presence_data = self.set_bits(presence_data, 0, 1, self.REPUBLISHED_BIT)

            return True
        return False

    def has_data(self):
        return self.has_latency() or \
               self.has_xml_payload_hash() or \
               self.has_bin_attach_hash() or \
               self.has_message_identifier() or \
               self.has_stream_id() or \
               self.get_republished()

    def set_latency(self, latency):
        self._has_latency = True
        self._latency = latency

    def get_latency(self):
        return self._latency

    def has_latency(self):
        return self._has_latency

    def set_xml_payload_hash(self, xml_payload_hash):
        self._xml_payload_hash = xml_payload_hash

    def get_xml_payload_hash(self):
        return self._xml_payload_hash

    def has_xml_payload_hash(self):
        return self._xml_payload_hash > 0

    def set_bin_attach_hash(self, bin_attach_hash):
        self._bin_attach_hash = bin_attach_hash

    def get_bin_attach_hash(self):
        return self._bin_attach_hash

    def has_bin_attach_hash(self):
        return self._bin_attach_hash > 0

    def clear_bin_attach_hash(self):
        self._bin_attach_hash = -1

    def set_message_identifier(self, message_identifier):
        self._message_identifier = message_identifier

    def get_message_identifier(self):
        return self._message_identifier

    def has_message_identifier(self):
        return self._message_identifier != -1

    def set_stream_id(self, stream_id):
        self._stream_id = stream_id

    def get_stream_id(self):
        return self._stream_id

    def has_stream_id(self):
        return self._stream_id != -1

    def set_republished(self, republished):
        self._republished = republished

    def get_republished(self):
        return self._republished

    def get_decoded_size(self):
        if self._decoded_size < 0:
            return 0
        return self._decoded_size

    def get_encoded_size(self):
        tool_data_size = 0

        if not self.has_data():
            return 0

        # Tool data and presence bitmap are always first.
        tool_data_size += len(self.MAGIC_WORD) + 4

        tool_data_size += 8 if self.has_latency() else 0
        tool_data_size += 4 if self.has_xml_payload_hash() else 0
        tool_data_size += 4 if self.has_bin_attach_hash() else 0
        tool_data_size += 8 if self.has_message_identifier() else 0
        tool_data_size += 4 if self.has_stream_id() else 0

        return tool_data_size

    def __str__(self):
        buf = ""

        if self.has_xml_payload_hash():
            buf += f" XML Payload Integrity   = {self._xml_payload_hash}\n"
        if self.has_bin_attach_hash():
            buf += f" Bin attachment Integrity= {self._bin_attach_hash}\n"
        if self.has_stream_id():
            buf += f" Stream ID               = {self._stream_id}\n"
        if self.has_message_identifier():
            buf += f" Msg ID                  = {self._message_identifier}\n"
        if self.has_latency():
            buf += f" Latency Time Stamp      = {self._latency}\n"
        if self._republished:
            buf += f" Republished             = {self._republished}\n"
        return buf

    @staticmethod
    def set_bits(data, val_to_set, num_bits, shift_bits):
        """
        Apply interesting bits from 'valToSet' to data

        Pseudocode:
         Calculate interest mask for input value
         Apply interest mask and shift input value
         Shift interest mask and invert to get data clearing mask
         Clear target bits in data with data clearing mask
         data := data | shifted value
        """
        cur_mask = (1 << num_bits) - 1
        shifted_val = (val_to_set & cur_mask) << shift_bits
        data_clear_mask = ~(cur_mask << shift_bits)
        data = data & data_clear_mask
        return data | shifted_val

    @staticmethod
    def int_to_four_bytes(value, dest, offset):
        dest[offset + 0] = (value & 0xFF000000) >> 24
        dest[offset + 1] = (value & 0x00FF0000) >> 16
        dest[offset + 2] = (value & 0x0000FF00) >> 8
        dest[offset + 3] = value & 0x000000FF

    @staticmethod
    def long_to_eight_bytes(value, dest, offset):
        val1 = (value & 0xFFFFFFFF00000000) >> 32
        val2 = value & 0x00000000FFFFFFFF
        ToolData.int_to_four_bytes(val1, dest, 0 + offset)
        ToolData.int_to_four_bytes(val2, dest, 4 + offset)

    @staticmethod
    def four_bytes_to_int(value, offset=0):
        b0 = 0x000000FF & value[offset]
        b1 = 0x000000FF & value[offset + 1]
        b2 = 0x000000FF & value[offset + 2]
        b3 = 0x000000FF & value[offset + 3]

        return b0 << 24 | b1 << 16 | b2 << 8 | b3

    @staticmethod
    def eight_bytes_to_long(value, offset=0):
        i1 = ToolData.four_bytes_to_int(value, offset)
        i2 = ToolData.four_bytes_to_int(value, offset + 4)

        return i1 << 32 | i2
