import argparse
import os
import pkg_resources

from pypubsubtools.core.runtime_property_builder import RuntimePropertyBuilder
from pypubsubtools.utils.log import get_logger, init_logging
from sdkperf.config.authentication_scheme import AuthenticationScheme
from sdkperf.config.publisher_back_pressure_type import PublisherBackPressureType
from sdkperf.config.client_mode_type import ClientModeType
from sdkperf.config.endpoint_access_type import EndpointAccessType
from sdkperf.config.generic_message_delivery_mode import GenericMessageDeliveryMode
from sdkperf.config.subscriber_back_pressure_type import SubscriberBackPressureType

logger = get_logger()


class CliPropertiesParser:
    # noinspection PyTypeChecker
    def __init__(self):
        self._ap = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
        self._init = False

    def _init_options(self):
        self._ap.add_argument("-v", "--version", action='store_true', help="Print version number and exit")

        # Client Control
        self._ap.add_argument("-cip", metavar="ip[:port]", type=str,
                              help="IP and port of the client appliance. (e.g. -cip=192.168.160.101)")
        self._ap.add_argument("-cu", metavar="user[@vpn]", type=str,
                              help="Client username and optionally VPN name.")
        self._ap.add_argument("-cn", metavar="string", type=str,
                              help="Root name for clients, index appended (default in pubsub mode is '')")
        self._ap.add_argument("-cnbc", action='store_true', help="Enable non-blocking connect/disconnect for clients.")
        self._ap.add_argument("-cm", metavar="string", type=str, default='sink',
                              choices=['sink', 'reply'],
                              help="Client mode, one of {sink|reply}. (default is sink)\n"
                                   "In reply, all messages are reflected back to the client that send the message.")

        # Message Control
        self._ap.add_argument("-mr", metavar="number", type=int, default=1,
                              help="Publishing rate (msg/sec per client) (default 1)")
        self._ap.add_argument("-mn", metavar="number", type=int, default=0,
                              help="Total # messages to publish (across ALL clients). (default is 0)")
        self._ap.add_argument("-msa", metavar="list", type=str,
                              help="Comma sep size list in bytes for auto-generated binary attachment portion.")
        self._ap.add_argument("-mt", metavar="string", type=str, default='direct',
                              choices=['direct', 'nonpersistent', 'persistent'],
                              help="Message type, one of {direct|nonpersistent|persistent}. (default is direct)")
        self._ap.add_argument("-md", action='store_true', help="Dump all received messages to the screen as text")
        self._ap.add_argument("-mrt", metavar="string", type=str, default='avg', choices=['avg', 'max'],
                              help="Publishing rate targets one of 'avg' or 'max'. (default 'avg')")

        # Publishing
        self._ap.add_argument("-pal", metavar="list", type=str, help="Comma sep binary attachment list.")
        self._ap.add_argument("-ptl", metavar="list", type=str, help="Comma sep list of topics for publishing.")
        self._ap.add_argument("-ptf", metavar="file", type=str,
                              help="File containing list of topics for publishing.\n"
                                   "Note all publishers publish to all topics.")
        self._ap.add_argument("-pql", metavar="list", type=str, help="Comma sep list of queues for publishing.")
        self._ap.add_argument("-pqf", metavar="file", type=str,
                              help="File containing list of queues for publishing.\n"
                                   "Note all publishers publish to all queues.")
        self._ap.add_argument("-ptp", metavar="string", type=str,
                              help="Publish topic prefix. Use with -ptc to generate a large topic list.")
        self._ap.add_argument("-ptc", metavar="int", type=int,
                              help="Publish topic count. Use with -ptp to generate a large topic list.")
        self._ap.add_argument("-psm", action='store_true', help="Enable smooth publishers.  Used for latency testing.")
        self._ap.add_argument("-prt", action='store_true', help="Publish reply topic flag. Pubs will add reply to "
                                                                "topic in each send. (Only applies to direct messages)")
        self._ap.add_argument("-nmp", metavar="int", type=int, default=1,
                              help="Number of message producers per client. (Default 1)\n"
                                   "The first producer will always be the default producer.\n"
                                   "When publishing, the client will round-robin over all available producers.")
        self._ap.add_argument("-pmp", action='store_true',
                              help="Enables Publish Message Priority. "
                                   "Priority is randomly generated on every message sent.\n"
                                   "Equivalent to -pmpl=0,10,2,10,3,10,4,10,5,10,6,10,7,10,8,10,9,10\n"
                                   "Which translates to 10 percent probability for all priorities between 0 and 9\n"
                                   "Message Priority Stats (-mps) enabled when this parameter is used.")
        self._ap.add_argument("-pmpl", metavar="list", type=str,
                              help="Publish Message Priority List. Comma-separated, 'priority value,percentage'\n"
                                   "For example, -pmpl=0,50,9,50 -- 50 percent of messages will be published at priority 0\n"
                                   "while the other 50 percent will be published at priority 9. "
                                   "The priority is chosen at random\n"
                                   "before every message is published based on probabilities defined.\n"
                                   "Message Priority Stats (-mps) enabled when this parameter is used.")
        self._ap.add_argument("-bd", metavar="int", type=float, default=0,
                              help="Duration of burst (double value in seconds), enables bursty publishing.\n"
                                   "Use with inter-burst duration.")
        self._ap.add_argument("-ibd", metavar="int", type=float, default=0,
                              help="Optional Inter-burst duration (value in seconds), specifies duration of inter\n"
                                   "burst gaps. Use with burst duration")
        self._ap.add_argument("-httpce", metavar="string", type=str, help="HTTP content encoding.")
        self._ap.add_argument("-httpct", metavar="string", type=str, help="HTTP content type.")
        self._ap.add_argument("-ped", metavar="int", type=int, default=2,
                              help="Publish end delay, The time in seconds\n"
                                   "to wait following publishing the final message before\n"
                                   "bringing down subscriber data channels.")
        self._ap.add_argument("-pkl", metavar="string", type=str,
                              help="Partition Key List, comma separated list of keys to be "
                                   "looped through when publishing messages.\n")

        # Subscribing
        self._ap.add_argument("-stl", metavar="list", type=str, help="Comma sep list of topics for subscribing.")
        self._ap.add_argument("-stf", metavar="list", type=str, help="List of files containing topics for subscribing.")
        self._ap.add_argument("-sql", metavar="list", type=str,
                              help="Comma sep list of queues for subscribing applied round robin to clients.")
        self._ap.add_argument("-sqf", metavar="list", type=str,
                              help="List of files containing queues for subscribing.")
        self._ap.add_argument("-stp", metavar="string", type=str,
                              help="Subscription topic prefix. Use with -stc to generate a large topic list.")
        self._ap.add_argument("-stc", metavar="int", type=int,
                              help="Subscription topic count. Use with -stp to generate a large topic list.")
        self._ap.add_argument("-ssl", metavar="list", type=str,
                              help="List of selector strings applied to filter messages from flows")
        self._ap.add_argument("-pe", action='store_true',
                              help="Enables provisioning and cleanup of queues and topic endpoints\n")
        self._ap.add_argument("-pea", metavar="int", type=int,
                              help="Provisioned Endpoints Access Type. 0 for Non Exclusive, 1 for Exclusive\n")
        self._ap.add_argument("-nsr", action='store_true',
                              help="No subscription remove. Flag to indicate that sdkperf should\n"
                                   "not remove subscriptions on exit.\n")
        self._ap.add_argument("-mps", action='store_true',
                              help="Message Priority Stats will be tracked and printed on exit.")
        self._ap.add_argument("-fr", action='store_true',
                              help="Request replay when creating subscriber flow(s).\n"
                                   "If no date is provided (-fdr) then replay will be requested "
                                   "from beginning of replay log.")
        self._ap.add_argument("-fdr", metavar="string", type=str,
                              help="Request replay from provided date when creating subscriber flow(s). "
                                   "(e.g. 2018-07-30T10:00:00-04:00).")
        self._ap.add_argument("-rs", metavar="int", type=int, default=0,
                              help="Specify the receiver sleep time in seconds. Used to simulate a slow receiver")
        self._ap.add_argument("-fmidr", metavar="string", type=str,
                              help="Request replay from provided replication message id when creating subscriber flow(s). "
                                    "(eg. rmid1:0117b-75462900ca8-00000000-00135082)")

        # Performance measurement options
        self._ap.add_argument("-l", action='store_true',
                              help="Enable latency measurements  (default is no latency).")
        self._ap.add_argument("-lb", metavar="number", type=int, default=1024,
                              help="Enable latency measurement with # buckets (default 1024 buckets)")
        self._ap.add_argument("-lg", metavar="number", type=int, default=0,
                              help="Latency granularity factor (default 0)")
        self._ap.add_argument("-lwu", metavar="number", type=float, default=30.0,
                              help="Set latency warmup in seconds.  (default 30.0)\n")
        self._ap.add_argument("-crc", action='store_true',
                              help="Enable CRC-32 check of message payload (default off)")
        self._ap.add_argument("-oc", action='store_true',
                              help="Enable out-of-order checking and loss/duplicate detection on clients.")
        self._ap.add_argument("-di", action='store_true',
                              help="Enable Discard indication statistics.")

        # Tool Specific Options
        self._ap.add_argument("-ka", metavar="number", type=int,
                              help="Keep alive interval in milliseconds. (Default off)")
        self._ap.add_argument("-z", metavar="number", type=int,
                              help="Enable compression. (1..9) 1 is fastest, 9 max compression.")
        self._ap.add_argument("-pz", metavar="number", type=int,
                              help="Enable payload compression. (1..9) 1 is fastest, 9 max compression.")
        self._ap.add_argument("-rc", metavar="number", type=int, default=3,
                              help="Retry Count.  Number of times API will try to re-establish a failed connection. \n"
                                   "(Default 3)")
        self._ap.add_argument("-epl", metavar="list", type=str,
                              help="Comma sep list of extra Props passed blindly to the API")

        # Advanced Client control
        self._ap.add_argument("-cc", metavar="number", type=int, default=1,
                              help="Number of client connections (default 1)")
        self._ap.add_argument("-cp", metavar="password", type=str, default='', help="Client password (default '')")
        self._ap.add_argument("-cpc", action='store_true',
                              help="Context per client.  Enable 1 context for each client.")
        self._ap.add_argument("-cpt", metavar="number", type=int, default=1,
                              help="Number of publish threads per client (default 1)")
        self._ap.add_argument("-cptt", metavar="number", type=int, default=1,
                              help="Total number of publish threads (default 1). \nClients will be evenly spread out "
                                   "on all available threads.\nMutually exclusive with -cpt parameter.")

        # Next Gen Options
        self._ap.add_argument("-pbp", metavar="list", type=str,
                              help="Publisher back pressure types. \nOptions are [elastic, reject, wait, no] (default wait)\n"
                                   "If multiple backpressure types are specified they will be \n"
                                   "applied to producers in each client in a round robin method.\n"
                                   "Each backpressure type will use the same buffer size. \n"
                                   "(No always uses reject with buffer of 0)")
        self._ap.add_argument("-pbpb", metavar="number", type=int,
                              help="Publisher back pressure buffer size for reject or wait (default 10000)")

        self._ap.add_argument("-sbp", metavar="string", type=str,
                              help="Subscriber back pressure types. \nThe options are [elastic, oldest, latest] (default elastic\n")
        self._ap.add_argument("-sbpb", metavar="number", type=int,
                              help="Subscriber back pressure buffer size for drop latest or oldest (default 10000)")

        self._ap.add_argument("-gp", metavar="number", type=int, default=600000,
                              help="Grace period (ms) for shutting down producers and consumers. \n"
                                   "Set to 0 for no grace period (default 600000ms, ignored on ctrl+c)")
        self._ap.add_argument("-dmpl", action='store_true',
                              help="Disable message producer listeners and stats reporting")
        self._ap.add_argument("-paa", action='store_true',
                              help="Enable Publish Await Acknowledgement (For persistent) or "
                                   "Publish Await Response (For request/reply).")
        self._ap.add_argument("-dtl", action='store_true',
                              help="Disable Termination Listeners "
                                   "(don't set_termination_notification_listener on each pub/rec)")
        self._ap.add_argument("-wut", action='store_true',
                              help="Want Ungraceful Terminate (don't call terminate on pub/rec)")
        self._ap.add_argument("-wud", action='store_true',
                              help="Want Ungraceful Disconnect (don't call disconnect on service)")

        # Message Control
        self._ap.add_argument("-mdq", action='store_true',
                              help="Messages should be flagged eligible for the dead message queue")
        self._ap.add_argument("-mtl", metavar="int", type=int,
                              help="Messages will be tagged with time to live (millis) of this value")
        self._ap.add_argument("-mee", action='store_true',
                              help="Messages should be flagged eligible for message eliding")
        self._ap.add_argument("-sdm", metavar="list", type=str,
                              help="Comma sep list of input files to use in structured data messaging.")
        self._ap.add_argument("-mot", metavar="string", type=str,
                              help="Message Object Type pattern to be used for publishing messages. Pattern repeats after\n"
                                   "it has been traversed. Message Object Type, creates small, canned messages of the given type instead\n"
                                   "of using message size.  Will also enable checking of the type if used with -motc.\n"
                                   "Pattern Values: b - Bytes, t - Text, s - Stream, m - Map (Eg. -mot=tsbbms). ")
        self._ap.add_argument("-motc", action='store_true',
                              help="Message Object Type will be checked for each message sent with the -mot flag.")

        # Assured Delivery
        self._ap.add_argument("-ats", metavar="int", type=int, default=0,
                              help="AD transaction size. Default is 0 (disabled)")
        self._ap.add_argument("-atpc", action='store_true',
                              help="Producer and consumers on the same transaction. "
                                   "Default is on separate transactions.")
        self._ap.add_argument("-atr", action='store_true',
                              help="Flag to enable mode where every transaction is rolled back (default is to commit)")
        self._ap.add_argument("-atri", metavar="int", type=int, default=-1,
                              help="AD transaction rollback interval.  Used in combination with -ats to\n"
                                   "rollback rather than commit at the specified frequency (default is disabled)")
        self._ap.add_argument("-att", metavar="int", type=int, default=0,
                              help="AD transaction idle time.  Used in combination with -ats.  If a partial\n"
                                   "uncommitted transaction does not receive a message within <int> ms, a commit\n"
                                   "is triggered.  Default is 0 ms (disabled)")
        self._ap.add_argument("-afi", action='store_true',
                              help="Enable active flow indication on each flow")
        self._ap.add_argument("-aii", metavar="int", type=int, default=0,
                              help="ACK immediately interval.  Valid range is >= 0, default is 0 (disabled)")
        self._ap.add_argument("-sp", metavar="string", type=str,
                              help="Settlement Pattern to be used by each flow. Pattern repeats after it has been traversed.\n"
                                   "Pattern Values: a - Accepted, r - Rejected, f - Failed. (Eg. -sp=faraaa)\n")
        self._ap.add_argument("-rsp", metavar="string", type=str,
                              help="Required Settlements to be enabled on each flow.\n"
                                   "Accepted Values: a - Accepted, r - Rejected, f - Failed.\n")

        # Authentication and SSL
        self._ap.add_argument("-as", dest="as_", metavar="string", type=str, default="BASIC",
                              choices=['basic', 'client-certificate', 'kerberos', 'oauth2'],
                              help="Client authentication scheme.  Valid values are: [basic, client-certificate, kerberos, oauth2]")
        self._ap.add_argument("-oaat", metavar="string", type=str,
                              help="OAuth Access Token. (b64token encoded)")
        self._ap.add_argument("-oaii", metavar="string", type=str,
                              help=" OAuth Issuer Identifier.")
        self._ap.add_argument("-oidt", metavar="string", type=str,
                              help=" OAuth ID Token. (b64token encoded)")

        self._ap.add_argument("-sslp", metavar="string", type=str,
                              help="Comma separated list of the encryption protocol(s) to use.")
        self._ap.add_argument("-sslep", metavar="string", type=str,
                              help="Comma separated list of excluded SSL protocol(s).")
        self._ap.add_argument("-sslvc", action='store_true',
                              help="Enable validation of server certificates. (default is false)")
        self._ap.add_argument("-sslvcd", action='store_true',
                              help="Enable validation of certificate date.    (default is false)")
        self._ap.add_argument("-sslvch", action='store_true',
                              help="Enable validation of certificate hostname. (default is false)")
        self._ap.add_argument("-sslcs", metavar="string", type=str,
                              help="Comma separated list of cipher suites to enable.")
        self._ap.add_argument("-ssltsd", metavar="string", type=str,
                              help="Path to trust store directory.")
        self._ap.add_argument("-ssltcn", metavar="string", type=str,
                              help="Comma separated list of trusted common names.")
        self._ap.add_argument("-sslcf", metavar="string", type=str,
                              help="Client certificate file.")
        self._ap.add_argument("-sslpk", metavar="string", type=str,
                              help="Client private key file.")
        self._ap.add_argument("-sslpkp", metavar="string", type=str,
                              help="Client private key file password.")
        self._ap.add_argument("-sslcd", metavar="string", type=str, choices=["PLAIN_TEXT"],
                              help="SSL Connection Downgrade. Session property specifying a transport protocol\n"
                                   "that the connection will be downgraded to after client authentication.\n"
                                   "Default: API default (do not downgrade); Accepted Value: PLAIN_TEXT")
        self._ap.add_argument("-ksn", metavar="string", type=str,
                              help="Kerberos service name.")

        # Cache Options
        self._ap.add_argument("-chpn", metavar="string", type=str, help="Name of cache for requests.")
        self._ap.add_argument("-chpm", metavar="int", type=int, help="Max Msgs Per Topic.")
        self._ap.add_argument("-chpa", metavar="int", type=int, help="Max age for cached messages in secs.")
        self._ap.add_argument("-chpt", metavar="int", type=int, help="Timeout in msec for a given cache request.")
        self._ap.add_argument("-chrr", metavar="float", type=float, help="Request rate (requests/sec).")
        self._ap.add_argument("-chrn", metavar="int", type=int, help="Request number.  How many requests to send at the"
                                                                     " rate specified.")
        self._ap.add_argument("-chld", metavar="string", type=str,
                              choices=["QUEUE", "FULFILL", "FLOW_THRU", "CACHED_ONLY"],
                              help="Cache Live Data Action. One of QUEUE, FULFILL, FLOW_THRU, CACHED_ONLY.")
        self._ap.add_argument("-chrs", action='store_true', help="Want Request on Subscribe - Tool will perform cache "
                                                                 "request on every subscribe performed")

        # Tracing options
        self._ap.add_argument("-tcc", action='store_true', help="Inject creation context with pre-defined values for trace ID, sampled ID, trace state and sampled flag\ninto published messages and generate/export spans at the same time.")
        self._ap.add_argument("-bag", metavar="string", type=str, help="Creation context injected into messages will have specified baggage, Format: \"key1=val1,key2=val2\" (Default: Empty).")
        self._ap.add_argument("-ttc", action='store_true', help="Inject transport context with pre-defined values for trace ID, sampled ID, trace state and sampled flag\ninto published messages and generate/export spans at the same time.")
        self._ap.add_argument("-tcrc", action='store_true', help="Copy receive context into a new span to be exported. (Default: False).")
        self._ap.add_argument("-tecip", metavar="string", type=str, help="Trace Exporter Connection String, point to a OTLP receiver for spans generated\nvia -tcc, -ttc, or -tcrc. Example: \"http://192.168.0.1:4317\"\n(Default: Empty, uses InMemorySpanExporter instead)")
        self._ap.add_argument("-tebs", metavar="int", type=int, help="Trace Exporter Batch Size. (Default: 5).")
        self._ap.add_argument("-tesd", metavar="int", type=int, help="Trace Exporter Schedule Delay in millisecond. (Default: 100ms).")
        self._ap.add_argument("-teqs", metavar="int", type=int, help="Trace Exporter Queue Size. (Default: 2048).")

        # General Options
        self._ap.add_argument("-soe", action='store_true',
                              help="Stop client connections or publishing on errors")
        self._ap.add_argument("-log", metavar="string", type=str, default="WARNING",
                              help="Log level for python api and tool logs")

        self._ap.add_argument("-clog", metavar="string", type=str,
                              help="Log level for c api logs")
        self._ap.add_argument("-clogf", metavar="string", type=str,
                              help="File name for c api logs (only used if -clog is also used)")

        self._init = True

    def parse_cli(self, cli_opts):
        from pypubsubtools.core.runtime_property_builder import RuntimePropertyBuilder

        if not self._init:
            self._init_options()

        args = self._ap.parse_args(cli_opts)
        builder = RuntimePropertyBuilder()
        self.load_default_options(builder)

        if len(cli_opts) == 0:
            self._ap.print_help()
            exit(0)

        if args.version:
            tool_version = pkg_resources.require("sdkperf_py")[0].version
            api_version = pkg_resources.require("solace-pubsubplus")[0].version
            otel_version = pkg_resources.require("pubsubplus-opentelemetry-integration")[0].version
            print(f"sdkperf_py Version: {tool_version}")
            print(f"solace-pubsubplus Version: {api_version}")
            print(f"pubsubplus-opentelemetry-integration Version: {otel_version}")
            exit(0)

        init_logging(args.log, False, args.clog, args.clogf)

        # Client Control
        if args.cip is not None:
            builder.set_client_ip_addr(args.cip)
        if args.cu is not None:
            parts = args.cu.split("@")
            if len(parts) == 2:
                builder.set_client_vpn(parts[1])
            builder.set_client_username(parts[0])
        if args.cn is not None:
            builder.set_client_name_prefix(args.cn)
        if args.cnbc:
            builder.set_want_non_blocking_connect(True)
        if args.cm.upper() is not ClientModeType.SINK:
            builder.set_client_mode(args.cm)

        # Advanced Client control
        if args.cc > 0:
            builder.set_num_clients(args.cc)
        builder.set_client_password(args.cp)
        if args.cpc:
            builder.set_want_context_per_cient(True)
        if args.cpt > 1 and args.cptt > 1:
            self._ap.error("Cannot set -cpt and -cptt flags simultaneously, they are mutually exclusive.")
        if args.cpt > 1:
            builder.set_num_pubs_per_session(args.cpt)
        if args.cptt > 1:
            builder.set_num_pub_threads(args.cpt)

        # Message Control
        if args.mr >= 0:
            builder.set_publish_rate_per_pub(args.mr)
        else:
            self._ap.error("Message rate (mr) cannot be less than 0")
        if args.mn >= 0:
            builder.set_num_msgs_to_publish(args.mn)
        else:
            self._ap.error("Number of messages to publish (mn) cannot be less than 0")
        if args.msa is not None and len(args.msa) > 0:
            sizes = [int(x) for x in args.msa.split(",")]
            builder.set_attachment_size_list(sizes)
        if args.mt in ["direct", "nonpersistent", "persistent"]:
            builder.set_pub_message_type(args.mt)
        builder.set_want_msg_dump(args.md)
        if args.mrt == 'max':
            builder.set_msg_rate_is_max(True)
        if args.nsr:
            builder.set_want_remove_subscriber(False)

        # Publishing
        if args.pal is not None and len(args.pal) > 0:
            files = args.pal.split(",")
            builder.set_attachment_file_list(files)
        if args.ptl is not None and len(args.ptl) > 0:
            topics = args.ptl.split(",")
            builder.set_publish_topic_list(topics)
        if args.ptf is not None and len(args.ptf) > 0:
            if not os.path.exists(args.ptf):
                self._ap.error("File used for PTF does not exist")
            topics = list()
            with open(args.ptf, "r") as ptf_file:
                for line in ptf_file:
                    topics.append(line)
            builder.set_publish_topic_list(topics)
        if args.pql is not None and len(args.pql) > 0:
            queues = args.pql.split(",")
            builder.set_publish_queue_list(queues)
            builder.set_pub_message_type("persistent")
        if args.pqf is not None and len(args.pqf) > 0:
            if not os.path.exists(args.pqf):
                self._ap.error("File used for PQF does not exist")
            queues = list()
            with open(args.pqf, "r") as pqf_file:
                for line in pqf_file:
                    queues.append(line)
            builder.set_publish_queue_list(queues)
            builder.set_pub_message_type("persistent")
        if args.ptp is not None and args.ptc is not None and args.ptc > 0:
            prefix = args.ptp
            tc = args.ptc
            topics = list()
            for i in range(1, tc + 1):
                topics.append(f"{prefix}{i:06d}")
            builder.set_publish_topic_list(topics)
        elif (args.ptp is None and args.ptc is not None) or (args.ptp is not None and args.ptc is None):
            self._ap.error("Both ptp and ptc must be specified at the same time")
        if int(args.nmp) > 1:
            builder.set_num_message_producer(int(args.nmp))
        if args.pmp and args.pmpl is not None:
            self._ap.error("Only one of '-pmp' or '-pmpl' can be provided.")
        elif args.pmp:
            builder.set_want_message_priority_stats(True)
            builder.set_publish_message_priority([0, 10, 1, 10, 2, 10, 3, 10, 4, 10, 5, 10, 6, 10, 7, 10, 8, 10, 9, 10])
        elif args.pmpl is not None:
            pri = [int(x) for x in args.pmpl.split(",")]
            if len(pri) % 2 != 0:
                self._ap.error("Message priority list must be dividable by two.")
            builder.set_want_message_priority_stats(True)
            builder.set_publish_message_priority(pri)
        if args.httpce is not None:
            builder.set_http_content_encoding(args.httpce)
        if args.httpct is not None:
            builder.set_http_content_type(args.httpct)
        if args.psm:
            builder.set_want_smooth_pub_calc_lat(True)
        if args.prt:
            builder.set_want_reply_topic(True)
        if args.ped:
            if args.ped < 0:
                self._ap.error("-ped cannot be less than 0")
            builder.set_publish_end_delay_in_sec(args.ped)
        if args.pkl is not None:
            builder.set_partition_key_list(args.pkl)
        if args.bd > 0:
            builder.set_burst_duration(args.bd)
        if args.ibd > 0:
            builder.set_inter_burst_duration(args.ibd)

        # Subscribing
        if args.stl is not None and len(args.stl) > 0:
            builder.set_sub_topic_lists(args.stl)
        if args.stf is not None and len(args.stf) > 0:
            builder.set_sub_topic_list_from_file_list(args.stf)
        if args.sql is not None and len(args.sql) > 0:
            builder.set_sub_queue_lists(args.sql)
        if args.sqf is not None and len(args.sqf) > 0:
            builder.set_sub_queue_list_from_file_list(str.split(args.sqf, ','))
        if args.stp is not None and args.stc is not None and args.stc > 0:
            prefix = args.stp
            tc = args.stc
            topics = list()
            for i in range(1, tc + 1):
                topics.append(f"{prefix}{i:06d}")
            builder.set_sub_topic_lists(topics)
        elif (args.stp is None and args.stc is not None) or (args.stp is not None and args.stc is None):
            self._ap.error("Both ptp and ptc must be specified at the same time")
        if args.ssl is not None and len(args.ssl) > 0:
            selectors = args.ssl.split(",")
            builder.set_selector_list(selectors)
        if args.pe:
            builder.set_want_provisioned_endpoint(True)
        if args.pea is not None:
            if args.pea == 1:
                builder.set_pe_access_type(EndpointAccessType.ACCESSTYPE_EXCLUSIVE)
            elif args.pea == 0:
                builder.set_pe_access_type(EndpointAccessType.ACCESSTYPE_NONEXCLUSIVE)
            else:
                self._ap.error("Invalid Endpoint Access Type provided (-pea):" + str(args.pea))
        if args.mps:
            builder.set_want_message_priority_stats(True)
        if args.fr:
            builder.set_want_replay(True)
        if args.fdr is not None and len(args.fdr) > 0:
            builder.set_want_replay(True)
            builder.set_want_replay_from_date(args.fdr)
        if args.fmidr is not None and len(args.fmidr) > 0:
            builder.set_want_replay(True)
            builder.set_want_replay_from_msg_id(args.fmidr)

        # Performance measurement options
        if args.l:
            builder.set_want_latency(True)
            builder.set_latency_buckets(args.lb)
            builder.set_latency_granularity(args.lg)
            builder.set_latency_warmup_in_secs(args.lwu)
        if args.oc:
            builder.set_want_order_check(True)
        if args.crc:
            builder.set_want_payload_check(True)
        if args.di:
            builder.set_want_di_stats(True)

        # Tool Specific Options
        if args.ka is not None and (args.ka == 0 or args.ka >= 50):
            builder.set_keepalive_interval_msec(args.ka)
        elif args.ka is not None:
            logger.info("Keepalive interval must be 0 (disabling the feature) or >= 50")
        if args.z is not None and 1 <= args.z <= 9:
            builder.set_client_compression_level(args.z)
        if args.pz is not None:
            builder.set_payload_compression_level(args.pz)
        builder.set_reconnect_attempts(args.rc)
        if args.epl is not None:
            try:
                builder.set_extra_prop_dict(args.epl.split(","))
            except Exception as e:
                self._ap.error(str(e))

        # Next Gen Options
        if args.pbp and len(args.pbp) > 0:
            builder.set_publisher_backpressure_type(args.pbp.split(","))
        if args.pbpb is not None:
            if args.pbpb < 0:
                self._ap.error("Cannot set -pbpb lower than 0.")
            builder.set_publisher_backpressure_buffer_size(args.pbpb)

        if args.sbp and len(args.sbp) > 0:
            builder.set_subscriber_backpressure_type(args.sbp)
        if args.sbpb:
            if args.sbpb <= 0:
                self._ap.error("Value of -sbpb must be positive")
            builder.set_subscriber_backpressure_buffer_size(args.sbpb)

        if args.rs < 0:
            self._ap.error("Cannot set -rs lower than 0.")
        builder.set_receiver_sleep_time(args.rs)

        if 0 < args.gp < 1000:
            self._ap.error("Cannot set -gp between 0 and 1000ms (If no grace period use 0).")
        builder.set_grace_period(args.gp)
        if args.dmpl:
            builder.set_want_producer_listeners(False)
        if args.paa:
            builder.set_want_publish_await_ack(True)
        if args.dtl:
            builder.set_want_termination_listeners(False)
        if args.wut:
            builder.set_want_ungraceful_terminate(True)
        if args.wud:
            builder.set_want_ungraceful_disconnect(True)

        # Message Control
        if args.mdq:
            builder.set_msg_dmq_eligible(True)
        if args.mtl is not None and int(args.mtl) >= 0:
            builder.set_msg_ttl(int(args.mtl))
        if args.mee:
            builder.set_msg_eliding_eligible(True)
        if args.sdm is not None:
            builder.set_sdm_prop_list_from_files(args.sdm.split(","))
            builder.set_want_struct_msg_check(True)
        if args.mot is not None:
            builder.set_msg_obj_type_list(args.mot)
        if args.motc:
            builder.set_want_msg_obj_type_check(True)

        # Assured Delivery
        if args.ats is not None and args.ats > 0:
            builder.set_ad_transaction_size(args.ats)
        if args.atpc:
            builder.set_ad_want_producer_consumers_transaction(True)
        if args.atr:
            builder.set_ad_transaction_rollback_interval(1)
        if args.atri is not None and args.atri > -1:
            builder.set_ad_transaction_rollback_interval(args.atri)
        if args.att is not None and args.att > -1:
            builder.set_ad_transaction_idle_time(args.att)
        if args.afi:
            builder.set_ad_want_active_flow_indication(True)
        if args.aii is not None and args.aii > 0:
            builder.set_ad_ack_immediately_interval(args.aii)
        if args.sp is not None and len(args.sp) > 0:
            builder.set_want_client_ack(True)
            builder.set_settlement_pattern_string(args.sp)
        if args.rsp is not None and len(args.rsp) > 0:
            builder.set_required_settlement_string(args.rsp)



        # Authentication and SSL
        if args.as_ == "basic":
            builder.set_authentication_scheme(AuthenticationScheme.BASIC)
        elif args.as_ == "client-certificate":
            builder.set_authentication_scheme(AuthenticationScheme.CLIENT_CERTIFICATE)
        elif args.as_ == "kerberos":
            builder.set_authentication_scheme(AuthenticationScheme.GSS_KRB)
        elif args.as_ == "oauth2":
            builder.set_authentication_scheme(AuthenticationScheme.OAUTH2)

        if args.oaat is not None:
            builder.set_oauth_access_token(args.oaat)
        if args.oaii is not None:
            builder.set_oauth_issuer_id(args.oaii)
        if args.oidt is not None:
            builder.set_oauth_id_token(args.oidt)

        if args.sslp is None and args.sslep is None:
            pass  # Nothing to do here
        elif args.sslp is None and args.sslep is not None:
            builder.set_ssl_excluded_protocols(args.sslep)
        elif args.sslp is not None and args.sslep is None:
            builder.set_ssl_protocol(args.sslp)
        else:
            logger.warning("Options to force SSL protocol and exclude SSL protocol are conflicting. Using exclude list")

        if args.sslvc:
            builder.set_ssl_validate_certificate(True)
        if args.sslvcd:
            builder.set_ssl_validate_certificate_date(True)
        if args.sslvch:
            builder.set_ssl_validate_certificate_hostname(True)
        if args.sslcs is not None:
            builder.set_ssl_cipher_suites(args.sslcs)
        if args.ssltsd is not None:
            builder.set_ssl_trust_store_dir(args.ssltsd)
        if args.ssltcn is not None:
            builder.set_ssl_trusted_common_name_list(args.ssltcn)
        if args.sslcf is not None:
            builder.set_ssl_client_certificate_file(args.sslcf)
        if args.sslpk is not None:
            builder.set_ssl_client_private_key_file(args.sslpk)
        if args.sslpkp is not None:
            builder.set_ssl_client_private_key_file_password(args.sslpkp)
        if args.sslcd is not None:
            builder.set_ssl_connection_downgrade_to(args.sslcd)
        if args.ksn is not None:
            builder.set_krb_service_name(args.ksn)

        if args.chpn is not None:
            builder.set_cache_prop_name(args.chpn)
            builder.set_cache_want_stats(True)
        if args.chpm is not None and args.chpm > 0:
            builder.set_cache_prop_max_msgs_per_topic(args.chpm)
        if args.chpa is not None and args.chpa > 0:
            builder.set_cache_prop_max_age(args.chpa)
        if args.chpt is not None and args.chpt > 0:
            builder.set_cache_prop_timeout_in_msec(args.chpt)
        if args.chrr is not None and args.chrr > 0:
            builder.set_cache_req_msg_rate(args.chrr)
        if args.chrn is not None and args.chrn > 0:
            builder.set_cache_num_req(args.chrn)
        if args.chld is not None:
            builder.set_cache_live_data_action(args.chld)
        if args.chrs:
            builder.set_cache_want_requests_on_subscribe(True)

        # Tracing Options
        if args.tcc:
            builder.set_want_trace_context_prop(True)
            builder.set_want_trace_creation_context(True)
        if args.bag is not None and args.bag != "":
            builder.set_want_trace_context_prop(True)
            builder.set_trace_baggage(args.bag)
        if args.ttc:
            builder.set_want_trace_context_prop(True)
            builder.set_want_trace_transport_context(True)
        if args.tcrc:
            builder.set_want_trace_context_prop(True)
            builder.set_want_trace_copy_receive_context(True)
        if args.tecip is not None and args.tecip != "":
            builder.set_want_trace_context_prop(True)
            builder.set_trace_exporter_connection_string(args.tecip)
        if args.tebs is not None and args.tebs > 0:
            builder.set_trace_exporter_batch_size(args.tebs)
        if args.tesd is not None and args.tesd > 0:
            builder.set_trace_exporter_schedule_delay(args.tesd)
        if args.teqs is not None and args.teqs > 0:
            builder.set_trace_exporter_queue_size(args.teqs)

        # General Options
        if args.soe:
            builder.set_want_stop_on_error(True)

        return builder.props

    @staticmethod
    def load_default_options(builder: RuntimePropertyBuilder):
        builder.set_reconnect_attempts(3)
        builder.set_reconnect_interval_msec(3000)
        builder.set_num_clients(0)
        builder.set_tool_mode("sdkperf")
        # builder.setToolApi(InternalApiMode.JCSMP.toString())
        builder.set_want_payload_check(False)
        builder.set_want_ud(False)
        builder.set_want_order_check(False)
        builder.set_want_loss_and_duplicate_detection(False)
        builder.set_want_order_memory(False)
        builder.set_want_message_examine_content(False)
        builder.set_num_pubs_per_session(1)
        builder.set_num_msgs_to_publish(0)
        builder.set_publish_end_delay_in_sec(2)
        builder.set_client_username("")
        builder.set_client_vpn("")
        # builder.set_client_name_prefix("") #  In python we don't want to set it unless we actually need it
        builder.set_client_password("")
        builder.set_client_description("")  # Not accessible via cli.
        builder.set_change_client_names_flag(False)
        builder.set_want_tcp_nodelay(True)  # DO NOT set the nodelay option
        builder.set_pub_message_type(GenericMessageDeliveryMode.DIRECT)
        builder.set_subscriber_sub_delay(0)
        builder.set_slow_sub_delay_count(-1)
        builder.set_subscriber_rate_interval(1)
        builder.set_want_remove_subscriber(True)
        builder.set_want_subscription_memory(True)
        builder.set_want_per_sub_thru_stats(False)
        builder.set_want_subscription_rate_stats(False)
        builder.set_sub_notification_queue_depth(-1)
        # builder.setWantCidStat(False)
        builder.set_want_di_stats(False)
        builder.set_want_verbose(False)
        builder.set_want_quiet(False)
        builder.set_want_versionprint(False)
        builder.set_want_latency(False)
        builder.set_want_full_latency_stats(True)
        builder.set_latency_buckets(1024)
        builder.set_latency_granularity(0)
        builder.set_latency_warmup_in_secs(30.0)
        builder.set_want_smooth_pub_calc_lat(False)
        builder.set_want_struct_msg_check(False)
        builder.set_publish_rate_per_pub(1)
        builder.set_want_context_per_cient(False)
        builder.set_burst_duration(0)
        builder.set_inter_burst_duration(0)
        builder.set_want_provisioned_endpoint(False)
        builder.set_pe_access_type(EndpointAccessType.ACCESSTYPE_API_DEFAULT)
        builder.set_want_stop_on_error(False)
        builder.set_pub_send_vect_size(0)
        builder.set_ad_pub_window_size(50)
        builder.set_want_msg_dump(False)
        builder.set_msg_dump_dir("")
        builder.set_client_mode(ClientModeType.SINK)
        builder.set_pub_stream_id_offset(-1)
        builder.set_pub_order_offset(0)
        builder.set_subscription_client_name("")
        builder.set_want_cut_through_persistence(False)
        builder.set_want_user_prop_tool_data(False)
        builder.set_msg_rate_is_max(False)
        builder.set_ignore_exists_errors(True)
        builder.set_want_client_ack(False)
        builder.set_want_client_ack_thread(False)
        builder.set_client_ack_skip_num(0)
        builder.set_client_ack_flush_queue(False)
        builder.set_client_ack_queue_reverse(False)
        builder.set_client_ack_random_depth(0)
        builder.set_sub_msg_queue_depth(0)
        builder.set_want_sync_receive(False)
        builder.set_want_no_local(False)
        builder.set_want_quiet(False)
        builder.set_ad_ack_immediately_interval(0)
        builder.set_ssl_connection_downgrade_to("")
        builder.set_ssl_validate_certificate(False)
        builder.set_ssl_validate_certificate_date(False)
        builder.set_ssl_validate_certificate_hostname(False)
        builder.set_ssl_cipher_suites("")
        builder.set_ssl_trust_store("")
        builder.set_ssl_trust_store_dir("")
        builder.set_ssl_trust_store_password("")
        builder.set_ssl_trust_store_format("")
        builder.set_ssl_trusted_common_name_list("")
        builder.set_ssl_key_store("")
        builder.set_ssl_key_store_password("")
        builder.set_ssl_key_store_format("")
        builder.set_ssl_private_key_alias("")
        builder.set_ssl_private_key_password(None)
        builder.set_ssl_client_certificate_file("")
        builder.set_ssl_client_private_key_file("")
        builder.set_ssl_client_private_key_file_password(None)
        builder.set_authentication_scheme(AuthenticationScheme.BASIC)
        builder.set_oauth_id_token(None)
        builder.set_oauth_access_token(None)
        builder.set_oauth_issuer_id(None)
        builder.set_ad_want_active_flow_indication(False)
        builder.set_ad_want_pub_ack_order_check(True)
        # builder.setWantReplicationReconnect(False)
        # builder.setWantApplicationServerCommand("")
        # builder.setWantRATransaction(False)
        builder.set_num_pub_threads(1)
        # builder.setClientTransportProtocol("")
        # builder.setWebTransportProtocolList("")
        # builder.setClientTransportProtocolDowngradeTimeout("")
        builder.set_want_per_message_details(False)
        builder.set_num_message_producer(1)
        builder.set_correlation_id("")
        builder.set_message_id("")
        builder.set_http_content_type(None)
        builder.set_http_content_encoding(None)
        builder.set_want_replay(False)
        # builder.setWantMessagePriorityStats(False)
        # builder.setWantMessagePriorityOrderChecking(False)
        builder.set_reconnect_attempts(-1)
        builder.set_reconnect_retry_interval_in_msecs(3000)
        builder.set_publisher_backpressure_type([PublisherBackPressureType.default()])
        builder.set_publisher_backpressure_buffer_size(10000)
        builder.set_subscriber_backpressure_type(SubscriberBackPressureType.default())
        builder.set_subscriber_backpressure_buffer_size(10000)
        builder.set_want_producer_listeners(True)
        builder.set_want_publish_await_ack(False)
        builder.set_want_single_pub_type(True)
        builder.set_want_termination_listeners(True)
        builder.set_want_ungraceful_terminate(False)
        builder.set_want_ungraceful_disconnect(False)
        builder.set_cache_prop_max_msgs_per_topic(0)
        builder.set_cache_prop_max_age(0)
        builder.set_cache_req_msg_rate(0)
        builder.set_cache_num_req(0)
        builder.set_cache_want_requests_on_subscribe(False)