import os
import signal
import sys
import time

from solace.messaging.utils.manageable import Metric

from pypubsubtools.core.runtime_properties import RuntimeProperties
from pypubsubtools.utils.log import get_logger
from sdkperf.config.constants import Constants
from sdkperf.config.endpoint_access_type import EndpointAccessType
from sdkperf_py.cli_properties_parser import CliPropertiesParser
from sdkperf.config.ep_config_properties import EpConfigProperties
from sdkperf.config.subscriber_destinations_type import SubscriberDestinationsType
from sdkperf.core.perf_stats import PerfStats
from sdkperf.core.sdkperf_client_collection import SdkPerfClientCollection

logger = get_logger()


class SDKPerf_python:
    def __init__(self, sdk_props: RuntimeProperties):
        self._props = sdk_props
        self._pub_rate_points = list()
        self._sub_rate_points = list()

    def handler(self, signum, frame):
        logger.warning("Caught Ctrl-C, shutting down. (Ignoring terminate grace period)")
        signal.signal(signal.SIGINT, self._old_handler)
        self.shutdown(0, True)
        exit(1)

    def run(self):
        self._old_handler = signal.signal(signal.SIGINT, self.handler)
        self._collection: SdkPerfClientCollection = SdkPerfClientCollection(self._props)
        try:
            self._collection.connect()
        except Exception as e:
            logger.error(f"Failed to connect clients: {e}")
            return 1

        self._transacted_session_names = list()
        pub_transacted_session_names = list()
        if getattr(self._props, RuntimeProperties.AD_TRANSACTION_SIZE, 0) > 0:
            num_clients = getattr(self._props, RuntimeProperties.NUM_CLIENTS, 1)
            pubs_per_session = getattr(self._props, RuntimeProperties.NUM_PUBS_PER_SESSION, 0)
            for i in range(num_clients):
                self._transacted_session_names.append(self._collection.open_transacted_session(i, self._props))

            for i in range(num_clients):
                for j in range(pubs_per_session):
                    if getattr(self._props, RuntimeProperties.AD_WANT_PRODUCER_CONSUMERS_TRANSACTION, False):
                        pub_transacted_session_names.append(self._transacted_session_names[i])
                    else:
                        pub_transacted_session_names.append(self._collection.open_transacted_session(i, self._props))

        setattr(self._props, RuntimeProperties.TRANSACTED_SESSION_NAME_LIST, pub_transacted_session_names)

        logger.info("Verifying subscriptions")
        if self.verify_subscription() != 0:
            return 1

        logger.info("Starting clients")
        if self.start_clients() != 0:
            return 1

        logger.info("Starting to publish if needed")
        if self.start_publishers() != 0:
            return 1

        logger.info("Waiting to finish publishing")
        if self.pub_loop() != 0:
            return 1

        logger.info("Shutting down")
        return self.shutdown(getattr(self._props, RuntimeProperties.GRACE_PERIOD, 600000))

    def verify_subscription(self):
        # Don't do queue browsing yet just go straight to adding
        try:
            self.add_subscriptions()
        except Exception:
            return 1
        return 0

    def add_subscriptions(self):
        num_clients = getattr(self._props, RuntimeProperties.NUM_CLIENTS, 1)
        topics = getattr(self._props, RuntimeProperties.SUB_TOPIC_LISTS, [])
        queues = getattr(self._props, RuntimeProperties.SUB_QUEUE_LISTS, [])
        selectors = getattr(self._props, RuntimeProperties.SELECTOR_LIST, [])
        transaction_size = getattr(self._props, RuntimeProperties.AD_TRANSACTION_SIZE, 0)

        for i in range(num_clients):
            if queues:
                temp_selectors = []
                if selectors:
                    temp_selectors.append(selectors[i % len(selectors)])
                ep_props = EpConfigProperties.create_for_queue_add(queues[i % len(queues)], temp_selectors)
                if getattr(self._props, RuntimeProperties.AD_WANT_ACTIVE_FLOW_INDICATION):
                    ep_props.want_active_flow_indication = True
                if getattr(self._props, RuntimeProperties.WANT_REPLAY):
                    ep_props.want_replay = True
                if getattr(self._props, RuntimeProperties.WANT_REPLAY_FROM_DATE, "") != "":
                    ep_props.want_replay_from_date = getattr(self._props, RuntimeProperties.WANT_REPLAY_FROM_DATE)
                if getattr(self._props, RuntimeProperties.WANT_REPLAY_FROM_MSG_ID, "") != "":
                    ep_props.want_replay_from_msg_id = getattr(self._props, RuntimeProperties.WANT_REPLAY_FROM_MSG_ID)
                if getattr(self._props, RuntimeProperties.WANT_PROVISIONED_ENDPOINT, False):
                    ep_props.want_provision = getattr(self._props, RuntimeProperties.WANT_PROVISIONED_ENDPOINT)
                    ep_props.access_type = getattr(self._props,
                                                   RuntimeProperties.PE_ACCESS_TYPE,
                                                   EndpointAccessType.ACCESSTYPE_API_DEFAULT)
                if transaction_size > 0:
                    ep_props.transacted_session_name = self._transacted_session_names[i]
                    ep_props.transaction_size = transaction_size

                self._collection.queue_update(ep_props, i)

                if topics:  # When both are specified then map topics
                    for queue in queues[i % len(queues)]:
                        self._collection.map_topics(queue, topics[i % len(topics)], True, i)
            elif topics:
                if getattr(self._props, RuntimeProperties.CACHE_WANT_REQUESTS_ON_SUBSCRIBE, False):
                    want_subscribe = True
                    live_data_action = getattr(self._props, RuntimeProperties.CACHE_LIVE_DATA_ACTION, None)
                    if live_data_action == "CACHED_ONLY":
                        # Override since this behavior does not normally exist in legacy APIs
                        want_subscribe = False
                    self._collection.cache_request(topics[i % len(topics)],
                                                   want_subscribe,
                                                   live_data_action,
                                                   True,
                                                   i)
                else:
                    ep_props = EpConfigProperties.create_for_subscription_add(topics[i % len(topics)],
                                                                              getattr(self._props,
                                                                                      RuntimeProperties.
                                                                                      SUBSCRIPTION_CLIENT_NAME,
                                                                                      ""),
                                                                              SubscriberDestinationsType.TOPIC)
                    self._collection.subscription_update(ep_props, i)
        return 0

    def start_clients(self):
        try:
            self._collection.start()
        except Exception as e:
            logger.error(f"Failed to start clients: {e}")
            return 1
        return 0

    def start_publishers(self):
        if getattr(self._props, RuntimeProperties.NUM_MSGS_TO_PUBLISH, 0) > 0:
            try:
                self._collection.start_publishing(self._props)
            except Exception as e:
                logger.error(f"Failed to start publishing: {e}")
                return 1
        if getattr(self._props, RuntimeProperties.CACHE_NUM_REQ, 0) > 0:
            try:
                self._collection.start_cache_requesting(getattr(self._props,
                                                                RuntimeProperties.PUBLISH_TOPIC_LIST,
                                                                []),
                                                        False,
                                                        getattr(self._props,
                                                                RuntimeProperties.CACHE_LIVE_DATA_ACTION,
                                                                None),
                                                        False,
                                                        getattr(self._props, RuntimeProperties.CACHE_NUM_REQ, 0),
                                                        getattr(self._props, RuntimeProperties.CACHE_REQ_MSG_RATE, 0))
            except Exception as e:
                logger.error(f"Failed to start cache requesting: {e}")
                return 1
        return 0

    def pub_loop(self):
        while True:
            time.sleep(.5)
            self._print_stats()

            if getattr(self._props, RuntimeProperties.NUM_MSGS_TO_PUBLISH, 0) > 0 or \
                    getattr(self._props, RuntimeProperties.CACHE_NUM_REQ, 0) > 0:
                if self._collection is not None and \
                        not self._collection.is_publishing() and \
                        not self._collection.is_cache_requesting():
                    # If we're monitoring publish, we also need to check that the API actually sent the messages
                    # not just put them in the backpressure buffer
                    # If we're monitoring cache, we can just break once the collection says we're done
                    if (getattr(self._props, RuntimeProperties.NUM_MSGS_TO_PUBLISH, 0) > 0 and
                        (self._collection.get_sdk_stat(Metric.TOTAL_MESSAGES_SENT, -1) >=
                         getattr(self._props, RuntimeProperties.NUM_MSGS_TO_PUBLISH, 0))) or \
                            getattr(self._props, RuntimeProperties.CACHE_NUM_REQ, 0) > 0:
                        self._print_stats()
                        break
            else:
                if self._collection is not None and not self._collection.is_connected():
                    self._print_stats()
                    break
        return 0

    def _snapshot_stats(self, is_pub, stat):
        stat_snapshot = (stat, time.time_ns())
        rate_points = self._pub_rate_points if is_pub else self._sub_rate_points

        rate_points.append(stat_snapshot)
        old_snapshot = rate_points[0]
        if is_pub:
            self._pub_rate_points = rate_points if len(rate_points) < 10 else rate_points[1:]
        else:
            self._sub_rate_points = rate_points if len(rate_points) < 10 else rate_points[1:]

        msgs = stat_snapshot[0] - old_snapshot[0]
        nanos = stat_snapshot[1] - old_snapshot[1]
        if nanos <= 0:
            return 0
        else:
            return (msgs * 1000000000) // nanos

    def _print_stats(self):
        print(f"PUB MR(5s)= {self._snapshot_stats(True, self._collection.get_sdk_stat(Metric.TOTAL_MESSAGES_SENT, -1)):5d}, "
              f"SUB MR(5s)= {self._snapshot_stats(False, self._collection.get_perf_stat('rxMsgs', -1)):5d}")

    def _prepare_summary(self):
        rx_msg_rate = 0
        for client in self._collection.get_client_list():
            rx_msg_rate += client.get_rx_stats().get_thru_put(
                self._collection.get_perf_stat("rxMsgs", client._client_id - 1))
        stats = self._collection.get_rx_stats(-1)

        ret_str = f"\n-------------------------------------------------\n"
        ret_str += f"Aggregate Stats (Total # clients: {getattr(self._props, RuntimeProperties.NUM_CLIENTS, 1)}):\n"
        ret_str += f"-------------------------------------------------\n"
        ret_str += f"Total Messages transmitted = {self._collection.get_sdk_stat(Metric.TOTAL_MESSAGES_SENT, -1)}\n"
        ret_str += f"Computed publish rate (msg/sec) = {self._collection.get_tx_throughput():.2f}\n"
        ret_str += f"-------------------------------------------------\n"
        ret_str += f"Total Messages received across all subscribers = {self._collection.get_perf_stat('rxMsgs', -1)}\n"
        ret_str += f"Computed subscriber rate (msg/sec across all subscribers) = {rx_msg_rate:.2f}\n"

        if getattr(self._props, RuntimeProperties.WANT_MESSAGE_PRIORITY_STATS, False):
            perf_stat_tmp = PerfStats("pri_stats", getattr(self._props, RuntimeProperties.LATENCY_BUCKETS, 0),
                                      getattr(self._props, RuntimeProperties.LATENCY_GRANULARITY, 0),
                                      getattr(self._props, RuntimeProperties.LATENCY_WARMUP_IN_SECS, 0))
            for client in self._collection.get_client_list():
                perf_stat_tmp.aggregate(client.get_rx_stats())
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Message Priority Stats:\n"
            if len(perf_stat_tmp.tx_priority) > 0:
                ret_str += f"  Published:\n"
                for k, v in perf_stat_tmp.tx_priority.items():
                    ret_str += f"    Priority #{k}: {v}\n"
            if len(perf_stat_tmp.rx_priority) > 0:
                ret_str += f"  Received:\n"
                for k, v in perf_stat_tmp.rx_priority.items():
                    ret_str += f"    Priority #{k}: {v}\n"

        if getattr(self._props, RuntimeProperties.WANT_PAYLOAD_CHECK, False):
            crc_ok = self._collection.get_perf_stat("crcOk", -1)
            crc_fail = self._collection.get_perf_stat("crcFail", -1)
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Message CRC Stats:\n"
            ret_str += f"  Num Messages CRC OK: {crc_ok}\n"
            ret_str += f"  Num Messages CRC FAIL: {crc_fail}\n"

        if getattr(self._props, RuntimeProperties.SETTLEMENT_PATTERN_STRING, None) is not None:
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"AD Sub ACK/NACK stats:\n"
            ret_str += f"  Messages Received Accepted: {self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_ACCEPTED, -1)}\n"
            ret_str += f"  Messages Received Rejected: {self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_REJECTED, -1)}\n"
            ret_str += f"  Messages Received Failed: {self._collection.get_sdk_stat(Metric.PERSISTENT_MESSSAGES_FAILED, -1)}\n"

        if getattr(self._props, RuntimeProperties.WANT_ORDER_CHECK, False):
            num_ooo_msgs = 0
            num_miss_msgs = 0
            num_dup_msgs = 0

            for k, v in stats._ooo_data.items():
                num_ooo_msgs += len(v.ooo_msg_ids)
                num_miss_msgs += len(v.lost_msg_ids)
                num_dup_msgs += len(v.dup_msg_ids)

            order_checked = self._collection.get_perf_stat("numOrderCheckedMsgs", -1)
            num_redelivered = self._collection.get_perf_stat("numMsgRedelivered", -1)
            num_msg_id_duplicates = self._collection.get_perf_stat('numReplicationGroupMsgIdDuplicates', -1)
            num_msg_spool_changes = self._collection.get_perf_stat('numReplicationGroupMsgIdSpoolIdChanges', -1)
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Order Check Stats:\n"
            ret_str += f"  Num Order Checked Messages: {order_checked}\n"
            ret_str += f"  Num Out Of Order Messages : {num_ooo_msgs}\n"
            ret_str += f"  Total Missing Msgs        : {num_miss_msgs}\n"
            ret_str += f"  Total Duplicate Msgs      : {num_dup_msgs}\n"
            ret_str += f"  Total Redelivered Msgs    : {num_redelivered}\n"
            ret_str += f"  Total Replication Group Message Id Duplicate Msgs: {num_msg_id_duplicates}\n"
            ret_str += f"  Total Replication Group Msg Spool Id Changes: {num_msg_spool_changes}\n"

        if getattr(self._props, RuntimeProperties.WANT_LATENCY, False):
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Latency Info: \n"
            ret_str += f"Individual latency bucket size = {stats.get_latency_bucket_size():.2f} us\n"
            ret_str += f"Latency warmup                 = {getattr(self._props, RuntimeProperties.LATENCY_WARMUP_IN_SECS, 0)} s\n"
            ret_str += f"Latency Messages received      = {stats.get_stat('latMsgs')}\n"
            ret_str += f"Latency msg rate (msg/sec)     = {stats.get_lat_thru_put(stats.get_stat('latMsgs')):.2f}\n"
            ret_str += f"\n"
            ret_str += f"Latency Stats: \n"
            ret_str += f"Minimum latency for subs       = {stats.get_stat('minLat'):.2f} us\n"
            ret_str += f"Average latency for subs       = {stats.get_stat('latAvg'):.2f} us\n"
            ret_str += f"50th   percentile latency      = {stats.get_percentile_latency_in_usec(50)} us\n"
            for perc in [(95, "95th  "), (99, "95th  "), (99.9, "99.9th")]:
                lat = stats.get_percentile_latency_in_usec(perc[0])
                ret_str += f"{perc[1]} percentile latency      = "
                ret_str += "N/A" if lat == 0 else f"{lat} us"
                ret_str += "\n"
            ret_str += f"Maximum latency for subs       = {stats.get_stat('maxLat'):.2f} us\n"
            ret_str += f"Standard Deviation             = {stats.get_latency_std_dev_in_usec():.2f} us\n"

        if getattr(self._props, RuntimeProperties.CACHE_PROP_NAME, None) is not None:
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Cache Stats: \n"
            ret_str += f"Num Requests Sent           = {self._collection.get_sdk_stat(Metric.CACHE_REQUESTS_SENT, -1)}\n"
            ret_str += f"Num Requests Recv           = {stats.get_stat('numCacheRespRecv')}\n"
            ret_str += f"Num Requests Completed OK   = {stats.get_stat('numCacheReqOk')}\n"
            ret_str += f"Num Requests Incomp no data = {stats.get_stat('numCacheReqIncompNoData')}\n"
            ret_str += f"Num Requests Incomp suspect = {stats.get_stat('numCacheReqIncompSuspect')}\n"
            ret_str += f"Num Requests Incomp timeout = {stats.get_stat('numCacheReqIncompTimeout')}\n"
            ret_str += f"Num Requests Errored        = {stats.get_stat('numCacheReqError')}\n"
            ret_str += f"Num Live Data Msgs Recv     = {stats.get_stat('numLiveMsgsRecv')}\n"
            ret_str += f"Num Cached Data Msgs Recv   = {stats.get_stat('numCacheMsgsRecv')}\n"
            ret_str += f"Num Msgs Recv Suspect       = {stats.get_stat('numSuspectRecv')}\n"
            ret_str += f"Reqeust Failed (API Stat)   = {self._collection.get_sdk_stat(Metric.CACHE_REQUESTS_FAILED, -1)}\n"

        if getattr(self._props, RuntimeProperties.WANT_DI_STATS, False):
            broker_di = self._collection.get_perf_stat("numMsgDISet", -1)
            internal_di = self._collection.get_perf_stat("numMsgBpDiSet", -1)
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Discard Indication Stats:\n"
            ret_str += f"  Num msgs received with broker DI = {broker_di}\n"
            ret_str += f"  Num msgs received with internal DI = {internal_di}\n"

        if getattr(self._props, RuntimeProperties.WANT_MSG_OBJ_TYPE_CHECK, False):
            num_msgs_ok = self._collection.get_perf_stat("numMsgsObjTypeOk", -1)
            num_msgs_fail = self._collection.get_perf_stat("numMsgsObjTypeFail", -1)
            ret_str += f"-------------------------------------------------\n"
            ret_str += f"Message Object Type Checking:\n"
            ret_str += f"  Total Messages OK = {num_msgs_ok}\n"
            ret_str += f"  Total Messages Fail = {num_msgs_fail}\n"

        return ret_str

    def _get_prod_listener_str(self):
        prod_stats = ""
        prod_stats += f"-------------------------------------------------\n"
        prod_stats += f"Total persistent messages with confirmed success delivery = " \
                      f"{self._collection.get_perf_stat('adPubSuccess', -1)}\n"
        prod_stats += f"Total persistent messages with confirmed failed delivery = " \
                      f"{self._collection.get_perf_stat('adPubFail', -1)}\n"
        prod_stats += f"Total direct messages with failed publish = " \
                      f"{self._collection.get_perf_stat('directPubFail', -1)}\n"
        return prod_stats

    def shutdown(self, graceful, handler=False):
        if self._collection is not None and self._collection.is_publishing():
            try:
                self._collection.stop_publishing()
            except Exception as e:
                logger.error(e)

        if self._collection is not None and self._collection.is_cache_requesting():
            try:
                self._collection.stop_cache_requesting()
            except Exception as e:
                logger.error(e)

        if self._collection is not None and self._collection.is_connected():
            ped = getattr(self._props, RuntimeProperties.PUBLISH_END_DELAY_IN_SEC, 0)
            if ped > 0:
                logger.info(f"Pausing -ped time to allow clients to finish recv messages ({ped * 1000} ms)")
                time.sleep(ped)

        num_clients = getattr(self._props, RuntimeProperties.NUM_CLIENTS, 1)
        if getattr(self._props, RuntimeProperties.AD_TRANSACTION_SIZE, 0) > 0:
            want_rollback = getattr(self._props, RuntimeProperties.AD_TRANSACTION_ROLLBACK_INTERVAL, -1) == 1
            for i in range(num_clients):
                try:
                    self._collection.commit_transaction(i, self._transacted_session_names[i], want_rollback)
                except Exception as e:
                    logger.error(f"Failed to commit transactions during cleanup: {e}")

            try:
                self._collection.close_all_transacted_sessions(Constants.ALL_CLIENT_INDEX)
            except Exception as e:
                logger.error(f"Failed to close transactions during cleanup: {e}")

        if getattr(self._props, RuntimeProperties.WANT_REMOVE_SUBSCRIBER, True):
            queues = getattr(self._props, RuntimeProperties.SUB_QUEUE_LISTS, [])
            topics = getattr(self._props, RuntimeProperties.SUB_TOPIC_LISTS, [])

            # Remove topics from queue (if given) and then delete that consumer
            if queues:
                for i, client in enumerate(self._collection.get_client_list()):
                    if topics:
                        for queue in queues[i % len(queues)]:
                            self._collection.map_topics(queue, topics[i % len(topics)], False, i)

                    ep_props = EpConfigProperties.create_for_queue_remove(queues[i % len(queues)])
                    self._collection.queue_update(ep_props, i)

            # Remove subscriptions from the default consumer
            if topics:
                for i, client in enumerate(self._collection.get_client_list()):
                    ep_props = EpConfigProperties.create_for_subscription_remove(topics[i % len(topics)], "",
                                                                                 SubscriberDestinationsType.TOPIC)
                    self._collection.subscription_update(ep_props, i)

        final_info = self._prepare_summary()

        if self._collection is not None and self._collection.is_connected():
            try:
                self._collection.disconnect(graceful)
            except Exception as e:
                logger.error(e)

        if getattr(self._props, RuntimeProperties.WANT_PRODUCER_LISTENERS, True):
            final_info += self._get_prod_listener_str()

        if self._collection is not None:
            try:
                self._collection.destroy()
            except Exception as e:
                logger.error(e)

        print(final_info)
        if handler:
            os._exit(1)
        else:
            return 0


def generate_runtime_properties(args):
    parser = CliPropertiesParser()
    return parser.parse_cli(args)


def run():
    if os.getenv('SDKPERF_EXTRA_IMPORT') is not None:
        import importlib.util
        spec = importlib.util.spec_from_file_location("sdkperfpy.extra", os.getenv('SDKPERF_EXTRA_IMPORT'))
        foo = importlib.util.module_from_spec(spec)
        sys.modules["sdkperfpy.extra"] = foo
        spec.loader.exec_module(foo)

    props = generate_runtime_properties(sys.argv[1:])
    sdk = SDKPerf_python(props)
    if sdk.run() == 0:
        exit(0)
    else:
        exit(1)
